Package: nflreadr
Title: Download 'nflverse' Data
Version: 1.5.1
Authors@R: c(
    person("Tan", "Ho", , "tan@tanho.ca", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0001-8388-5155")),
    person("Sebastian", "Carl", , "mrcaseb@gmail.com", role = "aut"),
    person("John", "Edwards", , "edwards1860@gmail.com", role = "ctb"),
    person("Ben", "Baldwin", , "bbaldwin206@gmail.com", role = "ctb"),
    person("Thomas", "Mock", role = "ctb"),
    person("Lee", "Sharpe", role = "ctb"),
    person("Pranav", "Rajaram", , "pranavrajaram@gmail.com", role = "ctb")
  )
Description: A minimal package for downloading data from 'GitHub'
    repositories of the 'nflverse' project.
License: MIT + file LICENSE
URL: https://nflreadr.nflverse.com,
        https://github.com/nflverse/nflreadr
BugReports: https://github.com/nflverse/nflreadr/issues
Depends: R (>= 4.1.0)
Imports: cachem (>= 1.0.0), cli (>= 3.0.0), curl (>= 4.3.0), data.table
        (>= 1.14.0), glue (>= 1.4.0), lifecycle, memoise (>= 2.0.0),
        methods, rappdirs (>= 0.3.0), rlang (>= 0.4.10), tools, utils
Suggests: arrow (>= 6.0.0), covr (>= 3.0.0), DT (>= 0.15.0), fs (>=
        1.5.0), gh (>= 1.0.0), knitr (>= 1.0.0), piggyback (>= 0.1.2),
        progressr (>= 0.8.0), rmarkdown (>= 2.6.0), stringi, testthat
        (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-04-13 18:23:38 UTC; tan
Author: Tan Ho [aut, cre, cph] (ORCID: <https://orcid.org/0000-0001-8388-5155>),
  Sebastian Carl [aut],
  John Edwards [ctb],
  Ben Baldwin [ctb],
  Thomas Mock [ctb],
  Lee Sharpe [ctb],
  Pranav Rajaram [ctb]
Maintainer: Tan Ho <tan@tanho.ca>
Repository: CRAN
Date/Publication: 2026-04-13 19:30:02 UTC
Built: R 4.5.2; ; 2026-04-15 04:58:56 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,nflverse_data)
S3method(print,nflverse_sitrep)
export(.clear_cache)
export(.for_cran)
export(.sitrep)
export(as.nflverse_data)
export(clean_homeaway)
export(clean_player_names)
export(clean_team_abbrs)
export(clear_cache)
export(csv_from_url)
export(ffverse_sitrep)
export(get_current_season)
export(get_current_week)
export(get_latest_season)
export(join_coalesce)
export(load_combine)
export(load_contracts)
export(load_depth_charts)
export(load_draft_picks)
export(load_espn_qbr)
export(load_ff_opportunity)
export(load_ff_playerids)
export(load_ff_rankings)
export(load_from_url)
export(load_ftn_charting)
export(load_injuries)
export(load_nextgen_stats)
export(load_officials)
export(load_participation)
export(load_pbp)
export(load_pfr_advstats)
export(load_pfr_passing)
export(load_player_stats)
export(load_players)
export(load_rosters)
export(load_rosters_weekly)
export(load_schedules)
export(load_snap_counts)
export(load_team_stats)
export(load_teams)
export(load_trades)
export(most_recent_season)
export(nflverse_download)
export(nflverse_game_id)
export(nflverse_releases)
export(nflverse_sitrep)
export(parquet_from_url)
export(progressively)
export(qs_from_url)
export(raw_from_url)
export(rbindlist_with_attrs)
export(rds_from_url)
export(stat_mode)
exportClasses(nflverse_data)
exportClasses(nflverse_sitrep)
importFrom(data.table,`:=`)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# nflreadr <a href='https://nflreadr.nflverse.com'><img src='man/figures/logo.svg' align="right" width="25%" min-width="120px" /></a>

<!-- badges: start -->

[![CRAN
status](https://img.shields.io/cran/v/nflreadr?style=flat-square&logo=R&label=CRAN)](https://CRAN.R-project.org/package=nflreadr)
[![Codecov test
coverage](https://img.shields.io/codecov/c/github/nflverse/nflreadr?label=codecov&style=flat-square&logo=codecov)](https://app.codecov.io/gh/nflverse/nflreadr?branch=main)
[![Dev
status](https://img.shields.io/github/r-package/v/nflverse/nflreadr/main?label=dev%20version&style=flat-square&logo=github)](https://nflreadr.nflverse.com/)
[![Lifecycle:
stable](https://img.shields.io/badge/lifecycle-stable-green.svg?style=flat-square)](https://lifecycle.r-lib.org/articles/stages.html)
[![R build
status](https://img.shields.io/github/actions/workflow/status/nflverse/nflreadr/R-CMD-check.yaml?label=R%20check&style=flat-square&logo=github)](https://github.com/nflverse/nflreadr/actions)
[![nflverse
discord](https://img.shields.io/discord/789805604076126219?color=7289da&label=nflverse%20discord&logo=discord&logoColor=fff&style=flat-square)](https://discord.com/invite/5Er2FBnnQa)

<!-- badges: end -->

nflreadr is a minimal package for downloading data from nflverse
repositories. It includes caching, optional progress updates, and data
dictionaries.

Please note that nflverse data repositories have been reorganized and
pushed towards the
[nflverse-data](https://github.com/nflverse/nflverse-data) repo, and
v1.2.0+ is the minimum version that supports this change. We encourage
all users to upgrade to this version immediately.

For Python access to nflverse data, please check out
[nflreadpy](https://nflreadpy.nflverse.com).

## Installation

Install the stable version from CRAN with:

``` r
install.packages("nflreadr")
```

Install the development version from GitHub with:

``` r
install.packages("nflreadr", repos = c("https://nflverse.r-universe.dev", getOption("repos")))

# or use remotes/devtools
# install.packages("remotes")
remotes::install_github("nflverse/nflreadr")
```

## Usage

The main functions of `nflreadr` are prefixed with `load_`.

``` r
library(nflreadr)

load_pbp(2021)
#> ── nflverse play by play data ──────────────────────────────────────────────────
#> ℹ Data updated: 2026-01-08 14:12:35 CET
#> # A tibble: 49,922 × 372
#>    play_id game_id     old_game_id home_team away_team season_type  week posteam
#>      <dbl> <chr>       <chr>       <chr>     <chr>     <chr>       <int> <chr>  
#>  1       1 2021_01_AR… 2021091207  TEN       ARI       REG             1 <NA>   
#>  2      40 2021_01_AR… 2021091207  TEN       ARI       REG             1 TEN    
#>  3      55 2021_01_AR… 2021091207  TEN       ARI       REG             1 TEN    
#>  4      76 2021_01_AR… 2021091207  TEN       ARI       REG             1 TEN    
#>  5     100 2021_01_AR… 2021091207  TEN       ARI       REG             1 TEN    
#>  6     122 2021_01_AR… 2021091207  TEN       ARI       REG             1 TEN    
#>  7     152 2021_01_AR… 2021091207  TEN       ARI       REG             1 ARI    
#>  8     181 2021_01_AR… 2021091207  TEN       ARI       REG             1 ARI    
#>  9     218 2021_01_AR… 2021091207  TEN       ARI       REG             1 ARI    
#> 10     253 2021_01_AR… 2021091207  TEN       ARI       REG             1 ARI    
#> # ℹ 49,912 more rows
#> # ℹ 364 more variables: posteam_type <chr>, defteam <chr>, side_of_field <chr>,
#> #   yardline_100 <dbl>, game_date <chr>, quarter_seconds_remaining <dbl>,
#> #   half_seconds_remaining <dbl>, game_seconds_remaining <dbl>,
#> #   game_half <chr>, quarter_end <dbl>, …

load_player_stats(2021)
#> ── nflverse player stats: week level ───────────────────────────────────────────
#> ℹ Data updated: 2026-01-08 14:15:10 CET
#> # A tibble: 18,969 × 114
#>    player_id  player_name      player_display_name position position_group
#>    <chr>      <chr>            <chr>               <chr>    <chr>         
#>  1 00-0019596 T.Brady          Tom Brady           QB       QB            
#>  2 00-0022824 A.Lee            Andy Lee            P        SPEC          
#>  3 00-0022924 B.Roethlisberger Ben Roethlisberger  QB       QB            
#>  4 00-0023252 R.Gould          Robbie Gould        K        SPEC          
#>  5 00-0023459 Aa.Rodgers       Aaron Rodgers       QB       QB            
#>  6 00-0023682 R.Fitzpatrick    Ryan Fitzpatrick    QB       QB            
#>  7 00-0023853 M.Prater         Matt Prater         K        SPEC          
#>  8 00-0024243 M.Lewis          Marcedes Lewis      TE       TE            
#>  9 00-0024417 S.Koch           Sam Koch            P        SPEC          
#> 10 00-0025565 N.Folk           Nick Folk           K        SPEC          
#> # ℹ 18,959 more rows
#> # ℹ 109 more variables: headshot_url <chr>, season <int>, week <int>,
#> #   season_type <chr>, team <chr>, opponent_team <chr>, completions <int>,
#> #   attempts <int>, passing_yards <int>, passing_tds <int>, …
```

## Data Sources

Data accessed by this package is stored on GitHub and can typically be
found in one of the following repositories:

- [nflverse/nflverse-data](https://github.com/nflverse/nflverse-data)
- [nflverse/nfldata](https://github.com/nflverse/nfldata)
- [nflverse/espnscrapeR-data](https://github.com/nflverse/espnscrapeR-data)
- [dynastyprocess/data](https://github.com/dynastyprocess/data)
- [ffverse/ffopportunity](https://github.com/ffverse/ffopportunity)

For a full list of functions, please see the [reference
page](https://nflreadr.nflverse.com/reference/index.html).

This data is maintained by the nflverse project team and is primarily
automated via GitHub Actions. You can check the status and schedules
page here: <https://github.com/nflverse/nflverse-data>

## Configuration

The following options help configure default `nflreadr` behaviours.

``` r
options(nflreadr.verbose) 
# TRUE/FALSE to silence messages such as cache warnings
options(nflreadr.cache) 
# one of "memory", "filesystem", or "off"
options(nflreadr.prefer) 
# one of "rds", "parquet", or "csv"
options(nflreadr.download_path)
# defaults to current working directory - change to specify where `nflverse_download()` places data.
```

You can also configure `nflreadr` to display progress messages with the
[`progressr` package](https://progressr.futureverse.org), e.g.

``` r
progressr::with_progress(load_rosters(seasons = 2010:2020))
 |========            |  40%
```

## Getting help

The best places to get help on this package are:

- the [nflverse discord](https://discord.com/invite/5Er2FBnnQa) (for
  both this package as well as anything R/NFL related)
- opening [an
  issue](https://github.com/nflverse/nflreadr/issues/new/choose)

## Contributing

Many hands make light work! Here are some ways you can contribute to
this project:

- You can [open an
  issue](https://github.com/nflverse/nflreadr/issues/new/choose) if
  you’d like to request specific data or report a bug/error.

- If you’d like to contribute code, please check out [the contribution
  guidelines](https://nflreadr.nflverse.com/CONTRIBUTING.html).

## Terms of Use

The R code for this package is released as open source under the [MIT
License](https://nflreadr.nflverse.com/LICENSE.html). NFL data accessed
by this package belong to their respective owners, and are governed by
their terms of use.

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  echo = FALSE,
  comment = "#>"
)
use_dt <- FALSE
if(requireNamespace("DT", quietly = TRUE)) use_dt <- TRUE

## ----eval = use_dt------------------------------------------------------------
# DT::datatable(
#   nflreadr::dictionary_team_stats,
#   options = list(scrollX = TRUE, pageLength = 25),
#   filter = "top",
#   rownames = FALSE
# )

## ----eval = FALSE-------------------------------------------------------------
# #' @inherit nflreadr::load_nextgen_stats
# #' @export
# # Need to add own examples if the function name is different
# #' @examples
# #' load_ng_stats(2020)
# load_ng_stats <- nflreadr::load_nextgen_stats

## ----eval = FALSE-------------------------------------------------------------
# load_rosters <- function(seasons = 1999:2020){
# 
#   # Create a progressor function inside your function that knows how many "steps" there will be
#   p <- progressr::progressor(steps = length(seasons))
# 
#   # Form the URLs to pass into rds_from_url
#   urls <- paste0(
#     "https://github.com/nflverse/nflverse-rosters/",
#     "raw/master/data/seasons/roster_",
#     seasons,
#     ".rds")
# 
#   # Use `progressively()` to wrap the existing function rds_from_url and
#   # call the `p()` progressor function as a signal immediately after each iteration
# 
#   out <- purrr::map_dfr(urls, progressively(rds_from_url,p))
# 
#   return(out)
# }

## ----eval = FALSE-------------------------------------------------------------
# progressr::with_progress(load_rosters(2010:2020))

#' Load Rosters
#'
#' @param seasons a numeric vector of seasons to return, defaults to returning
#' this year's data if it is March or later. If set to `TRUE`, will return all available data.
#' Data available back to 1920.
#' @param file_type One of `c("rds", "csv", "parquet")`. Can also be set globally with
#' `options(nflreadr.prefer)`
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({ # prevents cran errors
#'   load_rosters(2020)
#' })
#' }
#'
#' @return A tibble of season-level roster data.
#'
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_rosters.html> for a web version of the data dictionary
#' @seealso [`dictionary_rosters`] for the data dictionary as a dataframe
#' @seealso Issues with this data should be filed here: <https://github.com/nflverse/nflverse-data>
#'
#' @export
load_rosters <- function(
  seasons = most_recent_season(roster = TRUE),
  file_type = getOption("nflreadr.prefer", default = "rds")
) {
  file_type <- rlang::arg_match0(file_type, c("rds", "csv", "parquet"))
  current_year <- most_recent_season(roster = TRUE)
  if (isTRUE(seasons)) {
    seasons <- 1920:current_year
  }
  stopifnot(is.numeric(seasons), seasons >= 1920, seasons <= current_year)

  urls <- paste0(
    "https://github.com/nflverse/nflverse-data/releases/download/rosters/roster_",
    seasons,
    ".",
    file_type
  )

  out <- load_from_url(urls, seasons = seasons, nflverse = TRUE)

  return(out)
}

#' Load ESPN's QBR
#'
#' @param seasons a numeric vector of seasons to return, data available since 2006. Defaults to latest season available. TRUE will select all seasons.
#' @param summary_type One of `"season"` or `"week"`, defaults to `"season"`
#' @param file_type One of `c("rds", "csv", "parquet")`. Can also be set globally with
#' `options(nflreadr.prefer)`
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#'     load_espn_qbr(2020)
#' }
#'
#' @return a tibble of ESPN QBR data, summarized according to `summary_type`
#'
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_espn_qbr.html> for a web version of the dictionary
#' @seealso [`dictionary_espn_qbr`] for the data dictionary as bundled within the package
#' @seealso Issues with this data should be filed here: <https://github.com/nflverse/espnscrapeR-data>
#'
#' @export
load_espn_qbr <- function(
  seasons = most_recent_season(),
  summary_type = c("season", "week"),
  file_type = getOption("nflreadr.prefer", default = "rds")
) {
  summary_type <- rlang::arg_match0(summary_type, c("season", "week"))
  file_type <- rlang::arg_match0(file_type, c("rds", "csv", "parquet"))

  if (!isTRUE(seasons)) {
    stopifnot(
      is.numeric(seasons),
      seasons >= 2006,
      seasons <= most_recent_season()
    )
  }

  url <- paste0(
    "https://github.com/nflverse/nflverse-data/releases/download/espn_data/qbr_",
    summary_type,
    "_level.",
    file_type
  )
  out <- load_from_url(
    url,
    seasons = seasons,
    nflverse = TRUE,
    nflverse_type = "QBR (via ESPN/espnscrapeR)"
  )
  return(out)
}

#' Load Combine Data from PFR
#'
#' @description Loads combine data since 2000 courtesy of PFR.
#'
#' @param seasons a numeric vector of seasons to return, default `TRUE` returns all available data
#' @param file_type One of `c("rds", "csv", "parquet")`. Can also be set globally with
#' `options(nflreadr.prefer)`
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({ # prevents cran errors
#'   load_combine()
#' })
#' }
#'
#' @return A tibble of NFL combine data provided by Pro Football Reference.
#' @seealso Issues with this data should be filed here: <https://github.com/nflverse/nflverse-data>
#'
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_combine.html> for a web version of the dictionary
#' @seealso [`dictionary_combine`] for the data dictionary as bundled within the package
#'
#' @export
load_combine <- function(
  seasons = TRUE,
  file_type = getOption("nflreadr.prefer", default = "rds")
) {
  file_type <- rlang::arg_match0(file_type, c("rds", "csv", "parquet"))
  url <- glue::glue(
    "https://github.com/nflverse/nflverse-data/releases/download/combine/combine.{file_type}"
  )
  out <- load_from_url(url, seasons = seasons, nflverse = TRUE)
  return(out)
}

#' Get a Situation Report on System, nflverse/ffverse Package Versions and Dependencies
#'
#' This function gives a quick overview of the versions of R and the operating
#' system as well as the versions of nflverse/ffverse packages, options, and their dependencies.
#' It's primarily designed to help you get a quick idea of what's going on when
#' you're helping someone else debug a problem.
#'
#' @param pkg a character vector naming installed packages, or `NULL`
#'   (the default) meaning all nflverse packages. The function checks internally
#'   if all packages are installed and informs if that is not the case.
#' @param recursive a logical indicating whether dependencies of `pkg` and their
#'  dependencies (and so on) should be included.
#'  Can also be a character vector listing the types of dependencies, a subset
#'  of `c("Depends", "Imports", "LinkingTo", "Suggests", "Enhances")`.
#'  Character string `"all"` is shorthand for that vector, character string
#'  `"most"` for the same vector without `"Enhances"`, character string `"strong"`
#'  (default) for the first three elements of that vector.
#' @param redact_path a logical indicating whether options that contain "path"
#' in the name should be redacted, default = TRUE
#' @param dev_repos Developmental cran-like repos to check, e.g. r-universe repos
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({
#' nflverse_sitrep()
#' ffverse_sitrep()
#' .sitrep("cachem")
#' })
#' }
#' @rdname sitrep
#' @export
nflverse_sitrep <- function(
  pkg = c("nflreadr", "nflfastR", "nflseedR", "nfl4th", "nflplotR", "nflverse"),
  recursive = TRUE,
  redact_path = TRUE
) {
  .sitrep(
    pkg = pkg,
    recursive = recursive,
    redact_path = redact_path,
    dev_repos = "https://nflverse.r-universe.dev"
  )
}

#' @rdname sitrep
#' @export
ffverse_sitrep <- function(
  pkg = c("ffscrapr", "ffsimulator", "ffpros", "ffopportunity"),
  recursive = TRUE,
  redact_path = TRUE
) {
  .sitrep(
    pkg = pkg,
    recursive = recursive,
    redact_path = redact_path,
    dev_repos = "https://ffverse.r-universe.dev"
  )
}

#' @rdname sitrep
#' @export
.sitrep <- function(
  pkg,
  recursive = TRUE,
  redact_path = TRUE,
  dev_repos = c(
    "https://nflverse.r-universe.dev",
    "https://ffverse.r-universe.dev"
  )
) {
  out <- structure(
    list(
      system_info = list(),
      installed = list(),
      dependencies = list(),
      package_options = list(),
      not_installed = list(),
      packages = pkg,
      timestamp = Sys.time()
    ),
    class = c("nflverse_sitrep", "list")
  )

  # OPERATING SYSTEM INFO
  out$system_info <- list(
    r_version = R.version$version.string,
    os_version = utils::osVersion
  )
  packages <- pkg

  inst_pkgs <- data.table::as.data.table(utils::installed.packages())

  out$not_installed <- packages[!packages %in% inst_pkgs$Package]
  packages <- packages[packages %in% inst_pkgs$Package]

  if (length(packages) == 0) {
    return(out)
  }

  out$installed <- .sitrep_pkg_status(
    packages = packages,
    inst_pkgs = inst_pkgs,
    dev_repos = dev_repos,
    check_latest = TRUE
  )

  # Relies on options for a given package being prefixed by said package name
  out$package_options <- .sitrep_pkg_opts(packages, redact_path = redact_path)

  # Exit here if we don't want pkg deps
  if (isFALSE(recursive)) {
    return(out)
  }

  out$dependencies <- .sitrep_pkg_deps(
    packages = packages,
    inst_pkgs = inst_pkgs,
    recursive = recursive
  )

  return(out)
}

.sitrep_pkg_status <- function(
  packages,
  inst_pkgs = utils::installed.packages(),
  dev_repos = c(
    "https://nflverse.r-universe.dev",
    "https://ffverse.r-universe.dev"
  ),
  check_latest = curl::has_internet()
) {
  inst_pkgs <- data.table::as.data.table(inst_pkgs)
  Package <- Version <- NULL
  inst <- inst_pkgs[Package %in% packages][, list(
    package = Package,
    installed = Version,
    cran = NA,
    dev = NA,
    behind = NA
  )]
  if (!curl::has_internet() || !check_latest) {
    return(as.data.frame(inst))
  }

  cran_repos <- c(
    "https://packagemanager.posit.co/cran/latest",
    "https://cloud.r-project.org",
    "https://cran.rstudio.com"
  )
  cran_pkgs <- data.table::as.data.table(utils::available.packages(
    repos = cran_repos
  ))[Package %in% packages, list(package = Package, cran = Version)]
  dev_pkgs <- data.table::as.data.table(utils::available.packages(
    repos = dev_repos
  ))[Package %in% packages, list(package = Package, dev = Version)]

  package <- installed <- cran <- dev <- behind_cran <- behind_dev <- behind <- NULL

  inst <- merge(
    merge(
      inst[, list(package, installed)],
      cran_pkgs,
      by = "package",
      all = TRUE
    ),
    dev_pkgs,
    all = TRUE,
    by = "package"
  )[,
    list(
      package,
      installed,
      cran,
      dev,
      behind_cran = ifelse(
        !is.na(cran),
        package_version(installed, strict = FALSE) <
          package_version(cran, strict = FALSE),
        FALSE
      ),
      behind_dev = ifelse(
        !is.na(dev),
        package_version(installed, strict = FALSE) <
          package_version(dev, strict = FALSE),
        FALSE
      )
    )
  ][,
    list(
      package,
      installed,
      cran,
      dev,
      behind = paste0(
        ifelse(behind_cran, "cran;", ""),
        ifelse(behind_dev, "dev", "")
      ),
      behind_cran = NULL,
      behind_dev = NULL
    )
  ]

  return(as.data.frame(inst))
}

#' Show options for installed packages
#' @keywords internal
.sitrep_pkg_opts <- function(packages, redact_path = TRUE) {
  opts <- options()
  pkg_search_string <- paste(packages, collapse = "|")
  package_options <- opts[grepl(pkg_search_string, x = names(opts))]
  if (redact_path) {
    package_options[grepl(
      "path|token|auth|directory",
      names(package_options)
    )] <- "{redacted, use redact_path = FALSE to show}"
  }
  return(package_options)
}

#' Show dependency versions of installed packages
#' @keywords internal
.sitrep_pkg_deps <- function(
  packages,
  inst_pkgs = utils::installed.packages(),
  recursive = TRUE
) {
  inst_pkgs <- data.table::as.data.table(inst_pkgs)
  .flatten <- function(x) sort(unique(unlist(x, use.names = FALSE)))
  deps <- .flatten(tools::package_dependencies(
    packages,
    db = inst_pkgs,
    recursive = recursive
  ))

  deps <- deps[!deps %in% packages]
  missing_pkgs <- setdiff(deps, inst_pkgs$Package)

  dep_status <- inst_pkgs[inst_pkgs$Package %in% deps][, c(
    "Package",
    "Version"
  )]
  data.table::setnames(dep_status, c("package", "version"))

  if (length(missing_pkgs) > 0) {
    dep_status <- data.table::rbindlist(
      list(
        dep_status,
        data.table::data.table(package = missing_pkgs, version = "missing")
      )
    )
  }

  return(as.data.frame(dep_status))
}

#' nflverse_sitrep data class
#'
#' `nflverse_sitrep()` returns an S3 class that was created to allow for custom
#' printing. It will otherwise dispatch to the `list` class.
#'
#' Re-exporting with `methods::setOldClass()` allows these dataframes to be used
#' in S4 generics like those from DBI and jsonlite.
#'
#' @name nflverse_sitrep-class
#' @keywords internal
#' @exportClass nflverse_sitrep
methods::setOldClass("nflverse_sitrep", "list")

#' @export
#' @noRd
print.nflverse_sitrep <- function(x, ...) {
  cli::cat_rule(
    cli::style_bold("System Info"),
    col = cli::make_ansi_style("cyan"),
    line = 1
  )
  cli::cat_bullet(glue::glue(
    "{x$system_info$r_version} {cli::symbol$bullet} Running under: {x$system_info$os_version}"
  )) # nolint

  cli::cat_rule(
    cli::style_bold("Package Status"),
    col = cli::make_ansi_style("cyan"),
    line = 1
  )
  if (nrow(x$installed) > 0) {
    print(x$installed)
  }

  cli::cat_rule(
    cli::style_bold("Package Options"),
    col = cli::make_ansi_style("cyan"),
    line = 1
  )
  if (length(x$package_options) == 0) {
    cli::cat_bullet("No options set for above packages")
  }
  if (length(x$package_options) > 0) {
    .cat_options(names(x$package_options), unname(x$package_options))
  }

  if (length(x$dependencies) >= 1) {
    cli::cat_rule(
      cli::style_bold("Package Dependencies"),
      col = cli::make_ansi_style("cyan"),
      line = 1
    )
    .cat_pkg(x$dependencies$package, x$dependencies$version)
  }

  if (length(x$not_installed) >= 1) {
    cli::cat_rule(
      cli::style_bold("Not Installed"),
      col = cli::make_ansi_style("cyan"),
      line = 1
    )
    .cat_pkg(x$not_installed, rep_len("", length(x$not_installed)))
  }

  cli::cat_rule(col = cli::make_ansi_style("cyan"), line = 1)

  return(invisible(x))
}

.cat_pkg <- function(packages, versions) {
  stopifnot(length(packages) == length(versions))

  if (length(packages) <= 2) {
    cli::cat_bullet(glue::glue("{format(packages)} ({format(versions)})"))
    return(invisible())
  }

  l <- length(packages)

  breaks <- cut(
    x = seq_along(packages),
    breaks = c(0, ceiling(l / 3), 2 * ceiling(l / 3), l + 1)
  )

  p <- split(packages, breaks)
  v <- split(versions, breaks)

  r <- length(p[[1]]) - length(p[[3]])

  if (r != 0) {
    p[[3]] <- c(p[[3]], rep("", r))
    v[[3]] <- c(v[[3]], rep("", r))
  }

  p <- lapply(p, function(x) {
    ifelse(x != "", format(paste(cli::symbol$bullet, x)), "")
  })
  v <- lapply(v, function(x) ifelse(x != "", format(paste0(" (", x, ")")), ""))

  cli::cat_line(
    paste0(
      p[[1]],
      v[[1]],
      "  ",
      p[[2]],
      v[[2]],
      "  ",
      p[[3]],
      v[[3]],
      "  "
    )
  )
}

.cat_options <- function(option_names, option_values) {
  stopifnot(length(option_names) == length(option_values))
  cli::cat_bullet(glue::glue("{format(option_names)}: {format(option_values)}"))
}

test_that("sitrep works", {
  skip_on_cran()
  nflreadr_sitrep <- .sitrep(c("nflreadr", "data.table"), recursive = FALSE)
  expect_named(
    nflreadr_sitrep,
    c(
      "system_info",
      "installed",
      "dependencies",
      "package_options",
      "not_installed",
      "packages",
      "timestamp"
    )
  )

  xyz_sitrep <- .sitrep("xyz", recursive = FALSE)

  expect_true(
    all(
      lengths(xyz_sitrep[c("installed", "dependencies", "package_options")]) ==
        0
    )
  )

  expect_output(
    print(nflreadr_sitrep),
    regexp = "System Info|R version |Running under|Packages|nflreadr \\("
  )
})

#' Load Weekly Rosters
#'
#' Returns week level rosters (rather than latest for a given season as returned by `load_rosters()`)
#'
#' @param seasons a numeric vector of seasons to return, defaults to returning
#' this year's data if it is March or later. If set to `TRUE`, will return all available data.
#' Data available back to 2002.
#' @param file_type One of `c("rds", "csv", "parquet")`. Can also be set globally with
#' `options(nflreadr.prefer)`
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({ # prevents cran errors
#'   load_rosters_weekly(2020)
#' })
#' }
#'
#' @return A tibble of weekly roster data.
#'
#' @seealso Issues with this data should be filed here: <https://github.com/nflverse/nflverse-data>
#'
#' @export
load_rosters_weekly <- function(
  seasons = most_recent_season(roster = TRUE),
  file_type = getOption("nflreadr.prefer", default = "rds")
) {
  file_type <- rlang::arg_match0(file_type, c("rds", "csv", "parquet"))

  current_year <- most_recent_season(roster = TRUE)

  if (isTRUE(seasons)) {
    seasons <- 2002:current_year
  }

  stopifnot(is.numeric(seasons), seasons >= 2002, seasons <= current_year)

  urls <- paste0(
    "https://github.com/nflverse/nflverse-data/releases/download/weekly_rosters/roster_weekly_",
    seasons,
    ".",
    file_type
  )

  out <- load_from_url(urls, seasons = seasons, nflverse = TRUE)

  return(out)
}

# nocov start
.onLoad <- function(libname, pkgname) {
  memoise_option <- getOption("nflreadr.cache", default = "memory")

  if (!memoise_option %in% c("memory", "filesystem", "off")) {
    memoise_option <- "memory"
  }

  if (memoise_option == "filesystem") {
    cache_dir <- rappdirs::user_cache_dir(appname = "nflreadr")
    dir.create(cache_dir, recursive = TRUE, showWarnings = FALSE)
    cache <- cachem::cache_disk(dir = cache_dir)
  }

  if (memoise_option == "memory") {
    cache <- cachem::cache_mem()
  }

  if (memoise_option != "off") {
    assign(
      x = "rds_from_url",
      value = memoise::memoise(
        rds_from_url,
        ~ memoise::timeout(86400),
        cache = cache
      ),
      envir = rlang::ns_env("nflreadr")
    )
    assign(
      x = "csv_from_url",
      value = memoise::memoise(
        csv_from_url,
        ~ memoise::timeout(86400),
        cache = cache
      ),
      envir = rlang::ns_env("nflreadr")
    )
    assign(
      x = "raw_from_url",
      value = memoise::memoise(
        raw_from_url,
        ~ memoise::timeout(86400),
        cache = cache
      ),
      envir = rlang::ns_env("nflreadr")
    )
    assign(
      x = "parquet_from_url",
      value = memoise::memoise(
        parquet_from_url,
        ~ memoise::timeout(86400),
        cache = cache
      ),
      envir = rlang::ns_env("nflreadr")
    )
  }

  if (
    is.null(getOption("nflreadr.verbose")) &&
      isFALSE(getOption("nflreadr.cache_warning"))
  ) {
    options("nflreadr.verbose" = TRUE)
  }

  # validate suggested package installation
  if (
    !is.null(getOption("nflreadr.prefer")) &&
      getOption("nflreadr.prefer") %in% c("parquet")
  ) {
    switch(
      getOption("nflreadr.prefer"),
      "parquet" = {
        rlang::check_installed(
          pkg = "arrow (>= 6.0.0)",
          reason = "these packages are required to use options(nflreadr.prefer = 'parquet')"
        )

        options("nflreadr.arrow_ok" = TRUE)
      }
    )
  }
}

.onAttach <- function(libname, pkgname) {
  # validate nflreadr.prefer

  prefer_type <- getOption("nflreadr.prefer", default = "rds")

  if (!prefer_type %in% c("rds", "parquet", "csv")) {
    packageStartupMessage(
      "Note: Option nflreadr.prefer is set to ",
      prefer_type,
      ' and should be one of "parquet", "rds", or "csv". \n',
      'Defaulting back to "rds"'
    )

    options("nflreadr.prefer" = "rds")
  }

  # validate nflreadr.cache

  memoise_option <- getOption("nflreadr.cache", default = "memory")

  if (!memoise_option %in% c("memory", "filesystem", "off")) {
    packageStartupMessage(
      'Note: nflreadr.cache is set to "',
      memoise_option,
      '" and should be one of c("memory","filesystem", "off"). \n',
      'Defaulting to "memory".'
    )

    options("nflreadr.cache" = "memory")
  }

  if (memoise_option == "off") {
    packageStartupMessage('Note: nflreadr.cache is set to "off"')
  }
}

# nocov end

#' Bulk download utilities via piggyback
#'
#' This function downloads or updates data from the nflverse-data repository releases,
#' creating subfolders that match the release structure.
#'
#' @param ... releases to download, provided in either unquoted or character format
#' (i.e. pbp or "pbp" are both fine). Available release names can be listed with
#' [`nflverse_releases()`]
#' @param folder_path a folder in which subfolders will be created for each release -
#' defaults to path specified in `options(nflreadr.download_path)` or "." (the current working directory)
#' @param file_type one of `c("rds","parquet", "csv")` -
#' defaults to file type specified in `options(nflreadr.prefer)` or "rds"
#' @param use_hive whether to create hive-style partition folders for each season, e.g. `"~/pbp/.season=2021/pbp.csv"`
#' @param .token a GitHub API token, `"default"` uses `gh::gh_token()`
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({
#'   ## could also set options like
#'   # options(nflreadr.download_path = tempdir(), nflreadr.prefer = "parquet")
#'
#'   nflverse_download(combine, contracts, folder_path = tempdir(), file_type = "parquet")
#'
#'   list.files(tempdir(),pattern = ".parquet$") # check that files were downloaded!
#' })
#' }
#' @export
nflverse_download <- function(
  ...,
  folder_path = getOption("nflreadr.download_path", default = "."),
  file_type = getOption("nflreadr.prefer", default = "rds"),
  use_hive = file_type %in% c("parquet", "csv"),
  .token = "default"
) {
  rlang::check_installed(c("piggyback (>= 0.1.2)", "fs", "gh"))

  if (.token == "default") {
    .token <- gh::gh_token()
  }

  releases <- as.character(rlang::ensyms(...))
  file_type <- rlang::arg_match0(file_type, c("rds", "csv", "parquet"))

  stopifnot(
    length(folder_path) == 1,
    is.character(folder_path),
    length(use_hive) == 1,
    is.logical(use_hive)
  )

  all_releases <- piggyback::pb_releases(
    repo = "nflverse/nflverse-data",
    verbose = FALSE,
    .token = .token
  )

  if (!isTRUE(releases) && any(!releases %in% all_releases$tag_name)) {
    missing <- releases[!releases %in% all_releases$tag_name]
    cli::cli_warn(
      "Could not find {.val {missing}} in nflverse-data releases. Skipping."
    )
    releases <- releases[releases %in% all_releases$tag_name]
  }

  if (isTRUE(releases)) {
    releases <- all_releases$tag_name
  }

  if (length(releases) == 0) {
    cli::cli_abort("No matching releases to download.")
  }

  fs::dir_create(file.path(folder_path, releases))

  file_list <- piggyback::pb_list(
    repo = "nflverse/nflverse-data",
    tag = releases,
    .token = .token
  )

  data.table::setDT(file_list)

  file_name <- NULL
  tag <- NULL
  path <- NULL
  .season <- NULL

  download_list <- file_list[
    grepl(paste0("\\.", file_type, "$"), file_name),
    list(file_name, tag, path = file.path(folder_path, tag))
  ]

  if (use_hive) {
    download_list <- download_list[,
      `:=`(
        .season = gsub(
          x = file_name,
          pattern = ".+([0-9]{4}).+",
          replacement = "\\1"
        )
      )
    ][, `:=`(
      path = ifelse(
        .season == file_name,
        path,
        file.path(path, paste0("season=", .season))
      )
    )]

    fs::dir_create(download_list$path)
  }

  if (any(!releases %in% download_list$tag)) {
    missing <- releases[!releases %in% download_list$tag]
    cli::cli_warn(
      "Could not find files of type {.val {file_type}} for the following {cli::qty(missing)} release{?s}: {.val {missing}}. Please try another file type."
    )
  }

  cli::cli_alert_info(
    "Now downloading {.val {nrow(download_list)}} file{?s} to {.file {folder_path}}."
  )

  lapply(
    split(x = download_list, by = "path", drop = TRUE, flatten = FALSE),
    function(x) {
      piggyback::pb_download(
        file = x$file_name,
        dest = x$path[[1]],
        repo = "nflverse/nflverse-data",
        tag = x$tag[[1]],
        overwrite = TRUE,
        .token = .token
      )
      invisible(NULL)
    }
  )

  cli::cli_alert_success(
    "Downloaded {.val {nrow(download_list)}} file{?s} to {.file {folder_path}}."
  )

  return(invisible(download_list))
}


#' List all available nflverse releases
#'
#' @description This functions lists all nflverse data releases that are available
#'   in the nflverse-data repo. Release names can be used for downloads in
#'   [`nflverse_download()`].
#'
#' @param .token a GitHub API token, `"default"` uses `gh::gh_token()`
#'
#' @examples
#' \dontshow{
#' .for_cran()
#' # Change option for better output
#' old <- options(piggyback.verbose = FALSE)
#' }
#' \donttest{
#'
#' try( # avoids cran failures, can skip in normal usage
#' nflverse_releases()
#' )
#' }
#' \dontshow{
#' # Restore old options
#' options(old)
#' }
#'
#' @return A dataframe containing release names, release descriptions, and
#'   other relevant release information.
#' @export
nflverse_releases <- function(.token = "default") {
  rlang::check_installed("piggyback (>= 0.1.2)", "gh")
  if (.token == "default") {
    # we don't set the default .token arg to gh::gh_token because
    # gh is a suggested dependency and not necessarily installed
    .token <- gh::gh_token()
  }
  releases <- piggyback::pb_releases(
    repo = "nflverse/nflverse-data",
    .token = .token
  )
  assets <- piggyback::pb_list(repo = "nflverse/nflverse-data", .token = .token)
  data.table::setDT(assets)

  # define variables to prevent NSE check NOTE
  timestamp <- NULL
  file_name <- NULL
  tag <- NULL

  release_summary <- assets[,
    list(
      timestamp = format(max(timestamp), tz = "America/Toronto", usetz = TRUE),
      rds = sum(grepl("rds$", file_name)),
      parquet = sum(grepl("parquet$", file_name)),
      csv = sum(grepl("csv$", file_name)),
      csv_gz = sum(grepl("csv.gz$", file_name)),
      zip = sum(grepl("zip$", file_name))
    ),
    by = tag
  ][order(timestamp, decreasing = TRUE)]

  out <- data.table::data.table(
    release_name = releases$tag_name,
    release_description = gsub("[\r\n]", "", releases$release_body)
  )[release_summary, on = c("release_name" = "tag")]

  out <- make_nflverse_data(out, "release listing")

  return(out)
}

#' Load NFL Team Graphics, Colors, and Logos
#'
#' @description Loads team graphics, colors, and logos - useful for plots!
#'
#' @param current If `TRUE` (the default), returns a standardized list of current teams only,
#' with abbreviations as per [nflreadr::team_abbr_mapping].
#' @param file_type One of `c("rds", "csv", "parquet")`. Can also be set globally with
#' `options(nflreadr.prefer)`
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({ # prevents cran errors
#'   load_teams()
#' })
#' }
#'
#' @return A tibble of team-level image URLs and hex color codes.
#'
#' @seealso Issues with this data should be filed here: <https://github.com/nflverse/nflverse-pbp>
#'
#' @export
load_teams <- function(
  current = TRUE,
  file_type = getOption("nflreadr.prefer", default = "rds")
) {
  file_type <- rlang::arg_match0(file_type, c("rds", "csv", "parquet"))

  urls <- paste0(
    "https://github.com/nflverse/nflverse-data/releases/download/teams/",
    "teams_colors_logos.",
    file_type
  )

  out <- load_from_url(urls, nflverse = TRUE)

  if (isTRUE(current)) {
    out <- out[out$team_abbr %in% nflreadr::team_abbr_mapping, ]
  }
  out
}

#' Load Play By Play
#'
#' @description Loads play by play seasons from the [nflverse-data repository](https://github.com/nflverse/nflverse-data)
#'
#' @param seasons A numeric vector of 4-digit years associated with given NFL seasons - defaults to latest season. If set to `TRUE`, returns all available data since 1999.
#' @param file_type One of `c("rds", "csv", "parquet")`. Can also be set globally with
#' `options(nflreadr.prefer)`
#'
#' @return The complete nflfastR dataset as returned by `nflfastR::build_nflfastR_pbp()`
#' (see below) for all given `seasons`
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({ # prevents cran errors
#'   load_pbp(2019:2020)
#' })
#' }
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_pbp.html> for a web version of the data dictionary
#' @seealso [`dictionary_pbp`] for the data dictionary bundled as a package dataframe
#' @seealso <https://nflfastr.com/reference/build_nflfastR_pbp.html> for the nflfastR function `nflfastR::build_nflfastR_pbp()`
#' @seealso Issues with this data should be filed here: <https://github.com/nflverse/nflverse-pbp>
#'
#' @export
load_pbp <- function(
  seasons = most_recent_season(),
  file_type = getOption("nflreadr.prefer", default = "rds")
) {
  file_type <- rlang::arg_match0(file_type, c("rds", "csv", "parquet"))

  if (isTRUE(seasons)) {
    seasons <- 1999:most_recent_season()
  }

  stopifnot(
    is.numeric(seasons),
    seasons >= 1999,
    seasons <= most_recent_season()
  )

  urls <- paste0(
    "https://github.com/nflverse/nflverse-data/releases/download/pbp/play_by_play_",
    seasons,
    ".",
    file_type
  )

  out <- load_from_url(urls, nflverse = TRUE, seasons = seasons)

  return(out)
}

#' Compute nflverse Game Identifiers
#'
#' @param season 4 digit season between 1999 and the output of [most_recent_season()]
#' @param week Numeric or character giving the week, between 1 and 22.
#' @param home,away Valid NFL team abbreviation as it can be found in [team_abbr_mapping]
#'
#' @return A character vector
#' @export
#'
#' @examples
#' \dontshow{.for_cran()}
#' nflverse_game_id(2022, 2, "LAC", "KC")
nflverse_game_id <- function(season, week, away, home) {
  # catch character input of season and week
  season <- as.numeric(season)
  week <- as.numeric(week)

  if (!all(season %in% 1999:most_recent_season(roster = TRUE))) {
    cli::cli_abort(
      "{.arg season} must be between 1999 and {most_recent_season(roster = TRUE)}"
    )
  }

  if (!all(week %in% 1:22)) {
    cli::cli_abort("{.arg week} must be between 1 and 22")
  }

  valid_names <- names(nflreadr::team_abbr_mapping_norelocate)

  if (!all(home %in% valid_names)) {
    invalid <- unique(home[!home %in% valid_names])
    cli::cli_abort(
      "{.val {invalid}} {?is an/are} invalid {.arg home} abbreviation{?s}."
    )
  }

  if (!all(away %in% valid_names)) {
    invalid <- unique(away[!away %in% valid_names])
    cli::cli_abort(
      "{.val {invalid}} {?is an/are} invalid {.arg away} abbreviation{?s}."
    )
  }

  away <- clean_team_abbrs(away, current_location = FALSE)
  home <- clean_team_abbrs(home, current_location = FALSE)

  arg_lengths <- lengths(list(season, week, away, home))

  if (!all(arg_lengths %in% c(1, max(arg_lengths)))) {
    cli::cli_warn(
      "Arguments should be of of length one or of length {max(arg_lengths)}. Recycling may produce unexpected results."
    )
  }

  paste(
    season,
    formatC(week, width = 2, flag = "0"),
    away,
    home,
    sep = "_"
  )
}

#' Progressively
#'
#' This function helps add progress-reporting to any function - given function `f()` and progressor `p()`,
#' it will return a new function that calls `f()` and then (on exiting) will call `p()` after every iteration.
#' This is inspired by purrr's `safely`, `quietly`, and `possibly` function decorators.
#'
#' @param f a function to add progressor functionality to.
#' @param p a function such as one created by `progressr::progressor()` - also accepts purrr-style lambda functions.
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({ # prevents cran errors
#'
#' urls <- rep("https://github.com/nflverse/nflverse-data/releases/download/test/combines.csv",3)
#'
#' lapply(urls, progressively(read.csv, ~cli::cli_progress_step('Loading...')))
#'
#' read_rosters <- function(urls){
#'   p <- progressr::progressor(along = urls)
#'   lapply(urls, progressively(read.csv, p))
#' }
#'
#' progressr::with_progress(read_rosters())
#'
#' })
#' }
#'
#' @return a function that does the same as `f` but it calls `p()` after iteration.
#'
#' @seealso <https://nflreadr.nflverse.com/articles/exporting_nflreadr.html> for vignette on exporting nflreadr in packages
#'
#' @export
progressively <- function(f, p = NULL) {
  if (is.null(p)) {
    p <- function(...) NULL
  }
  p <- rlang::as_function(p)
  f <- rlang::as_function(f)

  function(...) {
    on.exit(p())
    f(...)
  }
}

#' Statistical Mode
#'
#' Computes the statistical mode, i.e. the value that appears most often in a
#' vector. Returns the first match, if TRUE for multiple values.
#'
#' @param x A vector of data values.
#' @param ... Further arguments, currently unused.
#' @param na.rm a logical evaluating to `TRUE` or `FALSE` indicating whether
#'   `NA` values should be stripped before the computation proceeds.
#'
#' @return The statistical mode with the same type as the input vector x.
#' @export
#'
#' @examples
#' vector_numeric <- sample(1:5, 15, TRUE)
#' vector_numeric
#' stat_mode(vector_numeric)
#'
#' vector_character <- sample(LETTERS[1:5], 15, TRUE)
#' vector_character
#' stat_mode(vector_character)
stat_mode <- function(x, ..., na.rm = FALSE) {
  if (isTRUE(na.rm)) {
    x <- x[!is.na(x)]
  }
  ux <- unique(x)
  tbl <- tabulate(match(x, ux))
  ux[which.max(tbl)]
}

test_that("mode works", {
  # I am sure cran will break my seed somehow
  skip_on_cran()
  # need to set a seed for sampling
  set.seed(20230928)

  # test numerical values
  vector_numeric <- sample(1:5, 15, TRUE)
  mode_numeric <- stat_mode(vector_numeric)
  expect_identical(mode_numeric, 3L)

  # test numerical values with NA
  vector_numeric_na <- c(1, 3, 3, NA, NA, NA)
  mode_numeric_na <- stat_mode(vector_numeric_na)
  mode_numeric_na_removed <- stat_mode(vector_numeric_na, na.rm = TRUE)
  expect_identical(mode_numeric_na, NA_real_)
  expect_identical(mode_numeric_na_removed, 3)

  # test character values
  vector_character <- sample(LETTERS[1:5], 15, TRUE)
  mode_character <- stat_mode(vector_character)
  expect_identical(mode_character, "D")

  # test Dates
  vector_dates <- as.Date(c(
    "2023-09-28",
    "2023-09-29",
    "2023-09-29",
    "2023-09-28",
    "2023-09-28"
  ))
  mode_date <- stat_mode(vector_dates)
  expect_identical(mode_date, as.Date("2023-09-28"))
})

#' Load Snap Counts from PFR
#'
#' @description Loads game level snap counts stats provided by Pro Football Reference
#' starting with the 2012 season.
#'
#' @param seasons a numeric vector specifying what seasons to return, if `TRUE` returns all available data
#' @param file_type One of `c("rds", "csv", "parquet")`. Can also be set globally with
#' `options(nflreadr.prefer)`
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#'   try({ # prevents CRAN errors
#'   load_snap_counts()
#'   })
#' }
#'
#' @return A tibble of game-level snap counts provided by Pro Football Reference.
#'
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_snap_counts.html> for the web data dictionary
#' @seealso [`dictionary_snap_counts`] for the data dictionary as bundled within the package
#' @seealso Issues with this data should be filed here: <https://github.com/nflverse/nflverse-pfr>
#'
#' @export
load_snap_counts <- function(
  seasons = most_recent_season(),
  file_type = getOption("nflreadr.prefer", default = "rds")
) {
  file_type <- rlang::arg_match0(file_type, c("rds", "csv", "parquet"))
  if (isTRUE(seasons)) {
    seasons <- 2012:most_recent_season()
  }
  stopifnot(
    is.numeric(seasons),
    seasons >= 2012,
    seasons <= most_recent_season()
  )

  urls <- paste0(
    "https://github.com/nflverse/nflverse-data/releases/",
    "download/snap_counts/snap_counts_",
    seasons,
    ".",
    file_type
  )

  out <- load_from_url(urls, seasons = seasons, nflverse = TRUE)
  out
}

#' Load Game/Schedule Data
#'
#' This returns game/schedule information as maintained by Lee Sharpe.
#'
#' @param seasons a numeric vector of seasons to return, default `TRUE` returns all available data.
#'
#' @return A tibble of game information for past and/or future games.
#'
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_schedules.html> for a web version of the data dictionary
#' @seealso [`dictionary_schedules`] for the data dictionary as a dataframe
#' @seealso Issues with this data should be filed here: <https://github.com/nflverse/nfldata>
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({ # prevents cran errors
#'  load_schedules(2020)
#' })
#' }
#'
#' @export
load_schedules <- function(seasons = TRUE) {
  if (!isTRUE(seasons)) {
    stopifnot(is.numeric(seasons))
  }
  out <- load_from_url(
    "https://github.com/nflverse/nfldata/raw/master/data/games.rds",
    seasons = seasons,
    nflverse = TRUE,
    nflverse_type = "games and schedules"
  )
  # nflfastR and nfl4th use the "roof" variable in several models
  # These models expect one of c("closed", "outdoors", "open", "retractable") or NA
  # We ran into problems where roof was empty or something else several time
  # So we catch these cases here and change invalid data to NA
  valid_roof_values <- c(
    "closed",
    "dome",
    "outdoors",
    "open",
    "retractable",
    NA_character_
  )
  out$roof[!out$roof %in% valid_roof_values] <- NA_character_

  return(out)
}

#' Load Player Level Stats
#'
#' @param seasons a numeric vector of seasons to return, defaults to most recent
#' season. If set to `TRUE`, returns all available data.
#' @inheritParams rlang::args_dots_empty
#' @param summary_level choice: one of week (default), "reg" for regular season,
#' "post" for postseason, "reg+post" for combined regular season + postseason stats
#' @param file_type choice: one of `c("rds", "csv", "parquet")`. Can also
#' be set globally with `options(nflreadr.prefer)`
#' @param stat_type `r lifecycle::badge("deprecated")` - now returns all stat
#' types by default
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({
#'   load_player_stats()
#' })
#' }
#'
#' @return A tibble of player statistics that aims to match NFL official box
#' scores and season summaries
#'
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_player_stats.html>
#' for a web version of the data dictionary
#' @seealso [`dictionary_player_stats`] for the data dictionary
#'
#' @export
load_player_stats <- function(
  seasons = most_recent_season(),
  ...,
  summary_level = c("week", "reg", "post", "reg+post"),
  file_type = getOption("nflreadr.prefer", default = "rds"),
  stat_type = lifecycle::deprecated()
) {
  # deprecate stat_type arg
  if (lifecycle::is_present(stat_type)) {
    lifecycle::deprecate_warn(
      when = "1.5.0",
      what = "load_player_stats(stat_type)",
      with = "load_player_stats()"
    )
  }

  load_stats(
    .stat_type = "player",
    seasons = seasons,
    ...,
    summary_level = summary_level,
    file_type = file_type
  )
}

#' Load Team Level Stats
#'
#' @param seasons a numeric vector of seasons to return, defaults to most recent
#' season. If set to `TRUE`, returns all available data.
#' @inheritParams rlang::args_dots_empty
#' @param summary_level choice: one of week (default), "reg" for regular season,
#' "post" for postseason, "reg+post" for combined regular + post season stats
#' @param file_type choice: one of `c("rds", "csv", "parquet")`. Can also
#' be set globally with `options(nflreadr.prefer)`
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({
#'   load_team_stats()
#' })
#' }
#'
#' @return A tibble of team statistics that aims to match NFL official box scores
#' and season summaries
#'
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_team_stats.html>
#' for a web version of the data dictionary
#' @seealso [`dictionary_team_stats`] for the data dictionary
#'
#' @export
load_team_stats <- function(
  seasons = most_recent_season(),
  ...,
  summary_level = c("week", "reg", "post", "reg+post"),
  file_type = getOption("nflreadr.prefer", default = "rds")
) {
  load_stats(
    .stat_type = "team",
    seasons = seasons,
    ...,
    summary_level = summary_level,
    file_type = file_type
  )
}

load_stats <- function(
  .stat_type = c("player", "team"),
  seasons,
  ...,
  summary_level = c("week", "reg", "post", "reg+post"),
  file_type = getOption("nflreadr.prefer", default = "rds")
) {
  # ensure that all function args except seasons and .stat_type are named
  rlang::check_dots_empty()

  .stat_type <- rlang::arg_match(.stat_type)
  summary_level <- rlang::arg_match(summary_level)
  file_type <- rlang::arg_match0(file_type, c("rds", "csv", "parquet"))

  if (!isTRUE(seasons)) {
    stopifnot(
      is.numeric(seasons),
      seasons >= 1999,
      seasons <= most_recent_season()
    )
  }

  if (isTRUE(seasons)) {
    seasons <- 1999:most_recent_season()
  }

  summary_level <- switch(
    summary_level,
    "week" = "week_",
    "reg" = "reg_",
    "post" = "post_",
    "reg+post" = "regpost_"
  )

  url <- paste0(
    "https://github.com/nflverse/nflverse-data/releases/download/",
    "stats_",
    .stat_type,
    "/stats_",
    .stat_type,
    "_",
    summary_level,
    seasons,
    ".",
    file_type
  )

  out <- load_from_url(url, seasons = seasons, nflverse = TRUE)
  return(out)
}

test_that("load_pbp", {
  skip_on_cran()
  skip_if_offline("github.com")

  pbp <- load_pbp()
  pbp_years <- load_pbp(2019:2020)

  expect_s3_class(pbp, "nflverse_data")
  expect_s3_class(pbp_years, "nflverse_data")

  expect_gt(nrow(pbp_years), 90000)
})
test_that("load_participation", {
  skip_on_cran()
  skip_if_offline("github.com")

  participation <- load_participation(2019)
  pbp_participation <- load_participation(
    2020,
    include_pbp = TRUE,
    file_type = "parquet"
  )

  expect_s3_class(participation, "nflverse_data")
  expect_s3_class(pbp_participation, "nflverse_data")

  expect_gt(ncol(participation), 5)
  expect_gt(ncol(pbp_participation), 300)
  expect_gt(nrow(participation), 40000)
  expect_gt(nrow(pbp_participation), 40000)
})

test_that("load_player_stats", {
  skip_on_cran()
  skip_if_offline("github.com")

  ps <- load_player_stats()
  ps_2020 <- load_player_stats(2020)
  ps_csv <- load_player_stats(2020, file_type = "csv")
  ps_all <- load_player_stats(seasons = TRUE)
  ps_reg <- load_player_stats(2020, summary_level = "reg")
  ps_regpost <- load_player_stats(2020, summary_level = "reg+post")

  expect_s3_class(ps, "nflverse_data")
  expect_s3_class(ps_2020, "nflverse_data")
  expect_s3_class(ps_csv, "nflverse_data")
  expect_s3_class(ps_all, "nflverse_data")
  expect_s3_class(ps_reg, "nflverse_data")
  expect_s3_class(ps_regpost, "nflverse_data")

  expect_gt(nrow(ps_2020), 17000)
  expect_gt(nrow(ps_csv), 17000)
  expect_gt(nrow(ps_all), 400000)
  expect_gt(nrow(ps_reg), 1900)
  expect_gt(nrow(ps_regpost), 1900)
})

test_that("load_team_stats", {
  skip_on_cran()
  skip_if_offline("github.com")

  ts <- load_team_stats()
  ts_2020 <- load_team_stats(2020)
  ts_csv <- load_team_stats(2020, file_type = "csv")
  ts_all <- load_team_stats(seasons = TRUE)
  ts_reg <- load_team_stats(2020, summary_level = "reg")
  ts_regpost <- load_team_stats(2020, summary_level = "reg+post")

  expect_s3_class(ts, "nflverse_data")
  expect_s3_class(ts_2020, "nflverse_data")
  expect_s3_class(ts_csv, "nflverse_data")
  expect_s3_class(ts_all, "nflverse_data")
  expect_s3_class(ts_reg, "nflverse_data")
  expect_s3_class(ts_regpost, "nflverse_data")

  expect_gt(nrow(ts_2020), 500)
  expect_gt(nrow(ts_csv), 500)
  expect_gt(nrow(ts_all), 13000)
  expect_equal(nrow(ts_reg), 32)
  expect_equal(nrow(ts_regpost), 32)
})

test_that("load_schedules", {
  skip_on_cran()
  skip_if_offline("github.com")

  schedules <- load_schedules()

  expect_s3_class(schedules, "nflverse_data")

  expect_gt(nrow(schedules), 6000)

  valid_roof_values <- c(
    "closed",
    "dome",
    "outdoors",
    "open",
    "retractable",
    NA_character_
  )
  expect_true(all(schedules$roof %in% valid_roof_values))

  expect_error(load_schedules("2020"))

  sched_2020 <- load_schedules(2020)

  expect_s3_class(schedules, "nflverse_data")

  expect_lte(nrow(sched_2020), 269)
})

test_that("load_rosters", {
  skip_on_cran()
  skip_if_offline("github.com")

  rosters <- load_rosters()
  rosters_years <- load_rosters(seasons = 2019:2020)

  expect_s3_class(rosters, "nflverse_data")
  expect_s3_class(rosters_years, "nflverse_data")
  expect_gt(nrow(rosters), 2000)
  expect_gt(nrow(rosters_years), 5000)
})

test_that("load_rosters_weekly", {
  skip_on_cran()
  skip_if_offline("github.com")

  rosters <- load_rosters_weekly(
    seasons = nflreadr::most_recent_season(roster = FALSE)
  )
  rosters_years <- load_rosters_weekly(seasons = 2019:2020)

  expect_s3_class(rosters, "nflverse_data")
  expect_s3_class(rosters_years, "nflverse_data")
  expect_gt(nrow(rosters), 2000)
  expect_gt(nrow(rosters_years), 90000)
})

test_that("load_ngs", {
  skip_on_cran()
  skip_if_offline("github.com")

  ngs_passing <- load_nextgen_stats()
  ngs_receiving <- load_nextgen_stats(
    seasons = 2019:2020,
    stat_type = "receiving",
    file_type = "parquet"
  )
  ngs_rushing <- load_nextgen_stats(seasons = 2020, stat_type = "rushing")

  expect_s3_class(ngs_passing, "nflverse_data")
  expect_s3_class(ngs_receiving, "nflverse_data")
  expect_s3_class(ngs_rushing, "nflverse_data")

  expect_gt(nrow(ngs_receiving), 1500)
  expect_gt(nrow(ngs_rushing), 500)
})

test_that("load_teams", {
  skip_on_cran()
  skip_if_offline("github.com")

  team_graphics <- load_teams()
  team_graphics_all <- load_teams(current = FALSE)

  expect_s3_class(team_graphics, "nflverse_data")
  expect_s3_class(team_graphics_all, "nflverse_data")

  expect_gte(nrow(team_graphics), 32)
  expect_gt(nrow(team_graphics_all), 32)
})

test_that("load_depth_charts", {
  skip_on_cran()
  skip_if_offline("github.com")

  depth_charts <- load_depth_charts()
  depth_charts_years <- load_depth_charts(seasons = 2019:2020)

  expect_s3_class(depth_charts, "nflverse_data")
  expect_s3_class(depth_charts_years, "nflverse_data")
  expect_gt(nrow(depth_charts_years), 60000)
})

test_that("load_injuries", {
  skip_on_cran()
  skip_if_offline("github.com")

  injuries <- load_injuries()
  injuries_years <- load_injuries(seasons = 2019:2020)
  # no injury data 2025+
  injuries_all <- load_injuries(seasons = 2009:2024)

  expect_s3_class(injuries, "nflverse_data")
  expect_s3_class(injuries_years, "nflverse_data")
  expect_s3_class(injuries_all, "nflverse_data")
  expect_gt(nrow(injuries_years), 8000)
  expect_gt(nrow(injuries_all), 60000)
})

test_that("load_espn_qbr", {
  skip_on_cran()
  skip_if_offline("github.com")

  qbr_default <- load_espn_qbr()
  qbr_weekly <- load_espn_qbr(seasons = TRUE, summary_type = "week")

  expect_s3_class(qbr_default, "nflverse_data")
  expect_s3_class(qbr_weekly, "nflverse_data")
  expect_gt(nrow(qbr_weekly), 7500)
})

test_that("load_trades", {
  skip_on_cran()
  skip_if_offline("github.com")

  trades <- load_trades()
  trades_2020 <- load_trades(seasons = 2020)

  expect_s3_class(trades, "nflverse_data")
  expect_s3_class(trades_2020, "nflverse_data")

  expect_gt(nrow(trades), 3000)
  expect_lte(nrow(trades_2020), 215)

  expect_error(load_trades("2020"))
})

test_that("load_draft_picks", {
  skip_on_cran()
  skip_if_offline("github.com")

  picks <- load_draft_picks()
  picks_2020 <- load_draft_picks(2020)

  expect_s3_class(picks, "nflverse_data")
  expect_s3_class(picks_2020, "nflverse_data")

  expect_gt(nrow(picks), 10000)
  expect_lt(nrow(picks_2020), 260)

  expect_error(load_draft_picks("2020"))
})

test_that("load_combine", {
  skip_on_cran()
  skip_if_offline("github.com")

  combine <- load_combine()
  combine_2020 <- load_combine(2020)

  expect_s3_class(combine, "nflverse_data")
  expect_s3_class(combine_2020, "nflverse_data")

  expect_gt(nrow(combine), 7000)
  expect_gte(nrow(combine_2020), 337)

  expect_error(load_combine("2020"))
})

test_that("load_pfr_passing", {
  skip_on_cran()
  skip_if_offline("github.com")

  expect_warning(passing <- load_pfr_passing())

  expect_s3_class(passing, "nflverse_data")
  expect_gt(nrow(passing), 200)
})

test_that("load_pfr_advstats", {
  skip_on_cran()
  skip_if_offline("github.com")

  expect_error(load_pfr_advstats("2020"))

  pass <- load_pfr_advstats(seasons = 2020, stat_type = "pass")
  rush <- load_pfr_advstats(seasons = TRUE, stat_type = "rush")
  def <- load_pfr_advstats(
    seasons = TRUE,
    stat_type = "def",
    summary_level = "season"
  )

  expect_s3_class(pass, "nflverse_data")
  expect_s3_class(rush, "nflverse_data")

  expect_gt(nrow(pass), 600)
  expect_gt(nrow(rush), 5000)
})

test_that("load_snap_counts", {
  skip_on_cran()
  skip_if_offline("github.com")

  expect_error(load_snap_counts("2020"))

  counts <- load_snap_counts(2020)

  expect_s3_class(counts, "nflverse_data")
  expect_gt(nrow(counts), 20000)
})

test_that("load_contracts", {
  skip_on_cran()
  skip_if_offline("github.com")

  contracts <- load_contracts()

  expect_s3_class(contracts, "nflverse_data")
  expect_gt(nrow(contracts), 20000)
})

test_that("load_players", {
  skip_on_cran()
  skip_if_offline("github.com")

  players <- load_players()

  expect_s3_class(players, "nflverse_data")
  expect_gt(nrow(players), 5000)
})

test_that("load_officials", {
  skip_on_cran()
  skip_if_offline("github.com")

  officials <- load_officials()

  expect_s3_class(officials, "nflverse_data")
  expect_gt(nrow(officials), 10000)
})

test_that("load_ftn_charting", {
  skip_on_cran()
  skip_if_offline("github.com")

  ftn_charting <- load_ftn_charting(2022)

  expect_s3_class(ftn_charting, "nflverse_data")
  expect_gt(nrow(ftn_charting), 40000)
})


## NEW LOAD FUNCTIONS GO ABOVE THIS LINE ##

test_that("nflverse_data print method works", {
  skip_on_cran()
  skip_if_offline("github.com")

  contracts <- load_contracts()

  expect_output(
    expect_message(
      print(contracts),
      regexp = "nflverse Historical Contract Data from OverTheCap.com|Data updated"
    )
  )
})

test_that("Cache clearing works", {
  skip_on_cran()
  skip_if_offline("github.com")

  csv_from_url(
    "https://github.com/nflverse/nflverse-data/releases/download/draft_picks/draft_picks.csv"
  )

  expect(
    memoise::has_cache(csv_from_url)(
      "https://github.com/nflverse/nflverse-data/releases/download/draft_picks/draft_picks.csv"
    ),
    "Function was not memoised!"
  )

  expect_message(.clear_cache(), "nflreadr function cache cleared!")

  expect(
    !memoise::has_cache(csv_from_url)(
      "https://github.com/nflverse/nflverse-data/releases/download/draft_picks/draft_picks.csv"
    ),
    "Cache was not cleared!"
  )
})

test_that("nflverse_game_id works", {
  # test single value input
  expect_equal(nflverse_game_id(2022, 2, "LAC", "KC"), "2022_02_LAC_KC")

  # test vector input
  expect_equal(
    nflverse_game_id(rep(2022, 2), c(2, 13), c("LAC", "LAC"), c("KC", "LV")),
    c("2022_02_LAC_KC", "2022_13_LAC_LV")
  )

  # allow old names (SD) and clean team abbreviations (LAR -> LA)
  expect_equal(
    nflverse_game_id(rep(2022, 2), c(2, 13), c("SD", "LAC"), c("KC", "LAR")),
    c("2022_02_SD_KC", "2022_13_LAC_LA")
  )

  # error on season outside range
  expect_error(
    nflverse_game_id(1998, 2, "LAC", "KC"),
    "must be between 1999"
  )

  # error on week outside range
  expect_error(
    nflverse_game_id(2022, 30, "LAC", "KC"),
    "must be between 1 and"
  )

  # error on away team
  expect_error(
    nflverse_game_id(2022, 2, rep("LACC", 2), "KC"),
    "invalid `away` abbreviation"
  )

  # error on home team
  expect_error(
    nflverse_game_id(2022, 2, "LAC", rep("KKC", 2)),
    "invalid `home` abbreviation"
  )

  # warn on non smart input
  expect_warning(
    nflverse_game_id(2022, 2, rep("LAC", 2), rep("KC", 3)),
    "Arguments should be of of length one or of length"
  )
})

#### Internal UTILS ###

.datatable.aware <- TRUE

is_installed <- function(pkg) requireNamespace(pkg, quietly = TRUE)

#' @keywords internal
"_PACKAGE"

#' @importFrom data.table `:=`
NULL

`%c%` <- function(x, y) {
  data.table::fifelse(!is.na(x), x, y)
}

#' rbindlist but maintain attributes of last file
#'
#' @export
#' @keywords internal
rbindlist_with_attrs <- function(dflist) {
  nflverse_timestamp <- attr(dflist[[length(dflist)]], "nflverse_timestamp")
  nflverse_type <- attr(dflist[[length(dflist)]], "nflverse_type")
  out <- data.table::rbindlist(dflist, use.names = TRUE, fill = TRUE)
  data.table::setattr(out, "nflverse_timestamp", nflverse_timestamp)
  data.table::setattr(out, "nflverse_type", nflverse_type)
  out
}

.ignore_unused_imports <- function() {
  curl::curl_fetch_memory
  methods::setOldClass
}

my_rhub_check <- function() {
  cli::cli_text("Please run the following code")
  cli::cli_text(
    "{.run rhub::rhub_check(platforms = nflreadr:::rhub_check_platforms())}"
  )
}

rhub_check_platforms <- function() {
  # plts created with
  # out <- paste0('"', rhub::rhub_platforms()$name, '"', collapse = ",\n")
  # cli::cli_code(paste0(
  #   "plts <- c(\n", out, "\n)"
  # ))

  plts <- c(
    "linux",
    "m1-san",
    "macos",
    "macos-arm64",
    "windows",
    "atlas",
    "c23",
    "clang-asan",
    "clang-ubsan",
    "clang16",
    "clang17",
    "clang18",
    "clang19",
    "clang20",
    "donttest",
    "gcc-asan",
    "gcc13",
    "gcc14",
    "gcc15",
    "intel",
    "mkl",
    "nold",
    "noremap",
    "nosuggests",
    "rchk",
    "ubuntu-clang",
    "ubuntu-gcc12",
    "ubuntu-next",
    "ubuntu-release",
    "valgrind"
  )
  exclude <- c("rchk", "nosuggests", "valgrind", "macos")
  plts[!plts %in% exclude]
}

nflverse_thanks <- function() {
  cli::cli_text("Run the following code and copy/paste its output to NEWS.md")

  cli::cli_code(
    '
    contributors <- usethis::use_tidy_thanks()
    paste(
      "Thank you to",
      glue::glue_collapse(
        paste0("&#x0040;", contributors), sep = ", ", last = ", and "
      ),
      "for their questions, feedback, and contributions towards this release."
    )'
  )
}

# additional release issue bullets for usethis::use_release_issue()
release_bullets <- function() {
  c(
    '`devtools::check_mac_release()`',
    '`nflreadr:::my_rhub_check()`',
    '`pkgdown::check_pkgdown()`',
    '`nflreadr:::nflverse_thanks()`',
    NULL
  )
}

#' Load Advanced Passing Stats from PFR
#'
#' @description Loads player level season stats provided by Pro Football Reference
#' starting with the 2019 season.
#'
#' @param seasons a numeric vector specifying what seasons to return, if `TRUE` returns all available data
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#'   load_pfr_passing()
#' }
#'
#' @return A tibble of season-level player statistics provided by Pro Football Reference.
#'
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_pfr_passing.html> for the web data dictionary
#' @seealso [`dictionary_pfr_passing`] for the data dictionary as bundled within the package
#' @seealso Issues with this data should be filed here: <https://github.com/nflverse/nflverse-pfr>
#'
#' @keywords internal
#' @export
load_pfr_passing <- function(seasons = TRUE) {
  cli::cli_warn(
    "`load_pfr_passing()` has been deprecated, please use `load_pfr_advstats(stat_type = 'pass', summary_level = 'season')`"
  )

  load_pfr_advstats(
    seasons = seasons,
    stat_type = "pass",
    summary_level = "season"
  )
}

#' Load Advanced Stats from PFR
#'
#' @description Loads player level season stats provided by Pro Football Reference
#' starting with the 2018 season, primarily to augment existing nflverse data.
#'
#' @param seasons a numeric vector specifying what seasons to return, if `TRUE` returns all available data
#' @param stat_type one of "pass", "rush", "rec", "def"
#' @param summary_level one of "week" (default) or "season" - some data is only available at the season level
#' @param file_type One of `c("rds", "csv", "parquet")`. Can also be set globally with
#' `options(nflreadr.prefer)`
#'
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({ # prevents cran errors
#'   load_pfr_advstats()
#' })
#' }
#'
#' @return A tibble of player statistics provided by Pro Football Reference that supplements data in nflverse
#'
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_pfr_passing.html> for the web data dictionary
#' @seealso <https://www.pro-football-reference.com/years/2021/passing_advanced.htm>
#' @seealso Issues with this data should be filed here: <https://github.com/nflverse/nflverse-data>
#'
#' @export
load_pfr_advstats <- function(
  seasons = most_recent_season(),
  stat_type = c("pass", "rush", "rec", "def"),
  summary_level = c("week", "season"),
  file_type = getOption("nflreadr.prefer", default = "rds")
) {
  if (isTRUE(seasons)) {
    seasons <- 2018:most_recent_season()
  }
  stopifnot(
    is.numeric(seasons),
    seasons >= 2018,
    seasons <= most_recent_season()
  )

  stat_type <- rlang::arg_match0(stat_type, c("pass", "rush", "rec", "def"))
  summary_level <- rlang::arg_match0(summary_level, c("week", "season"))
  file_type <- rlang::arg_match0(file_type, c("rds", "csv", "parquet"))

  switch(
    summary_level,
    "week" = .pfr_advstats_week(seasons, stat_type, file_type),
    "season" = .pfr_advstats_season(seasons, stat_type, file_type)
  )
}

.pfr_advstats_week <- function(seasons, stat_type, file_type) {
  urls <- paste0(
    "https://github.com/nflverse/nflverse-data/releases/download/pfr_advstats/advstats_week_",
    stat_type,
    "_",
    seasons,
    ".",
    file_type
  )

  out <- load_from_url(urls, seasons = seasons, nflverse = TRUE)
  return(out)
}

.pfr_advstats_season <- function(seasons, stat_type, file_type) {
  data_url <- paste0(
    "https://github.com/nflverse/nflverse-data/releases/download/pfr_advstats/advstats_season_",
    stat_type,
    ".",
    file_type
  )

  out <- load_from_url(data_url, seasons = seasons, nflverse = TRUE)
  return(out)
}

#' nflverse data class
#'
#' This class has a special S3 print method that tries to read attached metadata and
#' provide timestamps and source attributes. It otherwise will dispatch to the
#' `data.frame` class.
#'
#' @name nflverse_data-class
#' @keywords internal
#' @exportClass nflverse_data
methods::setOldClass(c("nflverse_data", "data.frame"))

#' @export
#' @noRd
print.nflverse_data <- function(x, ...) {
  cli::cli_rule("{.emph nflverse {attr(x,'nflverse_type')}}")

  if (!is.null(attr(x, 'nflverse_timestamp'))) {
    # Print a formatted string in local timezone
    timestamp <- format(
      # the attribute is character so we have to convert it to POSIX to be able
      # to convert the timezone to the local one
      as.POSIXct(
        attr(x, 'nflverse_timestamp'),
        tz = nflreadr::nflverse_data_timezone
      ),
      tz = Sys.timezone(),
      usetz = TRUE
    )
    cli::cli_alert_info("Data updated: {.field {timestamp}}")
  }

  NextMethod(print, x)
  invisible(x)
}

#' @export
#' @aliases make_nflverse_data
#' @rdname nflverse_data-class
as.nflverse_data <- function(df, nflverse_type = NULL, ...) {
  df <- data.table::as.data.table(df)
  data.table::setattr(
    df,
    "class",
    c("nflverse_data", "tbl_df", "tbl", "data.table", "data.frame")
  )

  dots <- rlang::dots_list(..., .named = TRUE)

  for (i in seq_along(dots)) {
    data.table::setattr(df, names(dots)[[i]], dots[[i]])
  }

  if (!is.null(nflverse_type)) {
    data.table::setattr(df, "nflverse_type", nflverse_type)
  }

  if (
    (length(dots) > 0 | !is.null(nflverse_type)) &&
      is.null(attr(df, "nflverse_timestamp"))
  ) {
    data.table::setattr(
      df,
      "nflverse_timestamp",
      format(Sys.time(), tz = nflreadr::nflverse_data_timezone, usetz = TRUE)
    )
  }

  return(df)
}

make_nflverse_data <- as.nflverse_data

#' Standardize NFL Team Abbreviations
#'
#' This function standardizes NFL team names, abbreviations, and nicknames to
#' nflverse defaults.
#' This helps for joins and plotting, especially with the nflplotR package!
#'
#' @param abbr a character vector of abbreviations, full team names, or team nicknames
#' @param current_location If `TRUE` (the default), the abbreviation of the most
#'   recent team location will be used.
#' @param keep_non_matches If `TRUE` (the default) an element of `abbr` that can't
#'   be matched to any of the internal mapping vectors will be kept as is. Otherwise
#'   it will be replaced with `NA`.
#'
#' @return A character vector with the length of `abbr` and cleaned team abbreviations
#'   if they are included in [`team_abbr_mapping`] or [`team_abbr_mapping_norelocate`]
#'   (depending on the value of `current_location`). Non matches may be replaced
#'   with `NA` (depending on the value of `keep_non_matches`).
#' @export
#' @examples
#' \dontshow{.for_cran()}
#' x <- c("PIE", "LAR", "PIT", "CRD", "OAK", "SL", "Niners", "San Diego Chargers")
#' # use current location and keep non matches
#' clean_team_abbrs(x)
#'
#' # keep old location and replace non matches
#' clean_team_abbrs(x, current_location = FALSE, keep_non_matches = FALSE)
clean_team_abbrs <- function(
  abbr,
  current_location = TRUE,
  keep_non_matches = TRUE
) {
  stopifnot(
    is.character(abbr),
    is.logical(current_location)
  )

  m <- if (isTRUE(current_location)) {
    nflreadr::team_abbr_mapping
  } else {
    nflreadr::team_abbr_mapping_norelocate
  }

  a <- unname(m[toupper(abbr)])

  if (any(is.na(a)) && getOption("nflreadr.verbose", default = interactive())) {
    cli::cli_warn(
      "Abbreviations not found in `nflreadr::team_abbr_mapping`: {paste(utils::head(abbr[is.na(a)], 10), collapse = ', ')}"
    )
  }

  if (isTRUE(keep_non_matches)) {
    a <- a %c% abbr
  }

  a
}

#' Create Player Merge Names
#'
#' Applies some name-cleaning heuristics to facilitate joins. These heuristics may include:
#'   - removing periods and apostrophes
#'   - removing common suffixes, such as Jr, Sr, II, III, IV
#'   - converting to lowercase
#'   - using `ffscrapr::dp_name_mapping` to do common name substitutions, such as Mitch Trubisky to Mitchell Trubisky
#'
#' Equivalent to the operation done by `ffscrapr::dp_clean_names()` and uses the same player name database.
#'
#' @param player_name a character vector of player names
#' @param lowercase defaults to FALSE - if TRUE, converts to lowercase
#' @param convert_lastfirst defaults to TRUE - converts names from "Last, First" to "First Last"
#' @param use_name_database uses internal name database to do common substitutions (Mitchell Trubisky to Mitch Trubisky etc)
#' @param convert_to_ascii If `TRUE`, will transliterate to latin-ascii via the
#' stringi package. Defaults to `TRUE` if the stringi package is installed.
#'
#' @return a character vector of cleaned names
#'
#' @examples
#' \dontshow{.for_cran()}
#' clean_player_names(c("A.J. Green", "Odell Beckham Jr.  ", "Le'Veon Bell Sr."))
#'
#' clean_player_names(c("Trubisky,      Mitch", "Atwell, Chatarius", "Elliott, Zeke", "Elijah Moore"),
#'                   convert_lastfirst = TRUE)
#'
#' @export
clean_player_names <- function(
  player_name,
  lowercase = FALSE,
  convert_lastfirst = TRUE,
  use_name_database = TRUE,
  convert_to_ascii = rlang::is_installed("stringi")
) {
  stopifnot(
    is.character(player_name),
    is.logical(c(lowercase, convert_lastfirst, use_name_database)),
    length(c(lowercase, convert_lastfirst, use_name_database)) == 3,
    !any(is.na(c(lowercase, convert_lastfirst, use_name_database)))
  )

  n <- player_name

  # squish internal whitespace
  n <- gsub(pattern = "\\s+", replacement = " ", x = n)

  # trim whitespace
  n <- gsub(pattern = "^\\s|\\s$", replacement = "", x = n)

  # convert Last, First
  if (isTRUE(convert_lastfirst)) {
    n <- gsub(pattern = "^(.+), (.+)$", replacement = "\\2 \\1", x = n)
  }

  # suffix removal
  n <- gsub(
    pattern = " Jr\\.$| Sr\\.$| III$| II$| IV$| V$|'|\\.|,",
    replacement = "",
    x = n,
    ignore.case = TRUE
  )

  # transliterate to latin-ascii if stringi is available
  if (isTRUE(convert_to_ascii)) {
    rlang::check_installed("stringi", "to convert text to latin-ascii")
    n <- stringi::stri_trans_general(n, "latin-ascii")
  }

  if (isTRUE(use_name_database)) {
    n <- unname(nflreadr::player_name_mapping[n] %c% n)
  }

  if (isTRUE(lowercase)) {
    n <- tolower(n)
  }

  n
}

#' Clean Home/Away in dataframes into Team/Opponent dataframes
#'
#' This function converts dataframes with "home_" and "away_" prefixed columns to "team_" and "opponent_", and doubles the rows. This makes sure that there's one row for each team (as opposed to one row for each game).
#'
#' @param dataframe dataframe
#' @param invert a character vector of columns that gets inverted when referring to the away team (e.g. home spread = 1 gets converted to away_spread = -1)
#'
#' @examples
#' \dontshow{.for_cran()}
#' # a small example dataframe
#' s <- data.frame(
#'    game_id = c("2020_20_TB_GB", "2020_20_BUF_KC", "2020_21_KC_TB"),
#'    game_type = c("CON", "CON", "SB"),
#'    away_team = c("TB", "BUF", "KC"),
#'    away_score = c(31L, 24L, 9L),
#'    home_team = c("GB", "KC", "TB"),
#'    home_score = c(26L, 38L, 31L),
#'    location = c("Home", "Home", "Neutral"),
#'    result = c(-5L, 14L, 22L),
#'    spread_line = c(3, 3, -3)
#'  )
#'
#' clean_homeaway(s, invert = c("result","spread_line"))
#' @return a dataframe with one row per team (twice as long as the input dataframe)
#' @export
clean_homeaway <- function(dataframe, invert = NULL) {
  stopifnot(
    is.data.frame(dataframe),
    is.null(invert) || is.character(invert)
  )
  dataframe$.row_order <- seq_len(nrow(dataframe))
  home <- dataframe
  away <- dataframe

  names(home) <- gsub(
    x = names(home),
    pattern = "^home_",
    replacement = "team_"
  )
  names(home) <- gsub(
    x = names(home),
    pattern = "^away_",
    replacement = "opponent_"
  )
  names(home) <- gsub(x = names(home), pattern = "_home$", replacement = "")
  names(home) <- gsub(
    x = names(home),
    pattern = "_away$",
    replacement = "_opponent"
  )
  names(home) <- gsub(
    x = names(home),
    pattern = "team_team",
    replacement = "team"
  )
  names(home) <- gsub(
    x = names(home),
    pattern = "opponent_team",
    replacement = "opponent"
  )

  # The location variable depends on if there is location in input. If yes, preserve
  # "neutral" otherwise just set to home/away
  home$location <- if ("location" %in% names(home)) {
    data.table::fifelse(home$location == "Neutral", "neutral", "home", "home")
  } else {
    "home"
  }

  names(away) <- gsub(
    x = names(away),
    pattern = "^away_",
    replacement = "team_"
  )
  names(away) <- gsub(
    x = names(away),
    pattern = "^home_",
    replacement = "opponent_"
  )
  names(away) <- gsub(x = names(away), pattern = "_away$", replacement = "")
  names(away) <- gsub(
    x = names(away),
    pattern = "_home$",
    replacement = "_opponent"
  )
  names(away) <- gsub(
    x = names(away),
    pattern = "team_team",
    replacement = "team"
  )
  names(away) <- gsub(
    x = names(away),
    pattern = "opponent_team",
    replacement = "opponent"
  )

  # The location variable depends on if there is location in input. If yes, preserve
  # "neutral" otherwise just set to home/away
  away$location <- if ("location" %in% names(away)) {
    data.table::fifelse(away$location == "Neutral", "neutral", "away", "away")
  } else {
    "away"
  }

  if (!is.null(invert)) {
    data.table::setDF(away)
    away[, c(invert)] <- away[, c(invert)] * -1
  }

  .row_order <- NULL
  out <- data.table::rbindlist(list(home, away), use.names = TRUE)[
    order(.row_order),
    -c(".row_order")
  ]

  # we want to preserve the input class(es)
  class(out) <- class(dataframe)

  # input attributes shall be presrved mostly so we gotta do some acrobatic
  # and combine input attributes with new attributes
  df_attrs <- attributes(dataframe)
  df_attrs <- df_attrs[setdiff(
    names(df_attrs),
    c("row.names", "names", ".internal.selfref")
  )]
  out_attrs <- attributes(out)[c("names", "row.names", ".internal.selfref")]
  attributes(out) <- c(out_attrs, df_attrs)

  # the input may or may not have the nflverse_type attribute
  # if it is available, we'll add "by team" to make clear it's a different
  # nflverse_type
  if ("nflverse_type" %in% names(attributes(out))) {
    attr(out, "nflverse_type") <- paste(attr(out, "nflverse_type"), "by team")
  }
  out
}

test_that("cleaning team abbreviations", {
  o <- options(nflreadr.verbose = TRUE)

  expected_warning <- "Abbreviations not found in `nflreadr::team_abbr_mapping`: PIE"

  x <- c(
    "PIE",
    "LAR",
    "PIT",
    "CRD",
    "OAK",
    "CLV",
    "PHO",
    "Niners",
    "49ers",
    "San Diego Chargers",
    "Oakland Raiders",
    "St Louis Rams"
  )
  expect_warning(new_abbr <- clean_team_abbrs(x), regexp = expected_warning)
  expect_warning(
    old_abbr <- clean_team_abbrs(x, current_location = FALSE),
    regexp = expected_warning
  )
  expect_warning(
    new_abbr_drop <- clean_team_abbrs(x, keep_non_matches = FALSE),
    regexp = expected_warning
  )

  expect_equal(
    new_abbr,
    c(
      "PIE",
      "LA",
      "PIT",
      "ARI",
      "LV",
      "CLE",
      "ARI",
      "SF",
      "SF",
      "LAC",
      "LV",
      "LA"
    )
  )
  expect_equal(
    new_abbr_drop,
    c(NA, "LA", "PIT", "ARI", "LV", "CLE", "ARI", "SF", "SF", "LAC", "LV", "LA")
  )
  expect_equal(
    old_abbr,
    c(
      "PIE",
      "LA",
      "PIT",
      "ARI",
      "OAK",
      "CLE",
      "ARI",
      "SF",
      "SF",
      "SD",
      "OAK",
      "STL"
    )
  )
  options(o)
})

test_that("cleaning player names", {
  p <- c(
    "Trubisky,      Mitch",
    "Atwell, Chatarius",
    "Elliott, Zeke",
    "Elijah Moore",
    "A.J. Green",
    "Odell Beckham Jr.   ",
    "Audric Estimé"
  )

  lower <- clean_player_names(
    p,
    lowercase = TRUE,
    use_name_database = TRUE,
    convert_lastfirst = TRUE
  )

  expect_equal(
    lower,
    c(
      "mitchell trubisky",
      "tutu atwell",
      "ezekiel elliott",
      "elijah moore",
      "aj green",
      "odell beckham",
      "audric estime"
    )
  )

  otc <- c("Willie Gay, Jr.", "Jerry Rice, Jr.")

  otc_clean <- clean_player_names(otc, convert_lastfirst = FALSE)

  expect_equal(
    otc_clean,
    c("Willie Gay", "Jerry Rice")
  )
})

test_that("all caps names with suffix are cleaned correctly", {
  p <- c("LENO JR., CHARLES")

  lower <- clean_player_names(
    p,
    lowercase = TRUE,
    use_name_database = TRUE,
    convert_lastfirst = TRUE
  )
  allcaps <- clean_player_names(
    p,
    lowercase = FALSE,
    use_name_database = TRUE,
    convert_lastfirst = TRUE
  )

  expect_equal(lower, c("charles leno"))
  expect_equal(allcaps, c("CHARLES LENO"))
})

test_that("cleaning home and away columns", {
  s <- data.frame(
    game_id = c("2020_20_TB_GB", "2020_20_BUF_KC", "2020_21_KC_TB"),
    game_type = c("CON", "CON", "SB"),
    home_team = c("GB", "KC", "TB"),
    home_score = c(26L, 38L, 31L),
    team_id_home = c("001", "002", "003"),
    away_team = c("TB", "BUF", "KC"),
    away_score = c(31L, 24L, 9L),
    team_id_away = c("003", "004", "002"),
    location = c("Home", "Home", "Neutral"),
    result = c(-5L, 14L, 22L),
    spread_line = c(3, 3, -3)
  )
  c <- clean_homeaway(s, invert = c("result", "spread_line"))

  expect_equal(nrow(c), nrow(s) * 2)
  expect(
    all(!grepl(x = names(c), pattern = "^home_")),
    "Error: `home_` was found in `names(c)`"
  )
  expect(
    all(!grepl(x = names(c), pattern = "_home$")),
    "Error: `_home` was found in `names(c)`"
  )
  expect_true("neutral" %in% c$location)
  expect_identical(class(c), class(s))
})

#' Coalescing join
#'
#' EXPERIMENTAL! This function joins two dataframes together by key, and then coalesces any
#' columns that have shared names (i.e. fills in NAs). A utility function primarily
#' used internally within nflverse to help build player IDs
#'
#' @param x,y dataframes. Will be coerced to data.table
#' @param by keys to join on, as a plain or named character vector
#' @param type one of "left" (all rows of x and matching rows of y),
#' "inner" (matching rows of x and y), "full" (all rows of x and y)
#' @param by.x,by.y alternate form of keys to join on - if provided, will override `by`.
#' @param incomparables keys to NOT match on, i.e. NA should not match on NA.
#' @param sort whether to sort output by the join keys
#' @param ... other args passed to `merge.data.frame()`
#'
#' @return a data.frame joining x and y dataframes together, with every column from
#' both x and y and patching NA values in x with those in y.
#'
#'
#' @examples
#' \dontshow{.for_cran()}
#' x <- data.frame(id1 = c(NA_character_,letters[1:4]), a = c(1,NA,3,NA,5), b = 1:5 * 10)
#' y <- data.frame(id2 = c(letters[3:11],NA_character_), a = -(1:10), c = 1:10 * 100)
#'
#' join_coalesce(x,y, by = c("id1"="id2"))
#' join_coalesce(x,y, by.x = "id1", by.y = "id2")
#' join_coalesce(x,y, by = c("id1"="id2"), type = "inner")
#' join_coalesce(x,y, by = c("id1"="id2"), type = "full")
#'
#' @export
join_coalesce <- function(
  x,
  y,
  by = NULL,
  type = c("left", "inner", "full"),
  ...,
  by.x = NULL,
  by.y = NULL,
  sort = TRUE,
  incomparables = c(NA, NaN)
) {
  type <- rlang::arg_match0(type, c("left", "inner", "full"))

  stopifnot(
    is.data.frame(x),
    is.data.frame(y),
    length(by.x) == length(by.y),
    (is.character(by) && length(by) >= 1) ||
      (is.character(by.x) && is.character(by.y) && length(by.x) >= 1)
  )
  x <- as.data.frame(x)
  y <- as.data.frame(y)

  keys_x <- if (!is.null(by.x)) {
    by.x
  } else if (is.null(names(by))) {
    by
  } else {
    ifelse(names(by) == "", by, names(by))
  }
  keys_y <- if (!is.null(by.y)) by.y else by

  check_keys <- c(
    "Join `by` keys in x are not unique" = nrow(x) != nrow(unique(x[keys_x])),
    "Join `by` keys in y are not unique" = nrow(y) != nrow(unique(y[keys_y])),
    "Join `by` keys in x have NAs" = any(is.na(x[keys_x])),
    "Join `by` keys in y have NAs" = any(is.na(y[keys_y]))
  )

  if (any(check_keys)) {
    cli::cli_warn(
      names(check_keys)[which(check_keys)]
    )
  }

  joined_cols <- c(setdiff(names(x), keys_x), setdiff(names(y), keys_y))
  dupl_cols <- joined_cols[duplicated(joined_cols)]

  # data.table's merge doesn't have an "incomparables" argument in the current prod version
  # this causes NA to match to NA
  merged_df <- merge.data.frame(
    x = x,
    y = y,
    by.x = keys_x,
    by.y = keys_y,
    all.x = type %in% c("left", "full"),
    all.y = type == "full",
    sort = sort,
    incomparables = incomparables,
    ...,
    suffixes = c("..x", "..y")
  )

  data.table::setDT(merged_df)

  for (col in dupl_cols) {
    data.table::set(
      merged_df,
      j = col,
      value = merged_df[[paste0(col, "..x")]] %c%
        merged_df[[paste0(col, "..y")]]
    )
    data.table::set(merged_df, j = paste0(col, "..x"), value = NULL)
    data.table::set(merged_df, j = paste0(col, "..y"), value = NULL)
  }

  out <- merged_df[, c(keys_x, unique(joined_cols)), with = FALSE]
  data.table::setDF(out)
  return(out)
}

test_that("join coalesce function works", {
  # two dataframes, some with missing variables (e.g. in `a`)
  x <- data.frame(
    id1 = c(NA_character_, letters[1:4]),
    a = c(1, NA, 3, NA, 5),
    b = 1:5 * 10
  )
  y <- data.frame(
    id2 = c(letters[3:11], NA_character_),
    a = -(1:10),
    c = 1:10 * 100
  )

  # Keys include NAs and trigger a warning
  expect_warning(
    left1 <- join_coalesce(x, y, by = c("id1" = "id2"))
  )
  expect_warning(
    left2 <- join_coalesce(x, y, by.x = "id1", by.y = "id2")
  )
  expect_warning(
    inner1 <- join_coalesce(x, y, by = c("id1" = "id2"), type = "inner")
  )
  expect_warning(
    full1 <- join_coalesce(x, y, by = c("id1" = "id2"), type = "full")
  )

  expect_equal(left1, left2) # both join-by methods work the same
  expect_equal(sum(is.na(left1$a)), 1) # one NA on left join
  expect_equal(sum(is.na(full1$a)), 1) # one NA on full join
  expect_equal(sum(is.na(inner1$a)), 0) # no NAs on inner join
  expect_equal(nrow(left1), nrow(x)) # left join creates same length as x
  expect_equal(nrow(left2), nrow(x))
  expect_equal(nrow(inner1), 2) # two rows in common
  expect_equal(nrow(full1), 13) # 13 rows on full join
})

# join_coalesce converted dates to integers because of the usage
# of ifelse in %c%. https://github.com/nflverse/nflreadr/issues/213
test_that("%c% keeps dates as dates", {
  d1 <- as.Date(c("1998-05-01", "1997-12-12", NA_character_))
  d2 <- as.Date(c("1995-01-31", NA_character_, "1997-12-12"))
  expect_identical(
    d1 %c% d2,
    as.Date(c("1998-05-01", "1997-12-12", "1997-12-12"))
  )
})

#' Load Participation Data
#'
#' @description Loads participation data from the [nflverse-data repository](https://github.com/nflverse/nflverse-data)
#'
#' Participation data prior to 2023 is from NFL NGS. Participation data from
#' 2023 onwards is courtesy of FTN and is provided after all post-season games are
#' completed. This data is released under the [CC-BY-SA 4.0](https://creativecommons.org/licenses/by-sa/4.0/)
#' Creative Commons license and attribution must be made to **FTN Data via nflverse** (from 2023 onwards)
#' or **NFL NextGenStats via nflverse** (for 2022 and earlier)
#'
#' @param seasons A numeric vector of 4-digit years associated with given NFL
#' seasons - defaults to latest season. If set to `TRUE`, returns all available
#' data since 2016.
#' @param include_pbp a logical: download and join pbp to this data?
#' @param file_type One of `c("rds", "csv", "parquet")`. Can also be
#' set globally with `options(nflreadr.prefer)`
#'
#' @return A dataframe of participation data, optionally merged with play by play
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({ # prevents cran errors
#'   load_participation(seasons = 2020, include_pbp = TRUE)
#' })
#' }
#'
#' @export
load_participation <- function(
  seasons = most_recent_season(),
  include_pbp = FALSE,
  file_type = getOption("nflreadr.prefer", default = "rds")
) {
  file_type <- rlang::arg_match0(file_type, c("rds", "csv", "parquet"))

  if (isTRUE(seasons)) {
    seasons <- 2016:most_recent_season()
  }

  stopifnot(
    is.numeric(seasons),
    seasons >= 2016,
    seasons <= most_recent_season(),
    is.logical(include_pbp),
    length(include_pbp) == 1
  )

  urls <- paste0(
    "https://github.com/nflverse/nflverse-data/releases/download/pbp_participation/pbp_participation_",
    seasons,
    ".",
    file_type
  )

  participation <- load_from_url(urls, nflverse = TRUE, seasons = seasons)

  if (!include_pbp) {
    return(participation)
  }

  pbp <- load_pbp(seasons = seasons, file_type = file_type)

  data.table::setDT(participation)
  data.table::setkeyv(participation, c("nflverse_game_id", "play_id"))
  data.table::setDT(pbp)
  data.table::setkeyv(pbp, c("game_id", "play_id"))

  pbp_participation <- data.table::merge.data.table(
    participation[, -c("old_game_id")],
    pbp,
    by.x = c("nflverse_game_id", "play_id"),
    by.y = c("game_id", "play_id")
  )

  pbp_participation <- make_nflverse_data(
    pbp_participation,
    nflverse_type = "play-by-play participation",
    nflverse_timestamp = attr(participation, "nflverse_timestamp")
  )

  return(pbp_participation)
}

#' Load FTN Charting Data
#'
#' FTN Data manually charts plays and has graciously provided a subset of their
#' charting data to be published via the nflverse. Data is available from the 2022
#' season onwards and is charted within 48 hours following each game. This data
#' is released under the [CC-BY-SA 4.0](https://creativecommons.org/licenses/by-sa/4.0/)
#' Creative Commons license and attribution must be made to **FTN Data via nflverse**
#'
#' @param seasons a numeric vector of seasons to return, defaults to most recent season.
#' If set to `TRUE`, returns all available data. Data available from 2022 onwards.
#' @param file_type One of `c("rds", "csv", "parquet")`. Can also be set globally with
#' `options(nflreadr.prefer)`
#'
#' @author FTN Data
#' @source FTNData.com
#'
#' @examples
#' \donttest{
#'   try({ # prevents cran errors
#'   load_ftn_charting()
#'   })
#' }
#'
#' @return Play-level manual charting data from FTN Data
#'
#' @seealso <https://ftnfantasy.com/stats/sports-data>
#' @seealso `vignette("Data Dictionary - FTN Charting")`
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_ftn_charting.html> for the web data dictionary
#' @family ftn_charting
#'
#' @export
load_ftn_charting <- function(
  seasons = most_recent_season(),
  file_type = getOption("nflreadr.prefer", default = "rds")
) {
  if (isTRUE(seasons)) {
    seasons <- 2022:most_recent_season()
  }

  stopifnot(
    is.numeric(seasons),
    seasons >= 2022,
    seasons <= most_recent_season()
  )

  file_type <- rlang::arg_match0(file_type, c("rds", "csv", "parquet"))

  urls <- glue::glue(
    "https://github.com/nflverse/nflverse-data/releases/download/ftn_charting/ftn_charting_{seasons}.{file_type}"
  )

  out <- load_from_url(
    urls,
    nflverse = TRUE,
    nflverse_type = glue::glue(
      "FTN Charting Data - please attribute to 'ftndata.com via nflverse'"
    )
  )
  return(out)
}

#' Data Dictionary: FTN Charting Data
#'
#' A dataframe containing the data dictionary for [`load_ftn_charting()`]
#'
#' @family ftn_charting
#' @seealso `vignette("Data Dictionary - FTN Charting")`
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_ftn_charting.html>
"dictionary_ftn_charting"

#' Load Players
#'
#' @description Load a dataframe of player-level information, including IDs and other mostly-immutable data (birthdates, college, draft position etc.)
#' @param file_type One of `c("rds", "csv", "parquet")`. Can also be set globally with
#' `options(nflreadr.prefer)`
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({ # prevents cran errors
#'   load_players()
#' })
#' }
#'
#' @return A tibble with one row per player.
#'
#' @details
#'
#' ## Contribute
#'
#' Please visit <https://github.com/nflverse/nflverse-players/blob/master/.github/CONTRIBUTING.md>
#' for a detailed explanation how to contribute (or correct) players data.
#'
#' ## Scope of Players Dataset
#'
#' The nflverse players dataset aims to be the single source of truth when it
#' comes to NFL player IDs across various sources (relevant to the nflverse).
#' The following source IDs and information are considered in-scope of the
#' nflverse players dataset:
#'
#' - Basic player information e.g. name, height, weight, age, date of birth,
#' years experience, college, status, position, headshot image (mostly from GSIS)
#' - Draft information: draft year, draft round, draft pick, draft team (from PFR)
#' - PFF position and status information
#' - Primary source IDs from:
#'   - NFL (`gsis_id`, `smart_id`) - `gsis_id` is the primary key
#'   - Pro Football Reference (`pfr_id`)
#'   - Pro Football Focus (`pff_id`)
#'   - Over The Cap (`otc_id`)
#'   - Elias Sports Bureau (`esb_id`)
#'   - ESPN (`espn_id`)
#'
#' @seealso Issues with this data should be filed here: <https://github.com/nflverse/nflverse-players>.
#' @seealso [`dictionary_players`] for the data dictionary as bundled within the package
#'
#' @export
load_players <- function(
  file_type = getOption("nflreadr.prefer", default = "rds")
) {
  file_type <- rlang::arg_match0(file_type, c("rds", "csv", "parquet"))
  url <- glue::glue(
    "https://github.com/nflverse/nflverse-data/releases/download/players/players.{file_type}"
  )
  out <- load_from_url(url, nflverse = TRUE)
  out
}

#' Clear function cache
#'
#' This function clears the memoised cache of all functions memoised by `nflreadr`.
#'
#' @rdname clear_cache
#' @export
#' @return A success message after clearing the cache.
#' @examples
#' clear_cache()
clear_cache <- function() {
  functions <- list(
    raw_from_url,
    csv_from_url,
    rds_from_url,
    # this is hard deprecated but there could still be some stuff in cache
    # we'll drop it here
    qs_from_url
  )

  lapply(functions, memoise::forget)

  cli::cli_alert_success("nflreadr function cache cleared!")

  invisible(TRUE)
}

#' @rdname clear_cache
#' @export
.clear_cache <- clear_cache

#' Load Historical Player Contracts from OverTheCap.com
#'
#' @description Loads player contracts from OverTheCap.com
#'
#' @param file_type One of `"rds"`, `"csv"`, or `"parquet"`.
#'   Can also be set globally with options(nflreadr.prefer)
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({ # prevents cran errors
#'   load_contracts()
#' })
#' }
#'
#' @return A tibble of active and non-active NFL player contracts.
#'
#' @seealso <https://overthecap.com/contract-history> for a web version of the data
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_contracts.html> for a web version of the dictionary
#' @seealso [`dictionary_contracts`] for the data dictionary as bundled within the package
#' @seealso Issues with this data should be filed here: <https://github.com/nflverse/rotc>
#'
#' @export
load_contracts <- function(
  file_type = getOption("nflreadr.prefer", default = "rds")
) {
  file_type <- rlang::arg_match0(file_type, c("rds", "parquet", "csv"))
  url <- glue::glue(
    "https://github.com/nflverse/nflverse-data/releases/download/contracts/historical_contracts.{file_type}"
  )
  out <- load_from_url(url, nflverse = TRUE)
  return(out)
}

#' Load any rds/csv/csv.gz/parquet file from a remote URL
#'
#' @param url a vector of URLs to load into memory. If more than one URL provided, will row-bind them.
#' @param seasons a numeric vector of years that will be used to filter the dataframe's `season` column. If `TRUE` (default), does not filter.
#' @param nflverse TRUE to add nflverse_data classing and attributes.
#' @param ... named arguments that will be added as attributes to the data, e.g. `nflverse_type` = "pbp"
#'
#' @export
#'
#' @return a dataframe, possibly of type `nflverse_data`
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({ # prevents cran errors
#'   urls <- c("https://github.com/nflverse/nflverse-data/releases/download/rosters/roster_2020.csv",
#'             "https://github.com/nflverse/nflverse-data/releases/download/rosters/roster_2021.csv")
#'  load_from_url(urls, nflverse = TRUE, nflverse_type = "rosters for 2020 & 2021")
#' })
#' }
load_from_url <- function(url, ..., seasons = TRUE, nflverse = FALSE) {
  url <- as.character(url)

  if (length(url) == 1) {
    out <- loader(url)
    if (!isTRUE(seasons)) {
      stopifnot(is.numeric(seasons))
    }
    if (!isTRUE(seasons) && "season" %in% names(out)) {
      out <- out[out$season %in% seasons]
    }
  }

  if (length(url) > 1) {
    p <- NULL
    if (is_installed("progressr")) {
      p <- progressr::progressor(along = url)
    }
    out <- lapply(url, progressively(loader, p))
    out <- rbindlist_with_attrs(out)
  }

  if (nflverse) {
    out <- make_nflverse_data(out, ...)
  }
  return(out)
}


#' Load .rds file from a remote connection
#'
#' @param url a character url
#'
#' @export
#'
#' @return a dataframe as created by [`readRDS()`]
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({ # prevents cran errors
#'   rds_from_url("https://github.com/nflverse/nflverse-data/releases/download/test/combines.rds")
#' })
#' }
rds_from_url <- function(url) {
  cache_message()
  con <- url(url)
  on.exit(close(con))
  load <- try(readRDS(con), silent = TRUE)

  if (inherits(load, "try-error")) {
    cli::cli_warn("Failed to readRDS from {.url {url}}")
    return(data.table::data.table())
  }

  data.table::setDT(load)
  load
}

#' Load .csv / .csv.gz file from a remote connection
#'
#' This is a thin wrapper on data.table::fread, but memoised & cached for twenty four hours.
#'
#' @param ... passed to data.table::fread
#' @inheritDotParams data.table::fread
#'
#' @export
#'
#' @return a dataframe as created by [`data.table::fread()`]
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({ # prevents cran errors
#'   csv_from_url("https://github.com/nflverse/nflverse-data/releases/download/test/combines.csv")
#' })
#' }
csv_from_url <- function(...) {
  cache_message()
  data.table::fread(...)
}

#' Load raw filedata from a remote connection
#'
#' This function allows you to retrieve data from a URL into raw format, which
#' can then be passed into the appropriate file-reading function. Data is memoised/cached for 24 hours.
#'
#' @param url a character url
#'
#' @export
#'
#' @return a raw vector
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({ # prevents CRAN errors
#' head(raw_from_url(
#'   "https://github.com/nflverse/nflverse-data/releases/download/test/combines.rds"
#'   ),
#' 50)
#' })
#' }
raw_from_url <- function(url) {
  cache_message()
  load <- try(curl::curl_fetch_memory(url), silent = TRUE)

  if (inherits(load, "try-error")) {
    cli::cli_warn("Failed to retrieve data from {.url {url}}")
    return(invisible(load))
  }

  if (load$status_code != 200) {
    cli::cli_warn(
      "HTTP error {.emph {load$status_code}} while retrieving data from {.url {url}}\nReturning request payload."
    )
    return(invisible(load))
  }

  load$content
}

#' Load .parquet file from a remote connection
#'
#' Retrieves a parquet file from URL. This function is cached
#'
#' @param url a character url
#'
#' @export
#'
#' @return a dataframe as parsed by [`arrow::read_parquet()`]
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({
#'   parquet_from_url(
#'     "https://github.com/nflverse/nflverse-data/releases/download/player_stats/player_stats.parquet"
#'   )
#' })
#' }
parquet_from_url <- function(url) {
  rlang::check_installed("arrow")
  cache_message()
  load <- try(curl::curl_fetch_memory(url), silent = TRUE)

  if (inherits(load, "try-error")) {
    cli::cli_warn("Failed to retrieve data from {.url {url}}")
    return(data.table::data.table())
  }

  content <- try(arrow::read_parquet(load$content), silent = TRUE)

  if (inherits(content, "try-error")) {
    cli::cli_warn(
      "Failed to parse file with {.fun arrow::read_parquet()} from {.url {url}}"
    )
    return(data.table::data.table())
  }

  data.table::setDT(content)
  return(content)
}

#' Load .qs file from a remote connection
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' The underlying package qs has been removed from CRAN on 2026-01-17 so
#' nflreadr stopped the qs support
#'
#' @param url a character url
#'
#' @export
#' @keywords internal
#' @return a dataframe
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({
#'   qs_from_url(
#'     "https://github.com/nflverse/nflverse-data/releases/download/player_stats/player_stats.qs"
#'   )
#' })
#' }
qs_from_url <- function(url) {
  lifecycle::deprecate_stop(
    "1.6.0", "qs_from_url()",
    details = "Please use one of the other `*_from_url()` functions."
  )
}

cache_message <- function() {
  do_it <- getOption("nflreadr.verbose", default = interactive()) &&
    getOption("nflreadr.cache_warning", default = interactive())
  if (isTRUE(do_it)) {
    rlang::inform(
      message = c(
        "Note: nflreadr caches (i.e., stores a saved version) data by default.\nIf you expect different output try one of the following:",
        i = "Restart your R Session or",
        i = "Run `nflreadr::.clear_cache()`.",
        "To disable this warning, run `options(nflreadr.verbose = FALSE)` or add it to your .Rprofile"
      ),
      .frequency = "regularly",
      .frequency_id = "cache_message"
    )
  }
}


loader <- function(url) {
  switch(
    detect_filetype(url),
    "rds" = rds_from_url(url),
    "parquet" = parquet_from_url(url),
    "csv" = csv_from_url(url)
  )
}

detect_filetype <- function(url) {
  tools::file_ext(gsub(x = url, pattern = ".gz$", replacement = ""))
}

test_that("from_url fails nicely", {
  skip_on_cran()
  skip_if_offline("github.com")

  expect_warning(
    rds_from_url(
      "https://github.com/nflverse/nfldata/blob/master/data/games.rds"
    ),
    regexp = "Failed to readRDS"
  )

  expect_warning(
    raw_from_url("https://github.com/nflverse/nfl/blob/master/data/games.rd"),
    regexp = "HTTP error"
  )

  expect_error(
    qs_from_url(
      "https://github.com/nflverse/nfldata/raw/master/data/games.rds"
    ),
    regexp = "deprecated in nflreadr 1.6.0"
  )
})

test_that("progress updates in raw_from_url work", {
  skip_on_cran()
  skip_if_offline("github.com")

  # enable progress updates in batch mode for testing the progress updates
  old <- options(progressr.enable = TRUE)

  urls <- rep(
    "https://github.com/nflverse/nflverse-pbp/raw/master/teams_colors_logos.rds",
    3
  )

  expect_condition(
    {
      load <- progressr::with_progress({
        p <- progressr::progressor(along = urls)
        lapply(urls, progressively(raw_from_url, p = p))
      })
    },
    class = "progression"
  )

  expect_null({
    capture_condition({
      load <- progressr::without_progress({
        p <- progressr::progressor(along = urls)
        lapply(urls, progressively(raw_from_url, p = p))
      })
    })
  })

  expect_true(is.list(load))
  expect_true(is.raw(unlist(load)))
  options(old)
})

test_that("nflverse_download downloads files", {
  skip_on_cran()
  skip_if_offline("github.com")

  temp_dir <- tempdir(check = TRUE)

  expect_error(
    expect_warning(
      nflverse_download(
        "asdf",
        folder_path = temp_dir,
        .token = Sys.getenv("NFLVERSE_GH_TOKEN", unset = gh::gh_token())
      ),
      "Could not find"
    ),
    "No matching releases"
  )

  nflverse_download(
    combine,
    "contracts",
    folder_path = temp_dir,
    file_type = "parquet",
    .token = Sys.getenv("NFLVERSE_GH_TOKEN", unset = gh::gh_token())
  )

  expect_true(
    length(list.files(temp_dir, recursive = TRUE)) >= 2
  )
})

test_that("nflverse_releases lists releases", {
  skip_on_cran()
  skip_if_offline("github.com")

  releases <- nflverse_releases(
    .token = Sys.getenv("NFLVERSE_GH_TOKEN", unset = gh::gh_token())
  )

  expect_true(is.data.frame(releases))
  expect_true(nrow(releases) >= 15)
  expect_true(all(c("pbp", "player_stats") %in% releases$release_name))
})

#' Load Fantasy Player IDs
#'
#' Accesses DynastyProcess.com's database of fantasy football player IDs, which help connect nflverse to various other platforms and IDs.
#'
#' @return a dataframe of player IDs
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({ # prevents cran errors
#' load_ff_playerids()
#' })
#' }
#'
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_ff_playerids.html> for the web data dictionary
#' @seealso Issues with this data should be filed here: <https://github.com/dynastyprocess/data>
#'
#' @export
load_ff_playerids <- function() {
  out <- load_from_url(
    "https://github.com/dynastyprocess/data/raw/master/files/db_playerids.rds",
    nflverse = TRUE,
    nflverse_type = "(ffverse) player IDs"
  )
  return(out)
}

#' Load Latest FantasyPros Rankings
#'
#' Accesses DynastyProcess.com's repository of the latest FP expert consensus rankings - updated on a weekly basis.
#'
#' @param type one of "draft" (preseason), "week" (this week, inseason), or "all" (full archive)
#'
#' @return a dataframe of expert consensus rankings
#'
#' @examples
#' \donttest{
#' try({ # prevents cran errors
#' load_ff_rankings()
#' })
#' }
#'
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_ff_rankings.html> for the web data dictionary
#' @seealso <https://www.fantasypros.com> for the source of data
#' @seealso Issues with this data should be filed here: <https://github.com/dynastyprocess/data>
#'
#' @export
load_ff_rankings <- function(type = c("draft", "week", "all")) {
  type <- rlang::arg_match0(type, c("draft", "week", "all"))

  url <- switch(
    type,
    draft = "https://github.com/dynastyprocess/data/raw/master/files/db_fpecr_latest.rds",
    week = "https://github.com/dynastyprocess/data/raw/master/files/fp_latest_weekly.rds",
    all = "https://github.com/dynastyprocess/data/raw/master/files/db_fpecr.rds"
  )

  out <- load_from_url(
    url,
    nflverse = TRUE,
    nflverse_type = "FP expert rankings"
  )
  return(out)
}

#' Load Expected Fantasy Points
#'
#' This function downloads precomputed expected points data from [{ffopportunity}](https://ffopportunity.ffverse.com) automated releases.
#'
#' @param seasons a numeric vector of seasons to return, defaults to most recent season. If set to `TRUE`, returns all available data.
#' @param stat_type one of `"weekly"`, `"pbp_pass"`, `"pbp_rush"`
#' @param model_version one of `"latest"` or `"v1.0.0"`
#'
#' @examples
#' \donttest{
#'   try({ # prevents cran errors
#'   load_ff_opportunity()
#'   load_ff_opportunity(seasons = 2021, stat_type = "pbp_pass", model_version = "v1.0.0")
#'   })
#' }
#'
#' @return Precomputed expected fantasy points data from the ffopportunity automated releases.
#'
#' @seealso <https://ffopportunity.ffverse.com> for more on the package, data, and modelling
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_ff_opportunity.html> for the web data dictionary
#' @seealso [`dictionary_ff_opportunity`] for the data dictionary bundled as a package data frame
#' @seealso Issues with this data should be filed here: <https://github.com/ffverse/ffopportunity>
#'
#' @export
load_ff_opportunity <- function(
  seasons = most_recent_season(),
  stat_type = c("weekly", "pbp_pass", "pbp_rush"),
  model_version = c("latest", "v1.0.0")
) {
  if (isTRUE(seasons)) {
    seasons <- 2006:most_recent_season()
  }

  stopifnot(
    is.numeric(seasons),
    seasons >= 2006,
    seasons <= most_recent_season()
  )

  stat_type <- rlang::arg_match0(stat_type, c("weekly", "pbp_pass", "pbp_rush"))
  model_version <- rlang::arg_match0(model_version, c("latest", "v1.0.0"))

  urls <- glue::glue(
    "https://github.com/ffverse/ffopportunity/releases/download/{model_version}-data/ep_{stat_type}_{seasons}.rds"
  )

  out <- load_from_url(
    urls,
    nflverse = TRUE,
    nflverse_type = glue::glue("ffopportunity expected points: {stat_type}")
  )
  return(out)
}

test_that("load_ff_rankings", {
  skip_if_offline("github.com")
  skip_on_cran()

  player_ids <- load_ff_playerids()

  expect_s3_class(player_ids, "tbl_df")

  expect_gt(nrow(player_ids), 6000)
})

test_that("load_ff_rankings", {
  skip_if_offline("github.com")
  skip_on_cran()

  rankings_draft <- load_ff_rankings(type = "draft")
  rankings_week <- load_ff_rankings(type = "week")

  expect_s3_class(rankings_draft, "tbl_df")
  expect_gt(nrow(rankings_draft), 500)
  expect_s3_class(rankings_week, "tbl_df")
  expect_gt(nrow(rankings_week), 10)
})

test_that("load_ff_opportunity", {
  skip_if_offline("github.com")
  skip_on_cran()

  ff_opps_week <- load_ff_opportunity(seasons = TRUE)
  ff_opps_pbp <- load_ff_opportunity(
    seasons = 2021,
    stat_type = "pbp_pass",
    model_version = "v1.0.0"
  )

  expect_s3_class(ff_opps_week, "tbl_df")
  expect_gt(nrow(ff_opps_week), 80000)
  expect_s3_class(ff_opps_pbp, "tbl_df")
  expect_gt(nrow(ff_opps_pbp), 19000)
})

#' Load Trades
#'
#' This returns a table of historical trades as maintained by Lee Sharpe.
#'
#' @param seasons a numeric vector of seasons to return, default `TRUE` returns all available data.
#'
#' @return A tibble of game information for past and/or future games.
#'
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_trades.html> for a web version of the dictionary
#' @seealso [`dictionary_trades`] for the data dictionary as bundled within the package
#' @seealso Issues with this data should be filed here: <https://github.com/nflverse/nfldata>
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#'  load_trades(2020)
#' }
#'
#' @export
load_trades <- function(seasons = TRUE) {
  if (!isTRUE(seasons)) {
    stopifnot(is.numeric(seasons))
  }

  out <- load_from_url(
    "https://github.com/nflverse/nfldata/raw/master/data/trades.rds",
    seasons = seasons,
    nflverse = TRUE,
    nflverse_type = "trades (via PFR / Lee Sharpe)"
  )

  return(out)
}

#' CRAN setup code
#'
#' Some code to comply with CRAN policies, e.g.
#' limiting data.table CPU cores for examples/tests
#'
#' @export
#' @keywords internal
.for_cran <- function() {
  cores <- min(
    getOption("Ncpus", default = 2L),
    as.integer(Sys.getenv("OMP_THREAD_LIMIT", unset = "2")),
    floor(as.integer(Sys.getenv(
      "_R_CHECK_EXAMPLE_TIMING_CPU_TO_ELAPSED_THRESHOLD_",
      unset = 2
    ))),
    2L,
    na.rm = TRUE
  )

  data.table::setDTthreads(cores)
}

#' Load Officials
#'
#' @description Loads data on which officials are assigned to oversee a specific game. Data available from 2015 onwards.
#'
#' @param seasons a numeric vector specifying what seasons to return, if `TRUE` returns all available data
#' @param file_type One of `c("rds", "csv", "parquet")`. Can also be set globally with
#' `options(nflreadr.prefer)`
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({ # prevents cran errors
#'   load_officials()
#' })
#' }
#'
#' @return A tibble with one row per game per official.
#'
#' @seealso Issues with this data should be filed here: <https://github.com/nflverse/nflreadr> and it will be triaged appropriately.
#'
#' @export
load_officials <- function(
  seasons = TRUE,
  file_type = getOption("nflreadr.prefer", default = "rds")
) {
  file_type <- rlang::arg_match0(file_type, c("rds", "csv", "parquet"))
  url <- glue::glue(
    "https://github.com/nflverse/nflverse-data/releases/download/officials/officials.{file_type}"
  )

  if (!isTRUE(seasons)) {
    stopifnot(
      is.numeric(seasons),
      all(seasons >= 2015),
      all(seasons <= nflreadr::most_recent_season())
    )
  }

  out <- load_from_url(url, seasons = seasons, nflverse = TRUE)
  return(out)
}

#' Data Dictionary: Play by Play
#'
#' A dataframe containing the data dictionary for [`load_pbp()`]
#'
#' @seealso `vignette("Data Dictionary - PBP")`
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_pbp.html>
"dictionary_pbp"

#' Data Dictionary: Player Stats
#'
#' A dataframe containing the data dictionary for [`load_player_stats()`]
#'
#' @seealso `vignette("Data Dictionary - Player Stats")`
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_player_stats.html>
"dictionary_player_stats"

#' Data Dictionary: Team Stats
#'
#' A dataframe containing the data dictionary for [`load_team_stats()`]
#'
#' @seealso `vignette("Data Dictionary - Team Stats")`
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_team_stats.html>
"dictionary_team_stats"

#' Data Dictionary: Fantasy Player IDs
#'
#' A dataframe containing the data dictionary for [`load_ff_playerids()`]
#'
#' @seealso `vignette("Data Dictionary - FF Player IDs")`
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_ff_playerids.html>
"dictionary_ff_playerids"

#' Data Dictionary: Fantasy Football Rankings
#'
#' A dataframe containing the data dictionary for [`load_ff_rankings()`]
#'
#' @seealso `vignette("Data Dictionary - FF Rankings")`
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_ff_rankings.html>
"dictionary_ff_rankings"

#' Data Dictionary: Expected Fantasy Points
#'
#' A dataframe containing the data dictionary for [`load_ff_opportunity()`]
#'
#' @seealso `vignette("Data Dictionary - Expected Fantasy Points")`
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_ff_opportunity.html>
"dictionary_ff_opportunity"

#' Data Dictionary: Rosters
#'
#' A dataframe containing the data dictionary for [`load_rosters()`]
#'
#' @seealso `vignette("Data Dictionary - Rosters")`
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_rosters.html>
"dictionary_rosters"

#' Data Dictionary: Schedules
#'
#' A dataframe containing the data dictionary for [`load_schedules()`]
#'
#' @seealso `vignette("Data Dictionary - Schedules")`
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_schedules.html>
"dictionary_schedules"

#' Data Dictionary: Next Gen Stats
#'
#' A dataframe containing the data dictionary for [`load_nextgen_stats()`]
#'
#' @seealso `vignette("Data Dictionary - Next Gen Stats")`
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_nextgen_stats.html>
#'
"dictionary_nextgen_stats"

#' Data Dictionary: PFR Passing
#'
#' A dataframe containing the data dictionary for [`load_pfr_passing()`]
#'
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_pfr_passing.html>
#' @seealso `vignette("Data Dictionary - PFR Passing")`
"dictionary_pfr_passing"

#' Data Dictionary: Draft Picks
#'
#' A dataframe containing the data dictionary for [`load_draft_picks()`]
#'
#' @seealso `vignette("Data Dictionary - Draft Picks")`
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_draft_picks.html>
"dictionary_draft_picks"

#' Data Dictionary: Snap Counts
#'
#' A dataframe containing the data dictionary for [`load_snap_counts()`]
#'
#' @seealso `vignette("Data Dictionary - Snap Counts")`
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_snap_counts.html>
"dictionary_snap_counts"

#' Data Dictionary: ESPN QBR
#'
#' A dataframe containing the data dictionary for [`load_espn_qbr()`]
#'
#' @seealso `vignette("Data Dictionary - ESPN QBR")`
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_espn_qbr.html>
"dictionary_espn_qbr"

#' Data Dictionary: Combine
#'
#' A dataframe containing the data dictionary for [`load_combine()`]
#'
#' @seealso `vignette("Data Dictionary - Combine")`
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_combine.html>
"dictionary_combine"

#' Data Dictionary: Injuries
#'
#' A dataframe containing the data dictionary for [`load_injuries()`]
#'
#' @seealso `vignette("Data Dictionary - Injuries")`
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_injuries.html>
"dictionary_injuries"

#' Data Dictionary: Depth Charts
#'
#' A dataframe containing the data dictionary for [`load_depth_charts()`]
#'
#' @seealso `vignette("Data Dictionary - Depth Charts")`
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_depth_charts.html>
"dictionary_depth_charts"

#' Data Dictionary: Trades
#'
#' A dataframe containing the data dictionary for [`load_trades()`]
#'
#' @seealso `vignette("Data Dictionary - Trades")`
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_trades.html>
"dictionary_trades"

#' Data Dictionary: Contracts
#'
#' A dataframe containing the data dictionary for [`load_contracts()`]
#'
#' @seealso `vignette("Data Dictionary - Contracts")`
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_contracts.html>
"dictionary_contracts"

#' Data Dictionary: Participation
#'
#' A dataframe containing the data dictionary for [`load_participation()`]
#'
#' @seealso `vignette("Data Dictionary - Participation")`
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_participation.html>
"dictionary_participation"

#' Alternate player name mappings
#'
#' A named character vector mapping common alternate names, re-exported from `ffscrapr`.
#'
#' You can suggest additions to this table by [opening an issue in ffscrapr](https://github.com/ffverse/ffscrapr/issues/new/choose).
#'
#' @examples
#' \donttest{
#' player_name_mapping[c("Chatarius Atwell", "Robert Kelley")]
#' }
#'
#' @format A named character vector
#' \describe{
#'   \item{name attribute}{The "alternate" name.}
#'   \item{value attribute}{The "correct" name.}
#' }
"player_name_mapping"

#' Alternate team abbreviation mappings
#'
#' A named character vector mapping common alternate team abbreviations.
#'
#' You can suggest additions to this table by [opening an issue in nflreadr](https://github.com/nflverse/nflreadr/issues/new/choose).
#'
#' @examples
#' \donttest{
#' team_abbr_mapping[c("STL", "OAK","CRD","BLT", "CLV")]
#' }
#'
#' @seealso `team_abbr_mapping_norelocate` for the same thing but relocations stay in their original cities.
#'
#' @format A named character vector
#' \describe{
#'   \item{name attribute}{The "alternate" name.}
#'   \item{value attribute}{The "correct" name.}
#' }
"team_abbr_mapping"

#' Alternate team abbreviation mappings, no relocation
#'
#' A named character vector mapping common alternate team abbreviations, but does not follow relocations to their current city.
#'
#' You can suggest additions to this table by [opening an issue in nflreadr](https://github.com/nflverse/nflreadr/issues/new/choose).
#'
#' @examples
#' \donttest{
#' team_abbr_mapping_norelocate[c("STL", "OAK","CRD","BLT", "CLV")]
#' }
#'
#' @format A named character vector
#' \describe{
#'   \item{name attribute}{The "alternate" name.}
#'   \item{value attribute}{The "correct" name.}
#' }
"team_abbr_mapping_norelocate"

#' nflverse Timezone
#'
#' A character string defining the default timezone for data across the nflverse
#'
"nflverse_data_timezone"

#' dictionary_roster_status
#'
#' A dictionary translating the shorthand roster status to more verbose explanations of what each status indicates.
"dictionary_roster_status"

#' Data Dictionary: Players
#'
#' A dataframe containing the data dictionary for [`load_players()`]
#'
#' @seealso `vignette("Data Dictionary - Players")`
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_players.html>
"dictionary_players"

#' Get Latest Season
#'
#' A helper function to choose the most recent season available for a given dataset.
#'
#' This used to return new season on Thursday following Labor Day but NFL
#' announced in March 2026 that the 2026 season will kick off on
#' Wednesday because the TNF game will be played in Australia. We expect this
#' to happen in future seasons as well and decided to flip the switch on
#' Wednesday going forward.
#'
#' @param roster Either `TRUE` or `FALSE`.
#' If `TRUE`, will return current year after March 15th, otherwise previous year.
#' If `FALSE`, will return current year on or after Wednesday following Labor Day,
#' i.e. Wednesday after the first Monday in September. Otherwise previous year.
#'
#' @rdname latest_season
#' @return most recent season (a four digit numeric)
#' @family Date utils
#' @export
most_recent_season <- function(roster = FALSE) {
  today <- Sys.Date()
  current_year <- as.integer(format(today, format = "%Y"))
  current_month <- as.integer(format(today, format = "%m"))
  current_day <- as.integer(format(today, format = "%d"))
  # First Monday of September
  labor_day <- compute_labor_day(current_year)
  # Wednesday following Labor Day is season opener (starting in 2026)
  season_opener <- labor_day + 2

  if (
    (isFALSE(roster) && today >= season_opener) ||
      (isTRUE(roster) && current_month == 3 && current_day >= 15) ||
      (isTRUE(roster) && current_month > 3)
  ) {
    return(current_year)
  }

  return(current_year - 1)
}

#' @rdname latest_season
#' @export
get_latest_season <- most_recent_season

#' @rdname latest_season
#' @export
get_current_season <- most_recent_season


#' Get Current Week
#'
#' A helper function that returns the upcoming NFL regular season week based on either
#' the nflverse schedules file (as found in `load_schedules()`)
#' or some date-based heuristics (number of weeks since the first Monday of September)
#'
#' Note that the date heuristic will count a new week starting on Thursdays, while
#' the schedule-based method will count a new week after the last game of the previous
#' week, e.g. after MNF is completed. Tan and Ben argued for a while about this.
#'
#' @param use_date a logical to determine whether to use date-based heuristics to
#' determine current week, default FALSE (i.e. uses schedule file)
#' @param ... Arguments passed on to [most_recent_season()]
#'
#' @examples {
#'   \donttest{
#'     try({ # schedules file as per default requires online access
#'     get_current_week()
#'     })
#'   }
#'   # using the date method works offline
#'   get_current_week(use_date = TRUE)
#' }
#'
#' @family Date utils
#' @return current nfl regular season week as a numeric
#' @export
get_current_week <- function(use_date = FALSE, ...) {
  if (!use_date) {
    season <- NULL
    week <- NULL
    result <- NULL
    current_season <- data.table::as.data.table(load_schedules())[
      season == most_recent_season(...)
    ]

    if (all(!is.na(current_season$result))) {
      return(max(current_season$week, na.rm = TRUE))
    }

    current_week <- current_season[is.na(result), week]
    return(min(current_week, na.rm = TRUE))
  }

  if (use_date) {
    # Find first Monday of September in current season
    week1_sep <- as.POSIXlt(
      paste0(most_recent_season(...), "-09-0", 1:7),
      tz = "GMT"
    )
    monday1_sep <- week1_sep[week1_sep$wday == 1]

    # NFL season starts 3 days later
    first_game <- monday1_sep
    first_game$mday <- first_game$mday + 3

    # current week number of nfl season is 1 + how many weeks have elapsed since first game
    current_week <- as.numeric(Sys.Date() - as.Date(first_game)) %/% 7 + 1

    # hardcoded week bounds because this whole date based thing has assumptions anyway
    if (current_week < 1) {
      current_week <- 1
    }
    if (current_week > 22) {
      current_week <- 22
    }

    return(current_week)
  }
}

#' Compute Date of Labor Day
#'
#' Computes the date of the Labor Day, i.e. the first Monday in September, in a given year.
#'
#' @param year Numeric or Character year. 4 Digits.
#' @noRd
#' @return Object of Class `Date`
#' @examples
#' # 2023 Labor Day was on Sep 4th
#' compute_labor_day(2023)
compute_labor_day <- function(year) {
  stopifnot(length(year) == 1)
  earliest <- as.Date(paste(year, "09", "01", sep = "-"))
  latest <- as.Date(paste(year, "09", "08", sep = "-"))
  range <- seq(earliest, latest, by = "day")
  numeric_wdays <- as.POSIXlt(range)$wday
  labor_day <- range[numeric_wdays == 1][1]
  labor_day
}

test_that("latest week methods are equivalent", {
  skip_if(
    packageVersion('nflreadr') == '1.3.2',
    "Skip date check for v1.3.2 because of 2022 Bengals/Bills game problems."
  )

  skip_on_cran()
  skip_if_offline("github.com")
  # This test fails during the season on Tuesdays and Wednesdays
  # Blame Tan and Ben smh
  skip_if(as.POSIXlt(Sys.Date())$wday %in% 2:3)

  expect_equal(
    get_current_week(use_date = TRUE),
    get_current_week(use_date = FALSE)
  )
})

#' Load Injury Reports
#'
#' Data collected from an API for weekly injury report data.
#'
#' @param seasons a numeric vector of seasons to return, data available since 2009. Defaults to latest season available.
#' @param file_type One of `c("rds", "csv", "parquet")`. Can also be set globally with
#' `options(nflreadr.prefer)`
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({# prevents cran errors
#'     load_injuries(2020)
#' })
#' }
#'
#' @return a tibble of season-level injury report data.
#'
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_injuries.html> for a web version of the dictionary
#' @seealso [`dictionary_injuries`] for the data dictionary as bundled within the package
#' @seealso Issues with this data should be filed here: <https://github.com/nflverse/nflverse-data>
#'
#' @export
load_injuries <- function(
  seasons = most_recent_season(),
  file_type = getOption("nflreadr.prefer", default = "rds")
) {
  if (isTRUE(seasons)) {
    seasons <- 2009:nflreadr::most_recent_season()
  }

  file_type <- rlang::arg_match0(file_type, c("rds", "csv", "parquet"))
  stopifnot(
    is.numeric(seasons),
    seasons >= 2009,
    seasons <= most_recent_season()
  )

  urls <- glue::glue(
    "https://github.com/nflverse/nflverse-data/releases/download/injuries/injuries_{seasons}.{file_type}"
  )

  out <- load_from_url(urls, seasons = seasons, nflverse = TRUE)

  return(out)
}

#' Load Draft Picks from PFR
#'
#' @description Loads every draft pick since 1980 courtesy of PFR.
#'
#' @param seasons a numeric vector of seasons to return, default `TRUE` returns all available data
#' @param file_type One of `c("rds", "csv", "parquet")`. Can also be set globally with
#' `options(nflreadr.prefer)`
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({ # prevents cran errors
#'   load_draft_picks()
#' })
#' }
#'
#' @return A tibble of NFL draft picks provided by Pro Football Reference.
#'
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_draft_picks.html> for the web data dictionary
#' @seealso [`dictionary_draft_picks`] for the data dictionary as bundled within the package
#' @seealso Issues with this data should be filed here: <https://github.com/nflverse/nflverse-data>
#'
#' @export
load_draft_picks <- function(
  seasons = TRUE,
  file_type = getOption("nflreadr.prefer", default = "rds")
) {
  file_type <- rlang::arg_match0(file_type, c("rds", "csv", "parquet"))
  url <- glue::glue(
    "https://github.com/nflverse/nflverse-data/releases/download/draft_picks/draft_picks.{file_type}"
  )
  out <- load_from_url(url, seasons = seasons, nflverse = TRUE)
  return(out)
}

#' Load Player Level Weekly NFL Next Gen Stats
#'
#' @description Loads player level weekly stats provided by NFL Next Gen Stats
#' starting with the 2016 season. Three different stat types are available and
#' the current season's data updates every night. NGS will only provide data for
#' players above a minimum number of pass/rush/rec attempts.
#'
#' @param seasons a numeric vector specifying what seasons to return, if `TRUE` returns all available data
#' @param stat_type one of `"passing"`, `"receiving"`, or `"rushing"`
#' @param file_type One of `c("rds", "csv", "parquet")`. Can also be set globally with
#' `options(nflreadr.prefer)`
#'
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({ # prevents cran errors
#'   load_nextgen_stats(stat_type = "passing")
#'   load_nextgen_stats(stat_type = "receiving")
#'   load_nextgen_stats(stat_type = "rushing")
#' })
#' }
#'
#' @return A tibble of week-level player statistics provided by NFL Next Gen Stats.
#' Regular season summary is given for `week == 0`.
#'
#' @seealso <https://nextgenstats.nfl.com/stats/passing> for `stat_type = "passing"`
#' @seealso <https://nextgenstats.nfl.com/stats/receiving> for `stat_type = "receiving"`
#' @seealso <https://nextgenstats.nfl.com/stats/rushing> for `stat_type = "rushing"`
#'
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_nextgen_stats.html> for a web version of the data dictionary
#' @seealso [`dictionary_nextgen_stats`] for the data dictionary as bundled within the package
#' @seealso Issues with this data should be filed here: <https://github.com/nflverse/nflverse-data>
#'
#' @export
load_nextgen_stats <- function(
  seasons = TRUE,
  stat_type = c("passing", "receiving", "rushing"),
  file_type = getOption("nflreadr.prefer", default = "rds")
) {
  if (!isTRUE(seasons)) {
    stopifnot(
      is.numeric(seasons),
      seasons >= 2016,
      seasons <= most_recent_season()
    )
  }

  file_type <- rlang::arg_match0(file_type, c("rds", "parquet", "csv"))
  stat_type <- rlang::arg_match0(
    stat_type,
    c("passing", "receiving", "rushing")
  )

  url <- paste0(
    "https://github.com/nflverse/nflverse-data/releases/download/",
    "nextgen_stats/ngs_",
    stat_type,
    ".",
    file_type
  )

  out <- load_from_url(url, seasons = seasons, nflverse = TRUE)
  out
}

#' Load Weekly Depth Charts
#'
#' @description Loads depth charts for each NFL team for each week back to 2001.
#'
#' @param seasons a numeric vector specifying what seasons to return, if `TRUE` returns all available data. Defaults to latest season with available rosters.
#' @param file_type One of `c("rds", "csv", "parquet")`. Can also be set globally with
#' `options(nflreadr.prefer)`
#'
#' @examples
#' \dontshow{.for_cran()}
#' \donttest{
#' try({ # prevents cran errors
#'   load_depth_charts(2020)
#' })
#' }
#'
#' @seealso <https://nflreadr.nflverse.com/articles/dictionary_depth_charts.html> for a web version of the dictionary
#' @seealso [`dictionary_depth_charts`] for the data dictionary as bundled within the package
#' @seealso Issues with this data should be filed here: <https://github.com/nflverse/nflverse-data>
#'
#' @return A tibble of week-level depth charts for each team.
#' @export
load_depth_charts <- function(
  seasons = most_recent_season(roster = TRUE),
  file_type = getOption("nflreadr.prefer", default = "rds")
) {
  if (isTRUE(seasons)) {
    seasons <- 2001:most_recent_season(roster = TRUE)
  }
  stopifnot(
    is.numeric(seasons),
    seasons >= 2001,
    seasons <= most_recent_season(roster = TRUE)
  )

  urls <- glue::glue(
    "https://github.com/nflverse/nflverse-data/releases/download/depth_charts/depth_charts_{seasons}.rds"
  )
  out <- load_from_url(urls, seasons = seasons, nflverse = TRUE)
  return(out)
}

---
title: "Data Dictionary - Players"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Data Dictionary - Players}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---
```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  echo = FALSE,
  comment = "#>"
)
use_dt <- FALSE
if(requireNamespace("DT", quietly = TRUE)) use_dt <- TRUE
```
```{r eval = use_dt}
DT::datatable(nflreadr::dictionary_players, 
              options = list(scrollX = TRUE, pageLength = 25),
              filter = "top",
              rownames = FALSE
              )
```
```{r eval = !use_dt}
knitr::kable(nflreadr::dictionary_players)
```

---
title: "Data Dictionary - Depth Charts"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Data Dictionary - Depth Charts}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

Please note that the data source has changed after the 2024 season. From 2025 onwards, the following variables and descriptions will apply.

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  echo = FALSE,
  comment = "#>"
)
```

```{r }
knitr::kable(nflreadr::dictionary_depth_charts)
```

---
title: "Data Dictionary - Team Stats"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Data Dictionary - Team Stats}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---
```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  echo = FALSE,
  comment = "#>"
)
use_dt <- FALSE
if(requireNamespace("DT", quietly = TRUE)) use_dt <- TRUE
```

```{r eval = use_dt}
DT::datatable(
  nflreadr::dictionary_team_stats,
  options = list(scrollX = TRUE, pageLength = 25),
  filter = "top",
  rownames = FALSE
)
```

---
title: "Data Dictionary - Combine"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Data Dictionary - Combine}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  echo = FALSE,
  comment = "#>"
)
```

```{r }
knitr::kable(nflreadr::dictionary_combine)
```

---
title: "Using nflreadr in packages"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Using nflreadr in packages}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

## Re-exporting nflreadr functions

You can re-export `nflreadr` functions in your package by using the following roxygen template:

```{r eval = FALSE}
#' @inherit nflreadr::load_nextgen_stats
#' @export
# Need to add own examples if the function name is different
#' @examples
#' load_ng_stats(2020)
load_ng_stats <- nflreadr::load_nextgen_stats
```

## Using progressr with from_url functions

The `from_url` family of functions can be adapted with the `progressively()` function decorator to signal progress. Here's an example, adapted from the `load_rosters()` source code:

```{r eval = FALSE}
load_rosters <- function(seasons = 1999:2020){

  # Create a progressor function inside your function that knows how many "steps" there will be
  p <- progressr::progressor(steps = length(seasons)) 

  # Form the URLs to pass into rds_from_url
  urls <- paste0(
    "https://github.com/nflverse/nflverse-rosters/",
    "raw/master/data/seasons/roster_",
    seasons, 
    ".rds")

  # Use `progressively()` to wrap the existing function rds_from_url and 
  # call the `p()` progressor function as a signal immediately after each iteration
  
  out <- purrr::map_dfr(urls, progressively(rds_from_url,p))
  
  return(out)
}
```

This will cause p() to execute once for each rds_from_url call, effectively "signalling" progress.

In order to receive progress updates, the user must wrap the function (load_rosters, in this case) with `progressr::with_progress()` as shown here:

```{r eval = FALSE}
progressr::with_progress(load_rosters(2010:2020))
```

For more information, please see the [progressr](https://progressr.futureverse.org) documentation.

---
title: "Data Dictionary - Schedules"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Data Dictionary - Schedules}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  echo = FALSE,
  comment = "#>"
)
use_dt <- FALSE
if(requireNamespace("DT", quietly = TRUE)) use_dt <- TRUE
```
```{r eval = use_dt}
DT::datatable(nflreadr::dictionary_schedules,
              options = list(scrollX = TRUE, pageLength = 25),
              filter = "top",
              rownames = FALSE
              )
```
```{r eval = !use_dt}
knitr::kable(nflreadr::dictionary_schedules)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_sitrep.R
\name{nflverse_sitrep}
\alias{nflverse_sitrep}
\alias{ffverse_sitrep}
\alias{.sitrep}
\title{Get a Situation Report on System, nflverse/ffverse Package Versions and Dependencies}
\usage{
nflverse_sitrep(
  pkg = c("nflreadr", "nflfastR", "nflseedR", "nfl4th", "nflplotR", "nflverse"),
  recursive = TRUE,
  redact_path = TRUE
)

ffverse_sitrep(
  pkg = c("ffscrapr", "ffsimulator", "ffpros", "ffopportunity"),
  recursive = TRUE,
  redact_path = TRUE
)

.sitrep(
  pkg,
  recursive = TRUE,
  redact_path = TRUE,
  dev_repos = c("https://nflverse.r-universe.dev", "https://ffverse.r-universe.dev")
)
}
\arguments{
\item{pkg}{a character vector naming installed packages, or \code{NULL}
(the default) meaning all nflverse packages. The function checks internally
if all packages are installed and informs if that is not the case.}

\item{recursive}{a logical indicating whether dependencies of \code{pkg} and their
dependencies (and so on) should be included.
Can also be a character vector listing the types of dependencies, a subset
of \code{c("Depends", "Imports", "LinkingTo", "Suggests", "Enhances")}.
Character string \code{"all"} is shorthand for that vector, character string
\code{"most"} for the same vector without \code{"Enhances"}, character string \code{"strong"}
(default) for the first three elements of that vector.}

\item{redact_path}{a logical indicating whether options that contain "path"
in the name should be redacted, default = TRUE}

\item{dev_repos}{Developmental cran-like repos to check, e.g. r-universe repos}
}
\description{
This function gives a quick overview of the versions of R and the operating
system as well as the versions of nflverse/ffverse packages, options, and their dependencies.
It's primarily designed to help you get a quick idea of what's going on when
you're helping someone else debug a problem.
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({
nflverse_sitrep()
ffverse_sitrep()
.sitrep("cachem")
})
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_name_cleaning.R
\name{clean_player_names}
\alias{clean_player_names}
\title{Create Player Merge Names}
\usage{
clean_player_names(
  player_name,
  lowercase = FALSE,
  convert_lastfirst = TRUE,
  use_name_database = TRUE,
  convert_to_ascii = rlang::is_installed("stringi")
)
}
\arguments{
\item{player_name}{a character vector of player names}

\item{lowercase}{defaults to FALSE - if TRUE, converts to lowercase}

\item{convert_lastfirst}{defaults to TRUE - converts names from "Last, First" to "First Last"}

\item{use_name_database}{uses internal name database to do common substitutions (Mitchell Trubisky to Mitch Trubisky etc)}

\item{convert_to_ascii}{If \code{TRUE}, will transliterate to latin-ascii via the
stringi package. Defaults to \code{TRUE} if the stringi package is installed.}
}
\value{
a character vector of cleaned names
}
\description{
Applies some name-cleaning heuristics to facilitate joins. These heuristics may include:
\itemize{
\item removing periods and apostrophes
\item removing common suffixes, such as Jr, Sr, II, III, IV
\item converting to lowercase
\item using \code{ffscrapr::dp_name_mapping} to do common name substitutions, such as Mitch Trubisky to Mitchell Trubisky
}
}
\details{
Equivalent to the operation done by \code{ffscrapr::dp_clean_names()} and uses the same player name database.
}
\examples{
\dontshow{.for_cran()}
clean_player_names(c("A.J. Green", "Odell Beckham Jr.  ", "Le'Veon Bell Sr."))

clean_player_names(c("Trubisky,      Mitch", "Atwell, Chatarius", "Elliott, Zeke", "Elijah Moore"),
                  convert_lastfirst = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{dictionary_depth_charts}
\alias{dictionary_depth_charts}
\title{Data Dictionary: Depth Charts}
\format{
An object of class \code{data.frame} with 12 rows and 3 columns.
}
\usage{
dictionary_depth_charts
}
\description{
A dataframe containing the data dictionary for \code{\link[=load_depth_charts]{load_depth_charts()}}
}
\seealso{
\code{vignette("Data Dictionary - Depth Charts")}

\url{https://nflreadr.nflverse.com/articles/dictionary_depth_charts.html}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{dictionary_nextgen_stats}
\alias{dictionary_nextgen_stats}
\title{Data Dictionary: Next Gen Stats}
\format{
An object of class \code{data.frame} with 51 rows and 3 columns.
}
\usage{
dictionary_nextgen_stats
}
\description{
A dataframe containing the data dictionary for \code{\link[=load_nextgen_stats]{load_nextgen_stats()}}
}
\seealso{
\code{vignette("Data Dictionary - Next Gen Stats")}

\url{https://nflreadr.nflverse.com/articles/dictionary_nextgen_stats.html}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_mode.R
\name{stat_mode}
\alias{stat_mode}
\title{Statistical Mode}
\usage{
stat_mode(x, ..., na.rm = FALSE)
}
\arguments{
\item{x}{A vector of data values.}

\item{...}{Further arguments, currently unused.}

\item{na.rm}{a logical evaluating to \code{TRUE} or \code{FALSE} indicating whether
\code{NA} values should be stripped before the computation proceeds.}
}
\value{
The statistical mode with the same type as the input vector x.
}
\description{
Computes the statistical mode, i.e. the value that appears most often in a
vector. Returns the first match, if TRUE for multiple values.
}
\examples{
vector_numeric <- sample(1:5, 15, TRUE)
vector_numeric
stat_mode(vector_numeric)

vector_character <- sample(LETTERS[1:5], 15, TRUE)
vector_character
stat_mode(vector_character)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_progressively.R
\name{progressively}
\alias{progressively}
\title{Progressively}
\usage{
progressively(f, p = NULL)
}
\arguments{
\item{f}{a function to add progressor functionality to.}

\item{p}{a function such as one created by \code{progressr::progressor()} - also accepts purrr-style lambda functions.}
}
\value{
a function that does the same as \code{f} but it calls \code{p()} after iteration.
}
\description{
This function helps add progress-reporting to any function - given function \code{f()} and progressor \code{p()},
it will return a new function that calls \code{f()} and then (on exiting) will call \code{p()} after every iteration.
This is inspired by purrr's \code{safely}, \code{quietly}, and \code{possibly} function decorators.
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({ # prevents cran errors

urls <- rep("https://github.com/nflverse/nflverse-data/releases/download/test/combines.csv",3)

lapply(urls, progressively(read.csv, ~cli::cli_progress_step('Loading...')))

read_rosters <- function(urls){
  p <- progressr::progressor(along = urls)
  lapply(urls, progressively(read.csv, p))
}

progressr::with_progress(read_rosters())

})
}

}
\seealso{
\url{https://nflreadr.nflverse.com/articles/exporting_nflreadr.html} for vignette on exporting nflreadr in packages
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_name_cleaning.R
\name{clean_homeaway}
\alias{clean_homeaway}
\title{Clean Home/Away in dataframes into Team/Opponent dataframes}
\usage{
clean_homeaway(dataframe, invert = NULL)
}
\arguments{
\item{dataframe}{dataframe}

\item{invert}{a character vector of columns that gets inverted when referring to the away team (e.g. home spread = 1 gets converted to away_spread = -1)}
}
\value{
a dataframe with one row per team (twice as long as the input dataframe)
}
\description{
This function converts dataframes with "home_" and "away_" prefixed columns to "team_" and "opponent_", and doubles the rows. This makes sure that there's one row for each team (as opposed to one row for each game).
}
\examples{
\dontshow{.for_cran()}
# a small example dataframe
s <- data.frame(
   game_id = c("2020_20_TB_GB", "2020_20_BUF_KC", "2020_21_KC_TB"),
   game_type = c("CON", "CON", "SB"),
   away_team = c("TB", "BUF", "KC"),
   away_score = c(31L, 24L, 9L),
   home_team = c("GB", "KC", "TB"),
   home_score = c(26L, 38L, 31L),
   location = c("Home", "Home", "Neutral"),
   result = c(-5L, 14L, 22L),
   spread_line = c(3, 3, -3)
 )

clean_homeaway(s, invert = c("result","spread_line"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_nextgen_stats.R
\name{load_nextgen_stats}
\alias{load_nextgen_stats}
\title{Load Player Level Weekly NFL Next Gen Stats}
\usage{
load_nextgen_stats(
  seasons = TRUE,
  stat_type = c("passing", "receiving", "rushing"),
  file_type = getOption("nflreadr.prefer", default = "rds")
)
}
\arguments{
\item{seasons}{a numeric vector specifying what seasons to return, if \code{TRUE} returns all available data}

\item{stat_type}{one of \code{"passing"}, \code{"receiving"}, or \code{"rushing"}}

\item{file_type}{One of \code{c("rds", "csv", "parquet")}. Can also be set globally with
\code{options(nflreadr.prefer)}}
}
\value{
A tibble of week-level player statistics provided by NFL Next Gen Stats.
Regular season summary is given for \code{week == 0}.
}
\description{
Loads player level weekly stats provided by NFL Next Gen Stats
starting with the 2016 season. Three different stat types are available and
the current season's data updates every night. NGS will only provide data for
players above a minimum number of pass/rush/rec attempts.
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({ # prevents cran errors
  load_nextgen_stats(stat_type = "passing")
  load_nextgen_stats(stat_type = "receiving")
  load_nextgen_stats(stat_type = "rushing")
})
}

}
\seealso{
\url{https://nextgenstats.nfl.com/stats/passing} for \code{stat_type = "passing"}

\url{https://nextgenstats.nfl.com/stats/receiving} for \code{stat_type = "receiving"}

\url{https://nextgenstats.nfl.com/stats/rushing} for \code{stat_type = "rushing"}

\url{https://nflreadr.nflverse.com/articles/dictionary_nextgen_stats.html} for a web version of the data dictionary

\code{\link{dictionary_nextgen_stats}} for the data dictionary as bundled within the package

Issues with this data should be filed here: \url{https://github.com/nflverse/nflverse-data}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/from_url.R
\name{qs_from_url}
\alias{qs_from_url}
\title{Load .qs file from a remote connection}
\usage{
qs_from_url(url)
}
\arguments{
\item{url}{a character url}
}
\value{
a dataframe
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

The underlying package qs has been removed from CRAN on 2026-01-17 so
nflreadr stopped the qs support
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({
  qs_from_url(
    "https://github.com/nflverse/nflverse-data/releases/download/player_stats/player_stats.qs"
  )
})
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{rbindlist_with_attrs}
\alias{rbindlist_with_attrs}
\title{rbindlist but maintain attributes of last file}
\usage{
rbindlist_with_attrs(dflist)
}
\description{
rbindlist but maintain attributes of last file
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{dictionary_trades}
\alias{dictionary_trades}
\title{Data Dictionary: Trades}
\format{
An object of class \code{data.frame} with 11 rows and 3 columns.
}
\usage{
dictionary_trades
}
\description{
A dataframe containing the data dictionary for \code{\link[=load_trades]{load_trades()}}
}
\seealso{
\code{vignette("Data Dictionary - Trades")}

\url{https://nflreadr.nflverse.com/articles/dictionary_trades.html}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{nflverse_data_timezone}
\alias{nflverse_data_timezone}
\title{nflverse Timezone}
\format{
An object of class \code{character} of length 1.
}
\usage{
nflverse_data_timezone
}
\description{
A character string defining the default timezone for data across the nflverse
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\docType{package}
\name{nflreadr-package}
\alias{nflreadr}
\alias{nflreadr-package}
\title{nflreadr: Download 'nflverse' Data}
\description{
A minimal package for downloading data from 'GitHub' repositories of the 'nflverse' project.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://nflreadr.nflverse.com}
  \item \url{https://github.com/nflverse/nflreadr}
  \item Report bugs at \url{https://github.com/nflverse/nflreadr/issues}
}

}
\author{
\strong{Maintainer}: Tan Ho \email{tan@tanho.ca} (\href{https://orcid.org/0000-0001-8388-5155}{ORCID}) [copyright holder]

Authors:
\itemize{
  \item Sebastian Carl \email{mrcaseb@gmail.com}
}

Other contributors:
\itemize{
  \item John Edwards \email{edwards1860@gmail.com} [contributor]
  \item Ben Baldwin \email{bbaldwin206@gmail.com} [contributor]
  \item Thomas Mock [contributor]
  \item Lee Sharpe [contributor]
  \item Pranav Rajaram \email{pranavrajaram@gmail.com} [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_participation.R
\name{load_participation}
\alias{load_participation}
\title{Load Participation Data}
\usage{
load_participation(
  seasons = most_recent_season(),
  include_pbp = FALSE,
  file_type = getOption("nflreadr.prefer", default = "rds")
)
}
\arguments{
\item{seasons}{A numeric vector of 4-digit years associated with given NFL
seasons - defaults to latest season. If set to \code{TRUE}, returns all available
data since 2016.}

\item{include_pbp}{a logical: download and join pbp to this data?}

\item{file_type}{One of \code{c("rds", "csv", "parquet")}. Can also be
set globally with \code{options(nflreadr.prefer)}}
}
\value{
A dataframe of participation data, optionally merged with play by play
}
\description{
Loads participation data from the \href{https://github.com/nflverse/nflverse-data}{nflverse-data repository}

Participation data prior to 2023 is from NFL NGS. Participation data from
2023 onwards is courtesy of FTN and is provided after all post-season games are
completed. This data is released under the \href{https://creativecommons.org/licenses/by-sa/4.0/}{CC-BY-SA 4.0}
Creative Commons license and attribution must be made to \strong{FTN Data via nflverse} (from 2023 onwards)
or \strong{NFL NextGenStats via nflverse} (for 2022 and earlier)
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({ # prevents cran errors
  load_participation(seasons = 2020, include_pbp = TRUE)
})
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_teams.R
\name{load_teams}
\alias{load_teams}
\title{Load NFL Team Graphics, Colors, and Logos}
\usage{
load_teams(
  current = TRUE,
  file_type = getOption("nflreadr.prefer", default = "rds")
)
}
\arguments{
\item{current}{If \code{TRUE} (the default), returns a standardized list of current teams only,
with abbreviations as per \link{team_abbr_mapping}.}

\item{file_type}{One of \code{c("rds", "csv", "parquet")}. Can also be set globally with
\code{options(nflreadr.prefer)}}
}
\value{
A tibble of team-level image URLs and hex color codes.
}
\description{
Loads team graphics, colors, and logos - useful for plots!
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({ # prevents cran errors
  load_teams()
})
}

}
\seealso{
Issues with this data should be filed here: \url{https://github.com/nflverse/nflverse-pbp}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_pbp.R
\name{load_pbp}
\alias{load_pbp}
\title{Load Play By Play}
\usage{
load_pbp(
  seasons = most_recent_season(),
  file_type = getOption("nflreadr.prefer", default = "rds")
)
}
\arguments{
\item{seasons}{A numeric vector of 4-digit years associated with given NFL seasons - defaults to latest season. If set to \code{TRUE}, returns all available data since 1999.}

\item{file_type}{One of \code{c("rds", "csv", "parquet")}. Can also be set globally with
\code{options(nflreadr.prefer)}}
}
\value{
The complete nflfastR dataset as returned by \code{nflfastR::build_nflfastR_pbp()}
(see below) for all given \code{seasons}
}
\description{
Loads play by play seasons from the \href{https://github.com/nflverse/nflverse-data}{nflverse-data repository}
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({ # prevents cran errors
  load_pbp(2019:2020)
})
}
}
\seealso{
\url{https://nflreadr.nflverse.com/articles/dictionary_pbp.html} for a web version of the data dictionary

\code{\link{dictionary_pbp}} for the data dictionary bundled as a package dataframe

\url{https://nflfastr.com/reference/build_nflfastR_pbp.html} for the nflfastR function \code{nflfastR::build_nflfastR_pbp()}

Issues with this data should be filed here: \url{https://github.com/nflverse/nflverse-pbp}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_trades.R
\name{load_trades}
\alias{load_trades}
\title{Load Trades}
\usage{
load_trades(seasons = TRUE)
}
\arguments{
\item{seasons}{a numeric vector of seasons to return, default \code{TRUE} returns all available data.}
}
\value{
A tibble of game information for past and/or future games.
}
\description{
This returns a table of historical trades as maintained by Lee Sharpe.
}
\examples{
\dontshow{.for_cran()}
\donttest{
 load_trades(2020)
}

}
\seealso{
\url{https://nflreadr.nflverse.com/articles/dictionary_trades.html} for a web version of the dictionary

\code{\link{dictionary_trades}} for the data dictionary as bundled within the package

Issues with this data should be filed here: \url{https://github.com/nflverse/nfldata}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_cache_clear.R
\name{clear_cache}
\alias{clear_cache}
\alias{.clear_cache}
\title{Clear function cache}
\usage{
clear_cache()

.clear_cache()
}
\value{
A success message after clearing the cache.
}
\description{
This function clears the memoised cache of all functions memoised by \code{nflreadr}.
}
\examples{
clear_cache()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_players.R
\name{load_players}
\alias{load_players}
\title{Load Players}
\usage{
load_players(file_type = getOption("nflreadr.prefer", default = "rds"))
}
\arguments{
\item{file_type}{One of \code{c("rds", "csv", "parquet")}. Can also be set globally with
\code{options(nflreadr.prefer)}}
}
\value{
A tibble with one row per player.
}
\description{
Load a dataframe of player-level information, including IDs and other mostly-immutable data (birthdates, college, draft position etc.)
}
\details{
\subsection{Contribute}{

Please visit \url{https://github.com/nflverse/nflverse-players/blob/master/.github/CONTRIBUTING.md}
for a detailed explanation how to contribute (or correct) players data.
}

\subsection{Scope of Players Dataset}{

The nflverse players dataset aims to be the single source of truth when it
comes to NFL player IDs across various sources (relevant to the nflverse).
The following source IDs and information are considered in-scope of the
nflverse players dataset:
\itemize{
\item Basic player information e.g. name, height, weight, age, date of birth,
years experience, college, status, position, headshot image (mostly from GSIS)
\item Draft information: draft year, draft round, draft pick, draft team (from PFR)
\item PFF position and status information
\item Primary source IDs from:
\itemize{
\item NFL (\code{gsis_id}, \code{smart_id}) - \code{gsis_id} is the primary key
\item Pro Football Reference (\code{pfr_id})
\item Pro Football Focus (\code{pff_id})
\item Over The Cap (\code{otc_id})
\item Elias Sports Bureau (\code{esb_id})
\item ESPN (\code{espn_id})
}
}
}
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({ # prevents cran errors
  load_players()
})
}

}
\seealso{
Issues with this data should be filed here: \url{https://github.com/nflverse/nflverse-players}.

\code{\link{dictionary_players}} for the data dictionary as bundled within the package
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_stats.R
\name{load_team_stats}
\alias{load_team_stats}
\title{Load Team Level Stats}
\usage{
load_team_stats(
  seasons = most_recent_season(),
  ...,
  summary_level = c("week", "reg", "post", "reg+post"),
  file_type = getOption("nflreadr.prefer", default = "rds")
)
}
\arguments{
\item{seasons}{a numeric vector of seasons to return, defaults to most recent
season. If set to \code{TRUE}, returns all available data.}

\item{...}{These dots are for future extensions and must be empty.}

\item{summary_level}{choice: one of week (default), "reg" for regular season,
"post" for postseason, "reg+post" for combined regular + post season stats}

\item{file_type}{choice: one of \code{c("rds", "csv", "parquet")}. Can also
be set globally with \code{options(nflreadr.prefer)}}
}
\value{
A tibble of team statistics that aims to match NFL official box scores
and season summaries
}
\description{
Load Team Level Stats
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({
  load_team_stats()
})
}

}
\seealso{
\url{https://nflreadr.nflverse.com/articles/dictionary_team_stats.html}
for a web version of the data dictionary

\code{\link{dictionary_team_stats}} for the data dictionary
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{dictionary_pfr_passing}
\alias{dictionary_pfr_passing}
\title{Data Dictionary: PFR Passing}
\format{
An object of class \code{data.frame} with 28 rows and 3 columns.
}
\usage{
dictionary_pfr_passing
}
\description{
A dataframe containing the data dictionary for \code{\link[=load_pfr_passing]{load_pfr_passing()}}
}
\seealso{
\url{https://nflreadr.nflverse.com/articles/dictionary_pfr_passing.html}

\code{vignette("Data Dictionary - PFR Passing")}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_sitrep.R
\name{.sitrep_pkg_opts}
\alias{.sitrep_pkg_opts}
\title{Show options for installed packages}
\usage{
.sitrep_pkg_opts(packages, redact_path = TRUE)
}
\description{
Show options for installed packages
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/from_url.R
\name{raw_from_url}
\alias{raw_from_url}
\title{Load raw filedata from a remote connection}
\usage{
raw_from_url(url)
}
\arguments{
\item{url}{a character url}
}
\value{
a raw vector
}
\description{
This function allows you to retrieve data from a URL into raw format, which
can then be passed into the appropriate file-reading function. Data is memoised/cached for 24 hours.
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({ # prevents CRAN errors
head(raw_from_url(
  "https://github.com/nflverse/nflverse-data/releases/download/test/combines.rds"
  ),
50)
})
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_contracts.R
\name{load_contracts}
\alias{load_contracts}
\title{Load Historical Player Contracts from OverTheCap.com}
\usage{
load_contracts(file_type = getOption("nflreadr.prefer", default = "rds"))
}
\arguments{
\item{file_type}{One of \code{"rds"}, \code{"csv"}, or \code{"parquet"}.
Can also be set globally with options(nflreadr.prefer)}
}
\value{
A tibble of active and non-active NFL player contracts.
}
\description{
Loads player contracts from OverTheCap.com
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({ # prevents cran errors
  load_contracts()
})
}

}
\seealso{
\url{https://overthecap.com/contract-history} for a web version of the data

\url{https://nflreadr.nflverse.com/articles/dictionary_contracts.html} for a web version of the dictionary

\code{\link{dictionary_contracts}} for the data dictionary as bundled within the package

Issues with this data should be filed here: \url{https://github.com/nflverse/rotc}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_ffverse.R
\name{load_ff_rankings}
\alias{load_ff_rankings}
\title{Load Latest FantasyPros Rankings}
\usage{
load_ff_rankings(type = c("draft", "week", "all"))
}
\arguments{
\item{type}{one of "draft" (preseason), "week" (this week, inseason), or "all" (full archive)}
}
\value{
a dataframe of expert consensus rankings
}
\description{
Accesses DynastyProcess.com's repository of the latest FP expert consensus rankings - updated on a weekly basis.
}
\examples{
\donttest{
try({ # prevents cran errors
load_ff_rankings()
})
}

}
\seealso{
\url{https://nflreadr.nflverse.com/articles/dictionary_ff_rankings.html} for the web data dictionary

\url{https://www.fantasypros.com} for the source of data

Issues with this data should be filed here: \url{https://github.com/dynastyprocess/data}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_snap_counts.R
\name{load_snap_counts}
\alias{load_snap_counts}
\title{Load Snap Counts from PFR}
\usage{
load_snap_counts(
  seasons = most_recent_season(),
  file_type = getOption("nflreadr.prefer", default = "rds")
)
}
\arguments{
\item{seasons}{a numeric vector specifying what seasons to return, if \code{TRUE} returns all available data}

\item{file_type}{One of \code{c("rds", "csv", "parquet")}. Can also be set globally with
\code{options(nflreadr.prefer)}}
}
\value{
A tibble of game-level snap counts provided by Pro Football Reference.
}
\description{
Loads game level snap counts stats provided by Pro Football Reference
starting with the 2012 season.
}
\examples{
\dontshow{.for_cran()}
\donttest{
  try({ # prevents CRAN errors
  load_snap_counts()
  })
}

}
\seealso{
\url{https://nflreadr.nflverse.com/articles/dictionary_snap_counts.html} for the web data dictionary

\code{\link{dictionary_snap_counts}} for the data dictionary as bundled within the package

Issues with this data should be filed here: \url{https://github.com/nflverse/nflverse-pfr}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_draft_picks.R
\name{load_draft_picks}
\alias{load_draft_picks}
\title{Load Draft Picks from PFR}
\usage{
load_draft_picks(
  seasons = TRUE,
  file_type = getOption("nflreadr.prefer", default = "rds")
)
}
\arguments{
\item{seasons}{a numeric vector of seasons to return, default \code{TRUE} returns all available data}

\item{file_type}{One of \code{c("rds", "csv", "parquet")}. Can also be set globally with
\code{options(nflreadr.prefer)}}
}
\value{
A tibble of NFL draft picks provided by Pro Football Reference.
}
\description{
Loads every draft pick since 1980 courtesy of PFR.
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({ # prevents cran errors
  load_draft_picks()
})
}

}
\seealso{
\url{https://nflreadr.nflverse.com/articles/dictionary_draft_picks.html} for the web data dictionary

\code{\link{dictionary_draft_picks}} for the data dictionary as bundled within the package

Issues with this data should be filed here: \url{https://github.com/nflverse/nflverse-data}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_officials.R
\name{load_officials}
\alias{load_officials}
\title{Load Officials}
\usage{
load_officials(
  seasons = TRUE,
  file_type = getOption("nflreadr.prefer", default = "rds")
)
}
\arguments{
\item{seasons}{a numeric vector specifying what seasons to return, if \code{TRUE} returns all available data}

\item{file_type}{One of \code{c("rds", "csv", "parquet")}. Can also be set globally with
\code{options(nflreadr.prefer)}}
}
\value{
A tibble with one row per game per official.
}
\description{
Loads data on which officials are assigned to oversee a specific game. Data available from 2015 onwards.
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({ # prevents cran errors
  load_officials()
})
}

}
\seealso{
Issues with this data should be filed here: \url{https://github.com/nflverse/nflreadr} and it will be triaged appropriately.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_rosters.R
\name{load_rosters}
\alias{load_rosters}
\title{Load Rosters}
\usage{
load_rosters(
  seasons = most_recent_season(roster = TRUE),
  file_type = getOption("nflreadr.prefer", default = "rds")
)
}
\arguments{
\item{seasons}{a numeric vector of seasons to return, defaults to returning
this year's data if it is March or later. If set to \code{TRUE}, will return all available data.
Data available back to 1920.}

\item{file_type}{One of \code{c("rds", "csv", "parquet")}. Can also be set globally with
\code{options(nflreadr.prefer)}}
}
\value{
A tibble of season-level roster data.
}
\description{
Load Rosters
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({ # prevents cran errors
  load_rosters(2020)
})
}

}
\seealso{
\url{https://nflreadr.nflverse.com/articles/dictionary_rosters.html} for a web version of the data dictionary

\code{\link{dictionary_rosters}} for the data dictionary as a dataframe

Issues with this data should be filed here: \url{https://github.com/nflverse/nflverse-data}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_combine.R
\name{load_combine}
\alias{load_combine}
\title{Load Combine Data from PFR}
\usage{
load_combine(
  seasons = TRUE,
  file_type = getOption("nflreadr.prefer", default = "rds")
)
}
\arguments{
\item{seasons}{a numeric vector of seasons to return, default \code{TRUE} returns all available data}

\item{file_type}{One of \code{c("rds", "csv", "parquet")}. Can also be set globally with
\code{options(nflreadr.prefer)}}
}
\value{
A tibble of NFL combine data provided by Pro Football Reference.
}
\description{
Loads combine data since 2000 courtesy of PFR.
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({ # prevents cran errors
  load_combine()
})
}

}
\seealso{
Issues with this data should be filed here: \url{https://github.com/nflverse/nflverse-data}

\url{https://nflreadr.nflverse.com/articles/dictionary_combine.html} for a web version of the dictionary

\code{\link{dictionary_combine}} for the data dictionary as bundled within the package
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_pfr_advstats.R
\name{load_pfr_passing}
\alias{load_pfr_passing}
\title{Load Advanced Passing Stats from PFR}
\usage{
load_pfr_passing(seasons = TRUE)
}
\arguments{
\item{seasons}{a numeric vector specifying what seasons to return, if \code{TRUE} returns all available data}
}
\value{
A tibble of season-level player statistics provided by Pro Football Reference.
}
\description{
Loads player level season stats provided by Pro Football Reference
starting with the 2019 season.
}
\examples{
\dontshow{.for_cran()}
\donttest{
  load_pfr_passing()
}

}
\seealso{
\url{https://nflreadr.nflverse.com/articles/dictionary_pfr_passing.html} for the web data dictionary

\code{\link{dictionary_pfr_passing}} for the data dictionary as bundled within the package

Issues with this data should be filed here: \url{https://github.com/nflverse/nflverse-pfr}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_download_nflverse.R
\name{nflverse_releases}
\alias{nflverse_releases}
\title{List all available nflverse releases}
\usage{
nflverse_releases(.token = "default")
}
\arguments{
\item{.token}{a GitHub API token, \code{"default"} uses \code{gh::gh_token()}}
}
\value{
A dataframe containing release names, release descriptions, and
other relevant release information.
}
\description{
This functions lists all nflverse data releases that are available
in the nflverse-data repo. Release names can be used for downloads in
\code{\link[=nflverse_download]{nflverse_download()}}.
}
\examples{
\dontshow{
.for_cran()
# Change option for better output
old <- options(piggyback.verbose = FALSE)
}
\donttest{

try( # avoids cran failures, can skip in normal usage
nflverse_releases()
)
}
\dontshow{
# Restore old options
options(old)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/from_url.R
\name{load_from_url}
\alias{load_from_url}
\title{Load any rds/csv/csv.gz/parquet file from a remote URL}
\usage{
load_from_url(url, ..., seasons = TRUE, nflverse = FALSE)
}
\arguments{
\item{url}{a vector of URLs to load into memory. If more than one URL provided, will row-bind them.}

\item{...}{named arguments that will be added as attributes to the data, e.g. \code{nflverse_type} = "pbp"}

\item{seasons}{a numeric vector of years that will be used to filter the dataframe's \code{season} column. If \code{TRUE} (default), does not filter.}

\item{nflverse}{TRUE to add nflverse_data classing and attributes.}
}
\value{
a dataframe, possibly of type \code{nflverse_data}
}
\description{
Load any rds/csv/csv.gz/parquet file from a remote URL
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({ # prevents cran errors
  urls <- c("https://github.com/nflverse/nflverse-data/releases/download/rosters/roster_2020.csv",
            "https://github.com/nflverse/nflverse-data/releases/download/rosters/roster_2021.csv")
 load_from_url(urls, nflverse = TRUE, nflverse_type = "rosters for 2020 & 2021")
})
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_sitrep.R
\name{nflverse_sitrep-class}
\alias{nflverse_sitrep-class}
\title{nflverse_sitrep data class}
\description{
\code{nflverse_sitrep()} returns an S3 class that was created to allow for custom
printing. It will otherwise dispatch to the \code{list} class.
}
\details{
Re-exporting with \code{methods::setOldClass()} allows these dataframes to be used
in S4 generics like those from DBI and jsonlite.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_espn_qbr.R
\name{load_espn_qbr}
\alias{load_espn_qbr}
\title{Load ESPN's QBR}
\usage{
load_espn_qbr(
  seasons = most_recent_season(),
  summary_type = c("season", "week"),
  file_type = getOption("nflreadr.prefer", default = "rds")
)
}
\arguments{
\item{seasons}{a numeric vector of seasons to return, data available since 2006. Defaults to latest season available. TRUE will select all seasons.}

\item{summary_type}{One of \code{"season"} or \code{"week"}, defaults to \code{"season"}}

\item{file_type}{One of \code{c("rds", "csv", "parquet")}. Can also be set globally with
\code{options(nflreadr.prefer)}}
}
\value{
a tibble of ESPN QBR data, summarized according to \code{summary_type}
}
\description{
Load ESPN's QBR
}
\examples{
\dontshow{.for_cran()}
\donttest{
    load_espn_qbr(2020)
}

}
\seealso{
\url{https://nflreadr.nflverse.com/articles/dictionary_espn_qbr.html} for a web version of the dictionary

\code{\link{dictionary_espn_qbr}} for the data dictionary as bundled within the package

Issues with this data should be filed here: \url{https://github.com/nflverse/espnscrapeR-data}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_ftn_charting.R
\name{load_ftn_charting}
\alias{load_ftn_charting}
\title{Load FTN Charting Data}
\source{
FTNData.com
}
\usage{
load_ftn_charting(
  seasons = most_recent_season(),
  file_type = getOption("nflreadr.prefer", default = "rds")
)
}
\arguments{
\item{seasons}{a numeric vector of seasons to return, defaults to most recent season.
If set to \code{TRUE}, returns all available data. Data available from 2022 onwards.}

\item{file_type}{One of \code{c("rds", "csv", "parquet")}. Can also be set globally with
\code{options(nflreadr.prefer)}}
}
\value{
Play-level manual charting data from FTN Data
}
\description{
FTN Data manually charts plays and has graciously provided a subset of their
charting data to be published via the nflverse. Data is available from the 2022
season onwards and is charted within 48 hours following each game. This data
is released under the \href{https://creativecommons.org/licenses/by-sa/4.0/}{CC-BY-SA 4.0}
Creative Commons license and attribution must be made to \strong{FTN Data via nflverse}
}
\examples{
\donttest{
  try({ # prevents cran errors
  load_ftn_charting()
  })
}

}
\seealso{
\url{https://ftnfantasy.com/stats/sports-data}

\code{vignette("Data Dictionary - FTN Charting")}

\url{https://nflreadr.nflverse.com/articles/dictionary_ftn_charting.html} for the web data dictionary

Other ftn_charting: 
\code{\link{dictionary_ftn_charting}}
}
\author{
FTN Data
}
\concept{ftn_charting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_nflverse_data_class.R
\name{nflverse_data-class}
\alias{nflverse_data-class}
\alias{as.nflverse_data}
\alias{make_nflverse_data}
\title{nflverse data class}
\usage{
as.nflverse_data(df, nflverse_type = NULL, ...)
}
\description{
This class has a special S3 print method that tries to read attached metadata and
provide timestamps and source attributes. It otherwise will dispatch to the
\code{data.frame} class.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{player_name_mapping}
\alias{player_name_mapping}
\title{Alternate player name mappings}
\format{
A named character vector
\describe{
\item{name attribute}{The "alternate" name.}
\item{value attribute}{The "correct" name.}
}
}
\usage{
player_name_mapping
}
\description{
A named character vector mapping common alternate names, re-exported from \code{ffscrapr}.
}
\details{
You can suggest additions to this table by \href{https://github.com/ffverse/ffscrapr/issues/new/choose}{opening an issue in ffscrapr}.
}
\examples{
\donttest{
player_name_mapping[c("Chatarius Atwell", "Robert Kelley")]
}

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_gameid.R
\name{nflverse_game_id}
\alias{nflverse_game_id}
\title{Compute nflverse Game Identifiers}
\usage{
nflverse_game_id(season, week, away, home)
}
\arguments{
\item{season}{4 digit season between 1999 and the output of \code{\link[=most_recent_season]{most_recent_season()}}}

\item{week}{Numeric or character giving the week, between 1 and 22.}

\item{home, away}{Valid NFL team abbreviation as it can be found in \link{team_abbr_mapping}}
}
\value{
A character vector
}
\description{
Compute nflverse Game Identifiers
}
\examples{
\dontshow{.for_cran()}
nflverse_game_id(2022, 2, "LAC", "KC")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/from_url.R
\name{rds_from_url}
\alias{rds_from_url}
\title{Load .rds file from a remote connection}
\usage{
rds_from_url(url)
}
\arguments{
\item{url}{a character url}
}
\value{
a dataframe as created by \code{\link[=readRDS]{readRDS()}}
}
\description{
Load .rds file from a remote connection
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({ # prevents cran errors
  rds_from_url("https://github.com/nflverse/nflverse-data/releases/download/test/combines.rds")
})
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_name_cleaning.R
\name{clean_team_abbrs}
\alias{clean_team_abbrs}
\title{Standardize NFL Team Abbreviations}
\usage{
clean_team_abbrs(abbr, current_location = TRUE, keep_non_matches = TRUE)
}
\arguments{
\item{abbr}{a character vector of abbreviations, full team names, or team nicknames}

\item{current_location}{If \code{TRUE} (the default), the abbreviation of the most
recent team location will be used.}

\item{keep_non_matches}{If \code{TRUE} (the default) an element of \code{abbr} that can't
be matched to any of the internal mapping vectors will be kept as is. Otherwise
it will be replaced with \code{NA}.}
}
\value{
A character vector with the length of \code{abbr} and cleaned team abbreviations
if they are included in \code{\link{team_abbr_mapping}} or \code{\link{team_abbr_mapping_norelocate}}
(depending on the value of \code{current_location}). Non matches may be replaced
with \code{NA} (depending on the value of \code{keep_non_matches}).
}
\description{
This function standardizes NFL team names, abbreviations, and nicknames to
nflverse defaults.
This helps for joins and plotting, especially with the nflplotR package!
}
\examples{
\dontshow{.for_cran()}
x <- c("PIE", "LAR", "PIT", "CRD", "OAK", "SL", "Niners", "San Diego Chargers")
# use current location and keep non matches
clean_team_abbrs(x)

# keep old location and replace non matches
clean_team_abbrs(x, current_location = FALSE, keep_non_matches = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{team_abbr_mapping}
\alias{team_abbr_mapping}
\title{Alternate team abbreviation mappings}
\format{
A named character vector
\describe{
\item{name attribute}{The "alternate" name.}
\item{value attribute}{The "correct" name.}
}
}
\usage{
team_abbr_mapping
}
\description{
A named character vector mapping common alternate team abbreviations.
}
\details{
You can suggest additions to this table by \href{https://github.com/nflverse/nflreadr/issues/new/choose}{opening an issue in nflreadr}.
}
\examples{
\donttest{
team_abbr_mapping[c("STL", "OAK","CRD","BLT", "CLV")]
}

}
\seealso{
\code{team_abbr_mapping_norelocate} for the same thing but relocations stay in their original cities.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_pfr_advstats.R
\name{load_pfr_advstats}
\alias{load_pfr_advstats}
\title{Load Advanced Stats from PFR}
\usage{
load_pfr_advstats(
  seasons = most_recent_season(),
  stat_type = c("pass", "rush", "rec", "def"),
  summary_level = c("week", "season"),
  file_type = getOption("nflreadr.prefer", default = "rds")
)
}
\arguments{
\item{seasons}{a numeric vector specifying what seasons to return, if \code{TRUE} returns all available data}

\item{stat_type}{one of "pass", "rush", "rec", "def"}

\item{summary_level}{one of "week" (default) or "season" - some data is only available at the season level}

\item{file_type}{One of \code{c("rds", "csv", "parquet")}. Can also be set globally with
\code{options(nflreadr.prefer)}}
}
\value{
A tibble of player statistics provided by Pro Football Reference that supplements data in nflverse
}
\description{
Loads player level season stats provided by Pro Football Reference
starting with the 2018 season, primarily to augment existing nflverse data.
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({ # prevents cran errors
  load_pfr_advstats()
})
}

}
\seealso{
\url{https://nflreadr.nflverse.com/articles/dictionary_pfr_passing.html} for the web data dictionary

\url{https://www.pro-football-reference.com/years/2021/passing_advanced.htm}

Issues with this data should be filed here: \url{https://github.com/nflverse/nflverse-data}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_ffverse.R
\name{load_ff_playerids}
\alias{load_ff_playerids}
\title{Load Fantasy Player IDs}
\usage{
load_ff_playerids()
}
\value{
a dataframe of player IDs
}
\description{
Accesses DynastyProcess.com's database of fantasy football player IDs, which help connect nflverse to various other platforms and IDs.
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({ # prevents cran errors
load_ff_playerids()
})
}

}
\seealso{
\url{https://nflreadr.nflverse.com/articles/dictionary_ff_playerids.html} for the web data dictionary

Issues with this data should be filed here: \url{https://github.com/dynastyprocess/data}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_download_nflverse.R
\name{nflverse_download}
\alias{nflverse_download}
\title{Bulk download utilities via piggyback}
\usage{
nflverse_download(
  ...,
  folder_path = getOption("nflreadr.download_path", default = "."),
  file_type = getOption("nflreadr.prefer", default = "rds"),
  use_hive = file_type \%in\% c("parquet", "csv"),
  .token = "default"
)
}
\arguments{
\item{...}{releases to download, provided in either unquoted or character format
(i.e. pbp or "pbp" are both fine). Available release names can be listed with
\code{\link[=nflverse_releases]{nflverse_releases()}}}

\item{folder_path}{a folder in which subfolders will be created for each release -
defaults to path specified in \code{options(nflreadr.download_path)} or "." (the current working directory)}

\item{file_type}{one of \code{c("rds","parquet", "csv")} -
defaults to file type specified in \code{options(nflreadr.prefer)} or "rds"}

\item{use_hive}{whether to create hive-style partition folders for each season, e.g. \code{"~/pbp/.season=2021/pbp.csv"}}

\item{.token}{a GitHub API token, \code{"default"} uses \code{gh::gh_token()}}
}
\description{
This function downloads or updates data from the nflverse-data repository releases,
creating subfolders that match the release structure.
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({
  ## could also set options like
  # options(nflreadr.download_path = tempdir(), nflreadr.prefer = "parquet")

  nflverse_download(combine, contracts, folder_path = tempdir(), file_type = "parquet")

  list.files(tempdir(),pattern = ".parquet$") # check that files were downloaded!
})
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_depth_charts.R
\name{load_depth_charts}
\alias{load_depth_charts}
\title{Load Weekly Depth Charts}
\usage{
load_depth_charts(
  seasons = most_recent_season(roster = TRUE),
  file_type = getOption("nflreadr.prefer", default = "rds")
)
}
\arguments{
\item{seasons}{a numeric vector specifying what seasons to return, if \code{TRUE} returns all available data. Defaults to latest season with available rosters.}

\item{file_type}{One of \code{c("rds", "csv", "parquet")}. Can also be set globally with
\code{options(nflreadr.prefer)}}
}
\value{
A tibble of week-level depth charts for each team.
}
\description{
Loads depth charts for each NFL team for each week back to 2001.
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({ # prevents cran errors
  load_depth_charts(2020)
})
}

}
\seealso{
\url{https://nflreadr.nflverse.com/articles/dictionary_depth_charts.html} for a web version of the dictionary

\code{\link{dictionary_depth_charts}} for the data dictionary as bundled within the package

Issues with this data should be filed here: \url{https://github.com/nflverse/nflverse-data}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/from_url.R
\name{parquet_from_url}
\alias{parquet_from_url}
\title{Load .parquet file from a remote connection}
\usage{
parquet_from_url(url)
}
\arguments{
\item{url}{a character url}
}
\value{
a dataframe as parsed by \code{\link[arrow:read_parquet]{arrow::read_parquet()}}
}
\description{
Retrieves a parquet file from URL. This function is cached
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({
  parquet_from_url(
    "https://github.com/nflverse/nflverse-data/releases/download/player_stats/player_stats.parquet"
  )
})
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_join_coalesce.R
\name{join_coalesce}
\alias{join_coalesce}
\title{Coalescing join}
\usage{
join_coalesce(
  x,
  y,
  by = NULL,
  type = c("left", "inner", "full"),
  ...,
  by.x = NULL,
  by.y = NULL,
  sort = TRUE,
  incomparables = c(NA, NaN)
)
}
\arguments{
\item{x, y}{dataframes. Will be coerced to data.table}

\item{by}{keys to join on, as a plain or named character vector}

\item{type}{one of "left" (all rows of x and matching rows of y),
"inner" (matching rows of x and y), "full" (all rows of x and y)}

\item{...}{other args passed to \code{merge.data.frame()}}

\item{by.x, by.y}{alternate form of keys to join on - if provided, will override \code{by}.}

\item{sort}{whether to sort output by the join keys}

\item{incomparables}{keys to NOT match on, i.e. NA should not match on NA.}
}
\value{
a data.frame joining x and y dataframes together, with every column from
both x and y and patching NA values in x with those in y.
}
\description{
EXPERIMENTAL! This function joins two dataframes together by key, and then coalesces any
columns that have shared names (i.e. fills in NAs). A utility function primarily
used internally within nflverse to help build player IDs
}
\examples{
\dontshow{.for_cran()}
x <- data.frame(id1 = c(NA_character_,letters[1:4]), a = c(1,NA,3,NA,5), b = 1:5 * 10)
y <- data.frame(id2 = c(letters[3:11],NA_character_), a = -(1:10), c = 1:10 * 100)

join_coalesce(x,y, by = c("id1"="id2"))
join_coalesce(x,y, by.x = "id1", by.y = "id2")
join_coalesce(x,y, by = c("id1"="id2"), type = "inner")
join_coalesce(x,y, by = c("id1"="id2"), type = "full")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_ffverse.R
\name{load_ff_opportunity}
\alias{load_ff_opportunity}
\title{Load Expected Fantasy Points}
\usage{
load_ff_opportunity(
  seasons = most_recent_season(),
  stat_type = c("weekly", "pbp_pass", "pbp_rush"),
  model_version = c("latest", "v1.0.0")
)
}
\arguments{
\item{seasons}{a numeric vector of seasons to return, defaults to most recent season. If set to \code{TRUE}, returns all available data.}

\item{stat_type}{one of \code{"weekly"}, \code{"pbp_pass"}, \code{"pbp_rush"}}

\item{model_version}{one of \code{"latest"} or \code{"v1.0.0"}}
}
\value{
Precomputed expected fantasy points data from the ffopportunity automated releases.
}
\description{
This function downloads precomputed expected points data from \href{https://ffopportunity.ffverse.com}{{ffopportunity}} automated releases.
}
\examples{
\donttest{
  try({ # prevents cran errors
  load_ff_opportunity()
  load_ff_opportunity(seasons = 2021, stat_type = "pbp_pass", model_version = "v1.0.0")
  })
}

}
\seealso{
\url{https://ffopportunity.ffverse.com} for more on the package, data, and modelling

\url{https://nflreadr.nflverse.com/articles/dictionary_ff_opportunity.html} for the web data dictionary

\code{\link{dictionary_ff_opportunity}} for the data dictionary bundled as a package data frame

Issues with this data should be filed here: \url{https://github.com/ffverse/ffopportunity}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_sitrep.R
\name{.sitrep_pkg_deps}
\alias{.sitrep_pkg_deps}
\title{Show dependency versions of installed packages}
\usage{
.sitrep_pkg_deps(
  packages,
  inst_pkgs = utils::installed.packages(),
  recursive = TRUE
)
}
\description{
Show dependency versions of installed packages
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_injuries.R
\name{load_injuries}
\alias{load_injuries}
\title{Load Injury Reports}
\usage{
load_injuries(
  seasons = most_recent_season(),
  file_type = getOption("nflreadr.prefer", default = "rds")
)
}
\arguments{
\item{seasons}{a numeric vector of seasons to return, data available since 2009. Defaults to latest season available.}

\item{file_type}{One of \code{c("rds", "csv", "parquet")}. Can also be set globally with
\code{options(nflreadr.prefer)}}
}
\value{
a tibble of season-level injury report data.
}
\description{
Data collected from an API for weekly injury report data.
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({# prevents cran errors
    load_injuries(2020)
})
}

}
\seealso{
\url{https://nflreadr.nflverse.com/articles/dictionary_injuries.html} for a web version of the dictionary

\code{\link{dictionary_injuries}} for the data dictionary as bundled within the package

Issues with this data should be filed here: \url{https://github.com/nflverse/nflverse-data}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_cran.R
\name{.for_cran}
\alias{.for_cran}
\title{CRAN setup code}
\usage{
.for_cran()
}
\description{
Some code to comply with CRAN policies, e.g.
limiting data.table CPU cores for examples/tests
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_date.R
\name{most_recent_season}
\alias{most_recent_season}
\alias{get_latest_season}
\alias{get_current_season}
\title{Get Latest Season}
\usage{
most_recent_season(roster = FALSE)

get_latest_season(roster = FALSE)

get_current_season(roster = FALSE)
}
\arguments{
\item{roster}{Either \code{TRUE} or \code{FALSE}.
If \code{TRUE}, will return current year after March 15th, otherwise previous year.
If \code{FALSE}, will return current year on or after Wednesday following Labor Day,
i.e. Wednesday after the first Monday in September. Otherwise previous year.}
}
\value{
most recent season (a four digit numeric)
}
\description{
A helper function to choose the most recent season available for a given dataset.
}
\details{
This used to return new season on Thursday following Labor Day but NFL
announced in March 2026 that the 2026 season will kick off on
Wednesday because the TNF game will be played in Australia. We expect this
to happen in future seasons as well and decided to flip the switch on
Wednesday going forward.
}
\seealso{
Other Date utils: 
\code{\link{get_current_week}()}
}
\concept{Date utils}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_date.R
\name{get_current_week}
\alias{get_current_week}
\title{Get Current Week}
\usage{
get_current_week(use_date = FALSE, ...)
}
\arguments{
\item{use_date}{a logical to determine whether to use date-based heuristics to
determine current week, default FALSE (i.e. uses schedule file)}

\item{...}{Arguments passed on to \code{\link[=most_recent_season]{most_recent_season()}}}
}
\value{
current nfl regular season week as a numeric
}
\description{
A helper function that returns the upcoming NFL regular season week based on either
the nflverse schedules file (as found in \code{load_schedules()})
or some date-based heuristics (number of weeks since the first Monday of September)
}
\details{
Note that the date heuristic will count a new week starting on Thursdays, while
the schedule-based method will count a new week after the last game of the previous
week, e.g. after MNF is completed. Tan and Ben argued for a while about this.
}
\examples{
{
  \donttest{
    try({ # schedules file as per default requires online access
    get_current_week()
    })
  }
  # using the date method works offline
  get_current_week(use_date = TRUE)
}

}
\seealso{
Other Date utils: 
\code{\link{most_recent_season}()}
}
\concept{Date utils}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{dictionary_roster_status}
\alias{dictionary_roster_status}
\title{dictionary_roster_status}
\format{
An object of class \code{data.frame} with 19 rows and 2 columns.
}
\usage{
dictionary_roster_status
}
\description{
A dictionary translating the shorthand roster status to more verbose explanations of what each status indicates.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/from_url.R
\name{csv_from_url}
\alias{csv_from_url}
\title{Load .csv / .csv.gz file from a remote connection}
\usage{
csv_from_url(...)
}
\arguments{
\item{...}{
  Arguments passed on to \code{\link[data.table:fread]{data.table::fread}}
  \describe{
    \item{\code{input}}{ A single character string. The value is inspected and deferred to either \code{file=} (if no \\n present), \code{text=} (if at least one \\n is present) or \code{cmd=} (if no \\n is present, at least one space is present, and it isn't a file name). Exactly one of \code{input=}, \code{file=}, \code{text=}, or \code{cmd=} should be used in the same call. }
    \item{\code{file}}{ File name in working directory, path to file (passed through \code{\link[base]{path.expand}} for convenience), or a URL starting http://, file://, etc. Compressed files with extension \file{.gz} and \file{.bz2} are supported if the \code{R.utils} package is installed. }
    \item{\code{text}}{ The input data itself as a character vector of one or more lines, for example as returned by \code{readLines()}. }
    \item{\code{cmd}}{ A shell command that pre-processes the file; e.g. \code{fread(cmd=paste("grep",word,"filename"))}. See Details. }
    \item{\code{sep}}{ The separator between columns. Defaults to the character in the set \code{[,\\t |;:]} that separates the sample of rows into the most number of lines with the same number of fields. Use \code{NULL} or \code{""} to specify no separator; i.e. each line a single character column like \code{base::readLines} does.}
    \item{\code{sep2}}{ The separator \emph{within} columns. A \code{list} column will be returned where each cell is a vector of values. This is much faster using less working memory than \code{strsplit} afterwards or similar techniques. For each column \code{sep2} can be different and is the first character in the same set above [\code{,\\t |;}], other than \code{sep}, that exists inside each field outside quoted regions in the sample. NB: \code{sep2} is not yet implemented. }
    \item{\code{nrows}}{ The maximum number of rows to read. Unlike \code{read.table}, you do not need to set this to an estimate of the number of rows in the file for better speed because that is already automatically determined by \code{fread} almost instantly using the large sample of lines. \code{nrows=0} returns the column names and typed empty columns determined by the large sample; useful for a dry run of a large file or to quickly check format consistency of a set of files before starting to read any of them. }
    \item{\code{header}}{ Does the first data line contain column names? Defaults according to whether every non-empty field on the first data line is type character. If so, or TRUE is supplied, any empty column names are given a default name. }
    \item{\code{na.strings}}{ A character vector of strings which are to be interpreted as \code{NA} values. By default, \code{",,"} for columns of all types, including type \code{character} is read as \code{NA} for consistency. \code{,"",} is unambiguous and read as an empty string. To read \code{,NA,} as \code{NA}, set \code{na.strings="NA"}. To read \code{,,} as blank string \code{""}, set \code{na.strings=NULL}. When they occur in the file, the strings in \code{na.strings} should not appear quoted since that is how the string literal \code{,"NA",} is distinguished from \code{,NA,}, for example, when \code{na.strings="NA"}. }
    \item{\code{stringsAsFactors}}{ Convert all or some character columns to factors? Acceptable inputs are \code{TRUE}, \code{FALSE}, or a decimal value between 0.0 and 1.0. For \code{stringsAsFactors = FALSE}, all string columns are stored as \code{character} vs. all stored as \code{factor} when \code{TRUE}. When \code{stringsAsFactors = p} for \code{0 <= p <= 1}, string columns \code{col} are stored as \code{factor} if \code{uniqueN(col)/nrow < p}. 
  }
    \item{\code{verbose}}{ Be chatty and report timings? }
    \item{\code{skip}}{ If 0 (default) start on the first line and from there finds the first row with a consistent number of columns. This automatically avoids irregular header information before the column names row. \code{skip>0} means ignore the first \code{skip} rows manually. \code{skip="string"} searches for \code{"string"} in the file (e.g. a substring of the column names row) and starts on that line (inspired by read.xls in package gdata). }
    \item{\code{select}}{ A vector of column names or numbers to keep, drop the rest. \code{select} may specify types too in the same way as \code{colClasses}; i.e., a vector of \code{colname=type} pairs, or a \code{list} of \code{type=col(s)} pairs. In all forms of \code{select}, the order that the columns are specified determines the order of the columns in the result. }
    \item{\code{drop}}{ Vector of column names or numbers to drop, keep the rest. }
    \item{\code{colClasses}}{ As in \code{\link[utils:read.table]{utils::read.csv}}; i.e., an unnamed vector of types corresponding to the columns in the file, or a named vector specifying types for a subset of the columns by name. The default, \code{NULL} means types are inferred from the data in the file. Further, \code{data.table} supports a named \code{list} of vectors of column names \emph{or numbers} where the \code{list} names are the class names; see examples. The \code{list} form makes it easier to set a batch of columns to be a particular class. When column numbers are used in the \code{list} form, they refer to the column number in the file not the column number after \code{select} or \code{drop} has been applied.
    If type coercion results in an error, introduces \code{NA}s, or would result in loss of accuracy, the coercion attempt is aborted for that column with warning and the column's type is left unchanged. If you really desire data loss (e.g. reading \code{3.14} as \code{integer}) you have to truncate such columns afterwards yourself explicitly so that this is clear to future readers of your code.
  }
    \item{\code{integer64}}{ "integer64" (default) reads columns detected as containing integers larger than 2^31 as type \code{bit64::integer64}. Alternatively, \code{"double"|"numeric"} reads as \code{utils::read.csv} does; i.e., possibly with loss of precision and if so silently. Or, "character". }
    \item{\code{dec}}{ The decimal separator as in \code{utils::read.csv}. When \code{"auto"} (the default), an attempt is made to decide whether \code{"."} or \code{","} is more suitable for this input. See details. }
    \item{\code{col.names}}{ A vector of optional names for the variables (columns). The default is to use the header column if present or detected, or if not "V" followed by the column number. This is applied after \code{check.names} and before \code{key} and \code{index}. }
    \item{\code{check.names}}{default is \code{FALSE}. If \code{TRUE} then the names of the variables in the \code{data.table} are checked to ensure that they are syntactically valid variable names. If necessary they are adjusted (by \code{\link{make.names}}) so that they are, and also to ensure that there are no duplicates.}
    \item{\code{encoding}}{ default is \code{"unknown"}. Other possible options are \code{"UTF-8"} and \code{"Latin-1"}.  Note: it is not used to re-encode the input, rather enables handling of encoded strings in their native encoding. }
    \item{\code{quote}}{ By default (\code{"\\""}), if a field starts with a double quote, \code{fread} handles embedded quotes robustly as explained under \code{Details}. If it fails, then another attempt is made to read the field \emph{as is}, i.e., as if quotes are disabled. By setting \code{quote=""}, the field is always read as if quotes are disabled. It is not expected to ever need to pass anything other than \\"\\" to quote; i.e., to turn it off. }
    \item{\code{strip.white}}{ Logical, default \code{TRUE}, in which case leading and trailing whitespace is stripped from unquoted \code{"character"} fields. \code{"numeric"} fields are always stripped of leading and trailing whitespace.}
    \item{\code{fill}}{logical or integer (default is \code{FALSE}). If \code{TRUE} then in case the rows have unequal length, number of columns is estimated and blank fields are implicitly filled. If an integer is provided it is used as an upper bound for the number of columns. If \code{fill=Inf} then the whole file is read for detecting the number of columns. }
    \item{\code{blank.lines.skip}}{\code{logical}, default is \code{FALSE}. If \code{TRUE} blank lines in the input are ignored.}
    \item{\code{key}}{Character vector of one or more column names which is passed to \code{\link[data.table]{setkey}}. Only valid when argument \code{data.table=TRUE}. Where applicable, this should refer to column names given in \code{col.names}. }
    \item{\code{index}}{ Character vector or list of character vectors of one or more column names which is passed to \code{\link[data.table]{setindexv}}. As with \code{key}, comma-separated notation like \code{index="x,y,z"} is accepted for convenience. Only valid when argument \code{data.table=TRUE}. Where applicable, this should refer to column names given in \code{col.names}. }
    \item{\code{showProgress}}{ \code{TRUE} displays progress on the console if the ETA is greater than 3 seconds. It is produced in fread's C code where the very nice (but R level) txtProgressBar and tkProgressBar are not easily available. }
    \item{\code{data.table}}{ TRUE returns a \code{data.table}. FALSE returns a \code{data.frame}. The default for this argument can be changed with \code{options(datatable.fread.datatable=FALSE)}.}
    \item{\code{nThread}}{The number of threads to use. Experiment to see what works best for your data on your hardware.}
    \item{\code{logical01}}{If TRUE a column containing only 0s and 1s will be read as logical, otherwise as integer.}
    \item{\code{logicalYN}}{If TRUE a column containing only Ys and Ns will be read as logical, otherwise as character.}
    \item{\code{keepLeadingZeros}}{If TRUE a column containing numeric data with leading zeros will be read as character, otherwise leading zeros will be removed and converted to numeric.}
    \item{\code{yaml}}{ If \code{TRUE}, \code{fread} will attempt to parse (using \code{\link[yaml]{yaml.load}}) the top of the input as YAML, and further to glean parameters relevant to improving the performance of \code{fread} on the data itself. The entire YAML section is returned as parsed into a \code{list} in the \code{yaml_metadata} attribute. See \code{Details}. }
    \item{\code{autostart}}{ Deprecated. Please use \code{skip} instead. }
    \item{\code{tmpdir}}{ Directory to use as the \code{tmpdir} argument for any \code{tempfile} calls, e.g. when the input is a URL or a shell command. The default is \code{tempdir()} which can be controlled by setting \code{TMPDIR} before starting the R session; see \code{\link[base:tempfile]{base::tempdir}}. }
    \item{\code{tz}}{ Relevant to datetime values which have no Z or UTC-offset at the end, i.e. \emph{unmarked} datetime, as written by \code{\link[utils:write.table]{utils::write.csv}}. The default \code{tz="UTC"} reads unmarked datetime as UTC POSIXct efficiently. \code{tz=""} reads unmarked datetime as type character (slowly) so that \code{as.POSIXct} can interpret (slowly) the character datetimes in local timezone; e.g. by using \code{"POSIXct"} in \code{colClasses=}. Note that \code{fwrite()} by default writes datetime in UTC including the final Z and therefore \code{fwrite}'s output will be read by \code{fread} consistently and quickly without needing to use \code{tz=} or \code{colClasses=}. If the \code{TZ} environment variable is set to \code{"UTC"} (or \code{""} on non-Windows where unset vs `""` is significant) then the R session's timezone is already UTC and \code{tz=""} will result in unmarked datetimes being read as UTC POSIXct. For more information, please see the news items from v1.13.0 and v1.14.0. }
  }}
}
\value{
a dataframe as created by \code{\link[data.table:fread]{data.table::fread()}}
}
\description{
This is a thin wrapper on data.table::fread, but memoised & cached for twenty four hours.
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({ # prevents cran errors
  csv_from_url("https://github.com/nflverse/nflverse-data/releases/download/test/combines.csv")
})
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_schedules.R
\name{load_schedules}
\alias{load_schedules}
\title{Load Game/Schedule Data}
\usage{
load_schedules(seasons = TRUE)
}
\arguments{
\item{seasons}{a numeric vector of seasons to return, default \code{TRUE} returns all available data.}
}
\value{
A tibble of game information for past and/or future games.
}
\description{
This returns game/schedule information as maintained by Lee Sharpe.
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({ # prevents cran errors
 load_schedules(2020)
})
}

}
\seealso{
\url{https://nflreadr.nflverse.com/articles/dictionary_schedules.html} for a web version of the data dictionary

\code{\link{dictionary_schedules}} for the data dictionary as a dataframe

Issues with this data should be filed here: \url{https://github.com/nflverse/nfldata}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_stats.R
\name{load_player_stats}
\alias{load_player_stats}
\title{Load Player Level Stats}
\usage{
load_player_stats(
  seasons = most_recent_season(),
  ...,
  summary_level = c("week", "reg", "post", "reg+post"),
  file_type = getOption("nflreadr.prefer", default = "rds"),
  stat_type = lifecycle::deprecated()
)
}
\arguments{
\item{seasons}{a numeric vector of seasons to return, defaults to most recent
season. If set to \code{TRUE}, returns all available data.}

\item{...}{These dots are for future extensions and must be empty.}

\item{summary_level}{choice: one of week (default), "reg" for regular season,
"post" for postseason, "reg+post" for combined regular season + postseason stats}

\item{file_type}{choice: one of \code{c("rds", "csv", "parquet")}. Can also
be set globally with \code{options(nflreadr.prefer)}}

\item{stat_type}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} - now returns all stat
types by default}
}
\value{
A tibble of player statistics that aims to match NFL official box
scores and season summaries
}
\description{
Load Player Level Stats
}
\examples{
\dontshow{.for_cran()}
\donttest{
try({
  load_player_stats()
})
}

}
\seealso{
\url{https://nflreadr.nflverse.com/articles/dictionary_player_stats.html}
for a web version of the data dictionary

\code{\link{dictionary_player_stats}} for the data dictionary
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_ftn_charting.R
\docType{data}
\name{dictionary_ftn_charting}
\alias{dictionary_ftn_charting}
\title{Data Dictionary: FTN Charting Data}
\format{
An object of class \code{data.frame} with 28 rows and 5 columns.
}
\usage{
dictionary_ftn_charting
}
\description{
A dataframe containing the data dictionary for \code{\link[=load_ftn_charting]{load_ftn_charting()}}
}
\seealso{
\code{vignette("Data Dictionary - FTN Charting")}

\url{https://nflreadr.nflverse.com/articles/dictionary_ftn_charting.html}

Other ftn_charting: 
\code{\link{load_ftn_charting}()}
}
\concept{ftn_charting}
\keyword{datasets}
